/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.generator.entity.grammardefnode;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import net.sf.parser4j.generator.service.GrammarDefNodeVisitException;
import net.sf.parser4j.generator.service.grammardefnode.IGrammarDefVisitor;
import net.sf.parser4j.kernelgenerator.service.GeneratorException;

/**
 * 
 * @author luc peuvrier
 * 
 */
public class StatementDef implements IGrammarDefNode {

	/** the type of this statement */
	private EnumStatementDefType statementDefType;// NOPMD

	/** match class name */
	private String matchClassName;// NOPMD

	/** name of non terminal defined by this statement */
	private String nonTerminalDefinedName;// NOPMD

	/** true if token definition, not a token by default */
	private boolean token = false; // NOPMD

	/** true is token son, it is not defined by default */
	private Boolean tokenSon;// NOPMD

	/**
	 * non terminal name for type {@link EnumStatementDefType#NON_TERMINAL}
	 */
	private final String nonTerminalName;// NOPMD

	/** terminal string for type {@link EnumStatementDefType#TERMINAL_STRING } */
	private final String terminalString;// NOPMD

	/** terminal character for type {@link EnumStatementDefType#TERMINAL_CHAR} */
	private final char terminalCharValue;// NOPMD

	/**
	 * set of terminal for type
	 * {@link EnumStatementDefType#TERMINAL_SET_OF_TERMINAL}
	 */
	private final List<StatementDef> setOfTerminal;// NOPMD

	/**
	 * first terminal value of range for type
	 * {@link EnumStatementDefType#TERMINAL_CHAR_RANGE}
	 */
	private final char firstTerminalValue;// NOPMD

	/**
	 * last terminal value of range for type
	 * {@link EnumStatementDefType#TERMINAL_CHAR_RANGE}
	 */
	private final char lastTerminalValue;// NOPMD

	/** terminal set for type {@link EnumStatementDefType#TERMINAL_CHAR_SET} */
	private final Set<Integer> terminalSet; // NOPMD

	/**
	 * operand list for type {@link EnumStatementDefType#ALTERNATIVE} and
	 * {@link EnumStatementDefType#CONCAT}
	 */
	private List<StatementDef> operandList;// NOPMD

	/** white space non terminal name, empty string if none */
	private String whiteSpaceNonTerminalName;// NOPMD

	private StatementDef fatherStatementDef;// NOPMD

	private Exception matchClassNameSetterTrace;

	public StatementDef(final StatementDef left, final StatementDef right,
			final EnumStatementDefType statementDefType) {
		operandList = new LinkedList<StatementDef>();
		switch (statementDefType) {
		case ALTERNATIVE:
			if (EnumStatementDefType.ALTERNATIVE.equals(left
					.getStatementDefType())) {
				final List<StatementDef> leftOperandList = left
						.getOperandList();
				operandList.addAll(leftOperandList);
				for (StatementDef son : leftOperandList) {
					son.fatherStatementDef = this;
				}
			} else {
				operandList.add(left);
			}
			if (EnumStatementDefType.ALTERNATIVE.equals(right
					.getStatementDefType())) {
				final List<StatementDef> rightOperandList = right
						.getOperandList();
				operandList.addAll(rightOperandList);
				for (StatementDef son : rightOperandList) {
					son.fatherStatementDef = this;
				}
			} else {
				operandList.add(right);
			}
			break;
		case CONCAT:
			if (EnumStatementDefType.CONCAT.equals(left.getStatementDefType())) {
				final List<StatementDef> leftOperandList = left
						.getOperandList();
				operandList.addAll(leftOperandList);
				for (StatementDef son : leftOperandList) {
					son.fatherStatementDef = this;
				}
			} else {
				operandList.add(left);
			}
			if (EnumStatementDefType.CONCAT.equals(right.getStatementDefType())) {
				final List<StatementDef> rightOperandList = right
						.getOperandList();
				operandList.addAll(rightOperandList);
				for (StatementDef son : rightOperandList) {
					son.fatherStatementDef = this;
				}
			} else {
				operandList.add(right);
			}
			break;
		default:
			throw new IllegalArgumentException("bad statement type "
					+ statementDefType);
		}
		this.statementDefType = statementDefType;
		nonTerminalName = null;
		terminalString = null;
		terminalCharValue = 0;
		setOfTerminal = null;
		firstTerminalValue = 0;
		lastTerminalValue = 0;
		terminalSet = null;
	}

	public StatementDef() {
		statementDefType = EnumStatementDefType.EMPTY;

		nonTerminalName = null;
		terminalString = null;
		terminalCharValue = 0;
		setOfTerminal = null;
		firstTerminalValue = 0;
		lastTerminalValue = 0;
		terminalSet = null;
	}

	public StatementDef(final String identifier) {
		statementDefType = EnumStatementDefType.NON_TERMINAL;
		nonTerminalName = identifier;

		terminalString = null;
		terminalCharValue = 0;
		setOfTerminal = null;
		firstTerminalValue = 0;
		lastTerminalValue = 0;
		terminalSet = null;
	}

	public StatementDef(final List<Character> charList) {
		statementDefType = EnumStatementDefType.TERMINAL_STRING;
		final StringBuilder stringBuilder = new StringBuilder(charList.size());
		for (char characterValue : charList) {
			stringBuilder.append(characterValue);
		}
		terminalString = stringBuilder.toString();

		nonTerminalName = null;
		terminalCharValue = 0;
		setOfTerminal = null;
		firstTerminalValue = 0;
		lastTerminalValue = 0;
		terminalSet = null;
	}

	public StatementDef(final char terminalValue) {
		statementDefType = EnumStatementDefType.TERMINAL_CHAR;
		terminalCharValue = terminalValue;

		nonTerminalName = null;
		terminalString = null;
		setOfTerminal = null;
		firstTerminalValue = 0;
		lastTerminalValue = 0;
		terminalSet = null;
	}

	public StatementDef(final StatementDef list, final StatementDef elt) {
		statementDefType = EnumStatementDefType.TERMINAL_SET_OF_TERMINAL;
		final EnumStatementDefType eltType = elt.getStatementDefType();
		switch (eltType) {
		case TERMINAL_CHAR_RANGE:
		case TERMINAL_CHAR_SET:
		case TERMINAL_CHAR:
			break;
		default:
			throw new IllegalArgumentException("bad type " + eltType
					+ " for element");
		}
		// FIXME remove assertion ?
		if (!elt.isTerminal()) {
			throw new IllegalArgumentException(elt + " is not a terminal");
		}
		if (!list.isTerminal()) {
			throw new IllegalArgumentException(list + " is not a terminal");
		}
		// end of assertion

		setOfTerminal = new LinkedList<StatementDef>();
		final EnumStatementDefType listType = list.getStatementDefType();
		switch (listType) {
		case TERMINAL_SET_OF_TERMINAL:
			final List<StatementDef> listSetOfTerminal = list
					.getSetOfTerminal();
			setOfTerminal.addAll(listSetOfTerminal);
			for (StatementDef son : listSetOfTerminal) {
				son.fatherStatementDef = this;
			}
			setOfTerminal.add(elt);
			elt.fatherStatementDef = this;
			break;
		case TERMINAL_CHAR:
		case TERMINAL_CHAR_RANGE:
		case TERMINAL_CHAR_SET:
			setOfTerminal.add(list);
			list.fatherStatementDef = this;
			setOfTerminal.add(elt);
			elt.fatherStatementDef = this;
			break;
		default:
			throw new IllegalArgumentException("bad type " + listType
					+ " for list");
		}

		nonTerminalName = null;
		terminalString = null;
		terminalCharValue = 0;
		firstTerminalValue = 0;
		lastTerminalValue = 0;
		terminalSet = null;
	}

	public StatementDef(final char first, final char last) {
		statementDefType = EnumStatementDefType.TERMINAL_CHAR_RANGE;
		firstTerminalValue = first;
		lastTerminalValue = last;

		nonTerminalName = null;
		terminalString = null;
		terminalCharValue = 0;
		setOfTerminal = null;
		terminalSet = null;
	}

	public StatementDef(final StatementDef list, final char terminalValue) {
		statementDefType = EnumStatementDefType.TERMINAL_CHAR_SET;

		terminalSet = new TreeSet<Integer>();

		final EnumStatementDefType listType = list.getStatementDefType();
		switch (listType) {
		case TERMINAL_CHAR:
			terminalSet.add(Integer.valueOf(list.getTerminalCharValue()));
			break;
		case TERMINAL_CHAR_SET:
			terminalSet.addAll(list.getTerminalSet());
			break;
		default:
			throw new IllegalArgumentException("bad type " + listType
					+ " for list");
		}
		terminalSet.add(Integer.valueOf(terminalValue));
		list.fatherStatementDef = this;
		nonTerminalName = null;
		terminalString = null;
		terminalCharValue = 0;
		setOfTerminal = null;
		firstTerminalValue = 0;
		lastTerminalValue = 0;
	}

	public StatementDef(final StatementDef statementDef,
			final String nonTerminalDefinedName) {
		this.statementDefType = statementDef.statementDefType;
		this.nonTerminalDefinedName = nonTerminalDefinedName;
		this.terminalSet = statementDef.terminalSet;
		this.nonTerminalName = statementDef.nonTerminalName;
		this.terminalString = statementDef.terminalString;
		this.terminalCharValue = statementDef.terminalCharValue;
		this.operandList = statementDef.operandList;
		this.setOfTerminal = statementDef.setOfTerminal;
		this.firstTerminalValue = statementDef.firstTerminalValue;
		this.lastTerminalValue = statementDef.lastTerminalValue;
		this.whiteSpaceNonTerminalName = statementDef.whiteSpaceNonTerminalName;
		this.tokenSon = statementDef.tokenSon;
		this.fatherStatementDef = statementDef.fatherStatementDef;
		if (operandList != null) {
			for (StatementDef son : operandList) {
				son.fatherStatementDef = this;
			}
		}
		if (setOfTerminal != null) {
			for (StatementDef son : setOfTerminal) {
				son.fatherStatementDef = this;
			}
		}
	}

	/**
	 * change this statement definition to concat statement definition
	 * 
	 */
	public void changeToConcat() {
		if (fatherStatementDef != null && fatherStatementDef.isTerminal()) {
			throw new IllegalStateException("father is terminal");
		}
		statementDefType = EnumStatementDefType.CONCAT;
		operandList = new LinkedList<StatementDef>();
	}

	/**
	 * change this statement definition to concat statement definition
	 * 
	 * @param sonStatementDef
	 *            son of this new concat statement definition
	 */
	public void changeToConcat(final StatementDef sonStatementDef) {
		if (fatherStatementDef != null && fatherStatementDef.isTerminal()) {
			throw new IllegalStateException("father is terminal");
		}
		statementDefType = EnumStatementDefType.CONCAT;
		operandList = new LinkedList<StatementDef>();
		operandList.add(sonStatementDef);
		sonStatementDef.fatherStatementDef = this;
	}

	public EnumStatementDefType getStatementDefType() {
		return statementDefType;
	}

	public void setMatchClassName(final String matchClassName) {
		if (this.matchClassName == null && matchClassName != null) {
			this.matchClassName = matchClassName;
			matchClassNameSetterTrace = new Exception("previous caller trace");
		} else if (matchClassName == null /* && this.matchClassName!=null */) {
			if (this.matchClassName != null) {
				redefineException(matchClassName);
			}
		} else if (!matchClassName.equals(this.matchClassName)) {
			redefineException(matchClassName);
		}
	}

	private void redefineException(final String matchClassName) {
		throw new IllegalStateException("for this statement definition "
				+ statementDefType + ", match class name already setted to "
				+ this.matchClassName + " can not set to " + matchClassName,
				matchClassNameSetterTrace);
	}

	// public void changeMatchClassName(final String matchClassName) {
	// this.matchClassName = matchClassName;
	// }

	public String getMatchClassName() {
		return matchClassName;
	}

	public void setNonTerminalDefinedName(final String nonTerminalDefinedName) {
		if (nonTerminalDefinedName == null) {
			throw new IllegalArgumentException(
					"non terminal defined name must have a value");
		}
		if (this.nonTerminalDefinedName == null) {
			this.nonTerminalDefinedName = nonTerminalDefinedName;
			int index = 0;
			switch (statementDefType) {
			case ALTERNATIVE:
			case CONCAT:
				for (StatementDef statementDef : operandList) {
					statementDef
							.setNonTerminalDefinedName(nonTerminalDefinedName
									+ '_' + index);
					index++;
				}
				break;
			case TERMINAL_SET_OF_TERMINAL:
				for (StatementDef statementDef : setOfTerminal) {
					statementDef
							.setNonTerminalDefinedName(nonTerminalDefinedName
									+ '_' + index);
					index++;
				}
				break;
			default:
				// no sons for witch set non terminal defined name needed
			}
		} else {
			if (!this.nonTerminalDefinedName.equals(nonTerminalDefinedName)) {
				throw new IllegalStateException(
						"this non terminal defined already set to "
								+ this.nonTerminalDefinedName
								+ ", can not be set to "
								+ nonTerminalDefinedName);
			}
		}
	}

	public String getNonTerminalDefinedName() {
		return nonTerminalDefinedName;
	}

	public boolean isToken() {
		return token;
	}

	/**
	 * @param tokenSon
	 *            the tokenSon to set
	 * @throws GeneratorException
	 */
	public void setTokenSon(final boolean tokenSon) throws GeneratorException {
		if (this.tokenSon != null && this.tokenSon != tokenSon) {
			throw new GeneratorException("token son state can not change to "
					+ this.tokenSon + " to " + tokenSon + " for "
					+ nonTerminalDefinedName);
		}
		this.tokenSon = tokenSon;
	}

	public boolean isTokenSon() throws GeneratorException {
		if (tokenSon == null) {
			throw new GeneratorException("token son state must be set for "
					+ nonTerminalDefinedName);
		}
		return tokenSon;
	}

	public boolean isTerminal() {
		return statementDefType.isTerminal();
	}

	public boolean isNonTerminalDefType() {
		return statementDefType.isNonTerminalDef();
	}

	public boolean isEmptyDefType() {
		return statementDefType.isEmptyDef();
	}

	public void setIsToken() {
		token = true;
	}

	public String getWhiteSpaceNonTerminalName() {
		return whiteSpaceNonTerminalName;
	}

	public void setWhiteSpaceNonTerminalName(
			final String whiteSpaceNonTerminalName) {
		this.whiteSpaceNonTerminalName = whiteSpaceNonTerminalName;
	}

	public String getNonTerminalName() {
		if (!EnumStatementDefType.NON_TERMINAL.equals(statementDefType)) {
			throw new UnsupportedOperationException("this is not a terminal");
		}
		return nonTerminalName;
	}

	public String getTerminalString() {
		if (!EnumStatementDefType.TERMINAL_STRING.equals(statementDefType)) {
			throw new UnsupportedOperationException(
					"this is not a terminal string");
		}
		return terminalString;
	}

	public char getTerminalCharValue() {
		if (!EnumStatementDefType.TERMINAL_CHAR.equals(statementDefType)) {
			throw new UnsupportedOperationException(
					"this is not a terminal string");
		}
		return terminalCharValue;
	}

	public List<StatementDef> getSetOfTerminal() {
		if (!EnumStatementDefType.TERMINAL_SET_OF_TERMINAL
				.equals(statementDefType)) {
			throw new UnsupportedOperationException(
					"this is not a terminal set of terminal");
		}
		return setOfTerminal;
	}

	public char getFirstTerminalValue() {
		if (!EnumStatementDefType.TERMINAL_CHAR_RANGE.equals(statementDefType)) {
			throw new UnsupportedOperationException(
					"this is not a terminal range");
		}
		return firstTerminalValue;
	}

	public char getLastTerminalValue() {
		if (!EnumStatementDefType.TERMINAL_CHAR_RANGE.equals(statementDefType)) {
			throw new UnsupportedOperationException(
					"this is not a terminal range");
		}
		return lastTerminalValue;
	}

	public Set<Integer> getTerminalSet() {
		if (!EnumStatementDefType.TERMINAL_CHAR_SET.equals(statementDefType)) {
			throw new UnsupportedOperationException(
					"this is not a terminal set");
		}
		return terminalSet;
	}

	public List<StatementDef> getOperandList() {
		if (!EnumStatementDefType.CONCAT.equals(statementDefType)
				&& !EnumStatementDefType.ALTERNATIVE.equals(statementDefType)) {
			throw new UnsupportedOperationException(
					"this is not a concat or alternative");
		}
		return operandList;
	}

	public StatementDef getFatherStatementDef() {
		return fatherStatementDef;
	}

	public void accept(final IGrammarDefVisitor visitor)
			throws GrammarDefNodeVisitException {
		if (visitor.beginVisit(this)) {
			switch (statementDefType) {
			case ALTERNATIVE:
			case CONCAT:
				for (StatementDef statementDef : operandList) {
					statementDef.accept(visitor);
				}
				break;
			case TERMINAL_SET_OF_TERMINAL:
				for (StatementDef statementDef : setOfTerminal) {
					statementDef.accept(visitor);
				}
				break;
			default:
				// no sons to visit
			}
			visitor.endVisit(this);
		}
	}

	@Override
	public String toString() {
		final StringBuilder stringBuilder = new StringBuilder();
		stringBuilder.append("StatementDef ");
		stringBuilder.append(statementDefType);
		stringBuilder.append(", match mgr: ");
		stringBuilder.append(matchClassName);
		stringBuilder.append(", nt defined: ");
		stringBuilder.append(nonTerminalDefinedName);
		switch (statementDefType) {
		// case ALTERNTIVE:
		// case CONCAT:
		// case EMPTY:

		case NON_TERMINAL:
			stringBuilder.append(", non terminal identifier: ");
			stringBuilder.append(nonTerminalName);
			break;

		case TERMINAL_CHAR:
			stringBuilder.append(", terminal character: ");
			if (terminalCharValue >= ' ' && terminalCharValue < 0x7f) {
				stringBuilder.append('\'');
				stringBuilder.append(terminalCharValue);
				stringBuilder.append('\'');
			} else {
				stringBuilder.append("0X");
				stringBuilder.append(Integer.toHexString(terminalCharValue));
			}
			break;

		case TERMINAL_CHAR_RANGE:
			stringBuilder.append(", terminal character range: ");
			if (firstTerminalValue >= ' ' && firstTerminalValue < 0x7f) {
				stringBuilder.append('\'');
				stringBuilder.append(firstTerminalValue);
				stringBuilder.append('\'');
			} else {
				stringBuilder.append("0X");
				stringBuilder.append(Integer.toHexString(firstTerminalValue));
			}
			stringBuilder.append(" - ");
			if (lastTerminalValue >= ' ' && lastTerminalValue < 0x7f) {
				stringBuilder.append('\'');
				stringBuilder.append(lastTerminalValue);
				stringBuilder.append('\'');
			} else {
				stringBuilder.append("0X");
				stringBuilder.append(Integer.toHexString(lastTerminalValue));
			}
			break;

		case TERMINAL_CHAR_SET:
			stringBuilder.append(", terminal character set:");
			for (int characterElt : terminalSet) {
				if (characterElt >= ' ' && characterElt < 0x7f) {
					stringBuilder.append(" '");
					stringBuilder.append((char) characterElt);
					stringBuilder.append('\'');
				} else {
					stringBuilder.append(" 0X");
					stringBuilder.append(Integer.toHexString(characterElt));
				}
			}
			break;

		// case TERMINAL_SET_OF_TERMINAL:
		// FIXME something to log ?
		// break

		case TERMINAL_STRING:
			stringBuilder.append(", terminal string: \"");
			stringBuilder.append(terminalString);
			stringBuilder.append('"');
			break;

		default:
		}
		return stringBuilder.toString();
	}
}
