/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.generator.service;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;

import net.sf.parser4j.generator.entity.grammardefnode.GrammarDef;
import net.sf.parser4j.generator.entity.grammardefnode.IGrammarDefNode;
import net.sf.parser4j.generator.entity.grammardefnode.RuleDef;
import net.sf.parser4j.generator.entity.grammardefnode.RulesDef;
import net.sf.parser4j.generator.entity.grammardefnode.RulesGroupListDef;
import net.sf.parser4j.generator.entity.grammardefnode.StatementDef;
import net.sf.parser4j.generator.service.grammardefnode.IGrammarDefVisitor;

/**
 * output grammar definition tree to text file
 * 
 * @author luc peuvrier
 * 
 */
public final class GrammarDefNodeToTextFile implements IGrammarDefVisitor {

	private static final GrammarDefNodeToTextFile INSTANCE =
	/**/new GrammarDefNodeToTextFile();

	private transient BufferedWriter bufferedWriter;

	private transient int numberOfPad;

	public static GrammarDefNodeToTextFile getInstance() {
		return INSTANCE;
	}

	private GrammarDefNodeToTextFile() {
		super();
	}

	public void generateTextFile(final IGrammarDefNode grammarDefNode,
			final File textFile) throws GrammarDefNodeVisitException {
		FileOutputStream fileOutputStream;
		try {
			fileOutputStream = new FileOutputStream(textFile);
		} catch (FileNotFoundException exception) {
			throw new GrammarDefNodeVisitException(exception);
		}
		Writer writer;
		try {
			writer = new OutputStreamWriter(fileOutputStream, "UTF8");
		} catch (UnsupportedEncodingException exception) {
			try {
				fileOutputStream.close();
			} catch (IOException exception2) {// NOPMD
				// ignore error
			}
			throw new GrammarDefNodeVisitException(exception);
		}
		generate(grammarDefNode, writer);
	}

	public String generateToString(final IGrammarDefNode grammarDefNode)
			throws GrammarDefNodeVisitException {
		final Writer writer = new StringWriter();
		generate(grammarDefNode, writer);
		return writer.toString();
	}

	private void generate(final IGrammarDefNode grammarDefNode,
			final Writer writer) throws GrammarDefNodeVisitException {
		bufferedWriter = new BufferedWriter(writer);
		numberOfPad = 0;
		grammarDefNode.accept(this);
		try {
			bufferedWriter.close();
		} catch (IOException exception) {
			throw new GrammarDefNodeVisitException(exception);
		}
	}

	private void cbeginVisit(final IGrammarDefNode def)
			throws GrammarDefNodeVisitException {
		try {
			addPad();
			bufferedWriter.write("begin of ");
			bufferedWriter.write(def.toString());
			bufferedWriter.write('\n');
			if (def.getClass().equals(StatementDef.class)) {
				addPad();
				final StatementDef fatherStatementDef = ((StatementDef) def)
						.getFatherStatementDef();
				if (fatherStatementDef == null) {
					bufferedWriter.write("root");
				} else {
					bufferedWriter.write("soon of ");
					bufferedWriter.write(fatherStatementDef
							.getNonTerminalDefinedName());
				}
				bufferedWriter.write('\n');
			}
			numberOfPad++;
		} catch (IOException exception) {
			throw new GrammarDefNodeVisitException(exception);
		}
	}

	private void cendVisit(final IGrammarDefNode def)
			throws GrammarDefNodeVisitException {
		try {
			numberOfPad--;
			addPad();
			bufferedWriter.write("end of ");
			if (def.getClass().equals(StatementDef.class)) {
				bufferedWriter.write(((StatementDef) def).getStatementDefType()
						.toString());
			} else {
				bufferedWriter.write(def.toString());
			}
			bufferedWriter.write('\n');
		} catch (IOException exception) {
			throw new GrammarDefNodeVisitException(exception);
		}
	}

	private void addPad() throws IOException {
		for (int count = 0; count < numberOfPad; count++) {
			bufferedWriter.write(' ');
		}
	}

	@Override
	public void beginVisit(final GrammarDef def)
			throws GrammarDefNodeVisitException {
		cbeginVisit(def);
	}

	@Override
	public void endVisit(final GrammarDef def)
			throws GrammarDefNodeVisitException {
		cendVisit(def);
	}

	@Override
	public void beginVisit(final RulesGroupListDef def)
			throws GrammarDefNodeVisitException {
		cbeginVisit(def);
	}

	@Override
	public void endVisit(final RulesGroupListDef def)
			throws GrammarDefNodeVisitException {
		cendVisit(def);
	}

	@Override
	public void beginVisit(final RulesDef def)
			throws GrammarDefNodeVisitException {
		cbeginVisit(def);
	}

	@Override
	public void endVisit(final RulesDef def)
			throws GrammarDefNodeVisitException {
		cendVisit(def);
	}

	@Override
	public void beginVisit(final RuleDef def)
			throws GrammarDefNodeVisitException {
		cbeginVisit(def);
	}

	@Override
	public void endVisit(final RuleDef def) throws GrammarDefNodeVisitException {
		cendVisit(def);
	}

	@Override
	public boolean beginVisit(final StatementDef def)
			throws GrammarDefNodeVisitException {
		cbeginVisit(def);
		return true;
	}

	@Override
	public void endVisit(final StatementDef def)
			throws GrammarDefNodeVisitException {
		cendVisit(def);
	}
}
