/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.generator.service;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.apache.log4j.Logger;

import net.sf.parser4j.generator.entity.grammardefnode.EnumStatementDefType;
import net.sf.parser4j.generator.entity.grammardefnode.IGrammarDefNode;
import net.sf.parser4j.generator.entity.grammardefnode.StatementDef;
import net.sf.parser4j.kernelgenerator.entity.GrammarTreeFactoryData;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.AbstractMularyGrammarNode;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.GrammarNodeNull;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.IGrammarNode;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeString;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeTerminalSetOfTerminal;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.OperandGrammarNode;
import net.sf.parser4j.kernelgenerator.service.GeneratorException;
import net.sf.parser4j.parser.entity.EnumNodeType;
import net.sf.parser4j.parser.entity.data.NonTerminalMap;
import net.sf.parser4j.parser.entity.data.Pair;

/**
 * create grammar tree from root data node
 * 
 * @author luc peuvrier
 * 
 */
public final class GrammarTreeFactory {

	private static final Logger _log = Logger
			.getLogger(GrammarTreeFactory.class);

	private static final String START_NON_TERMINAL = "Start";

	private static final String EMPTY_STRING = "";

	private static final GrammarTreeFactory INSTANCE =
	/**/new GrammarTreeFactory();

	private transient NonTerminalMap nonTerminalMap;

	private transient Map<String, Pair<StatementDef, IGrammarNode>>
	/**/statementDefAnGrammarNodeMap;

	public static GrammarTreeFactory getInstance() {
		return INSTANCE;
	}

	private GrammarTreeFactory() {
		super();
	}

	/**
	 * create grammar tree
	 * 
	 * @param rootData
	 *            root data node of grammar definition text file parsing
	 * @return
	 * @throws GeneratorException
	 */
	public GrammarTreeFactoryData createGrammar(final IGrammarDefNode rootData)
			throws GeneratorException {
		nonTerminalMap = new NonTerminalMap();
		statementDefTokenSonUpdate(rootData);
		computeStatementDefAnGrammarNodeMap(rootData);
		linkGrammarNode();
		final Pair<StatementDef, IGrammarNode> pair = statementDefAnGrammarNodeMap
				.get(START_NON_TERMINAL);
		final StatementDef statementDef = pair.getMemberA();
		if (statementDef.isToken()) {
			throw new GeneratorException(
					"Start non terminal can not be a token");
		}
		final IGrammarNode rootGrammarNode = pair.getMemberB();
		if (rootGrammarNode == null) {
			throw new GeneratorException("Start non terminal is not defined");
		}
		return new GrammarTreeFactoryData(nonTerminalMap, rootGrammarNode);
	}

	private void statementDefTokenSonUpdate(final IGrammarDefNode rootData)
			throws GeneratorException {
		final StatementDefTokenSonUpdater updater =
		/**/new StatementDefTokenSonUpdater();
		updater.update(rootData);
		final Set<String> notReferenced = updater.getNotReferenced();
		if (!notReferenced.remove(START_NON_TERMINAL)) {
			throw new GeneratorException(
					"Start must not be referenced by other rules");
		}
		if (!notReferenced.isEmpty()) {
			final StringBuilder builder = new StringBuilder();
			builder.append("not referenced non terminal:\n");
			for (String nonTerminalName : notReferenced) {
				builder.append(nonTerminalName);
				builder.append('\n');
			}
			// FIXME return this list in result
			_log.warn(builder.toString());
		}
	}

	private void computeStatementDefAnGrammarNodeMap(
			final IGrammarDefNode rootData) throws GeneratorException {
		final StatementDefToGrammarNodeAssocier associer =
		/**/new StatementDefToGrammarNodeAssocier(nonTerminalMap);
		try {
			/* do the visit */
			rootData.accept(associer);
		} catch (GrammarDefNodeVisitException exception) {
			throw new GeneratorException(exception);
		}
		statementDefAnGrammarNodeMap = associer
				.getStatementDefAnGrammarNodeMap();
		if (_log.isDebugEnabled()) {
			try {
				_log.debug("transformed grammar def\n"
						+ GrammarDefNodeToTextFile.getInstance()
								.generateToString(rootData));
			} catch (GrammarDefNodeVisitException exception) {
				_log.error("visiting transformed grammar def error", exception);
			}
		}
	}

	private void linkGrammarNode() throws GeneratorException {
		for (Pair<StatementDef, IGrammarNode> pair : statementDefAnGrammarNodeMap
				.values()) {
			final StatementDef statementDef = pair.getMemberA();
			final IGrammarNode grammarNode = pair.getMemberB();
			final EnumStatementDefType statementDefType = statementDef
					.getStatementDefType();
			final EnumNodeType grammarNodeType = grammarNode
					.getGrammarNodeType();
			switch (grammarNodeType) {

			case ALTERNATIVE:
				if (EnumStatementDefType.ALTERNATIVE.equals(statementDefType)) {
					linkGrammarNode(grammarNodeType, statementDef,
							(AbstractMularyGrammarNode) grammarNode);
				}
				break;

			case CONCAT:
				if (EnumStatementDefType.CONCAT.equals(statementDefType)) {
					linkGrammarNode(grammarNodeType, statementDef,
							(AbstractMularyGrammarNode) grammarNode);
				} else if (EnumStatementDefType.NON_TERMINAL
						.equals(statementDefType)) {
					final String name = statementDef.getNonTerminalName();
					final List<OperandGrammarNode> operandList =
					/**/new LinkedList<OperandGrammarNode>();// NOPMD
					final IGrammarNode nonTerminalGrammarNode =
					/**/statementDefAnGrammarNodeMap.get(name).getMemberB();
					final IGrammarNode whiteSpaceGrammarNode =
					/**/getWhiteSpaceGrammarNode(statementDef);
					if (whiteSpaceGrammarNode != null
							&& (nonTerminalGrammarNode.isTerminal() || nonTerminalGrammarNode
									.isToken())) {
						addToOperandList(operandList, whiteSpaceGrammarNode,
								true);
					}
					addToOperandList(operandList, nonTerminalGrammarNode, false);
					/* add white space at end of concat start rules */
					if (whiteSpaceGrammarNode != null
							&& START_NON_TERMINAL.equals(grammarNode
									.getNonTerminalName())) {
						addToOperandList(operandList, whiteSpaceGrammarNode,
								true);
					}
					grammarNode.setOperandList(operandList);
				} else if (EnumStatementDefType.EMPTY.equals(statementDefType)) {
					final IGrammarNode whiteSpaceGrammarNode =
					/**/getWhiteSpaceGrammarNode(statementDef);
					if (whiteSpaceGrammarNode != null) {
						final List<OperandGrammarNode> operandList =
						/**/new LinkedList<OperandGrammarNode>();// NOPMD
						addToOperandList(operandList, whiteSpaceGrammarNode,
								true);
						grammarNode.setOperandList(operandList);
					}
				}
				break;

			case EMPTY:
				/* have not son(s) */
				break;

			case CHARACTER_TERMINAL:
				if (EnumStatementDefType.TERMINAL_SET_OF_TERMINAL
						.equals(statementDefType)) {
					final List<StatementDef> list = statementDef
							.getSetOfTerminal();

					// FIXME assertion to remove
					if (list.isEmpty()) {
						throw new GeneratorException(
								"statement def have not son statement def");
					}
					for (StatementDef sonStatementDef : list) {
						if (!sonStatementDef.isTerminal()) {
							throw new GeneratorException(
									"son of statement def for non terminal "
											+ statementDef
													.getNonTerminalDefinedName()
											+ " is not a terminal, but is "
											+ sonStatementDef);
						}
					}
					// end assertion to remove

					final Set<IGrammarNode> setOfTerminalNode =
					/**/new TreeSet<IGrammarNode>();// NOPMD
					for (StatementDef subs : list) {
						final String name = subs.getNonTerminalDefinedName();
						setOfTerminalNode.add(statementDefAnGrammarNodeMap.get(
								name).getMemberB());
					}
					((GrammarNodeTerminalSetOfTerminal) grammarNode)
							.setSetOfTerminalNode(setOfTerminalNode);

				}
				break;

			case STRING_TERMINAL:
				final String stringValue = statementDef.getTerminalString();
				((GrammarNodeString) grammarNode).setStringValue(stringValue,
						nonTerminalMap);
				break;

			case NULL:
				/* have not son(s), identifier statement special case */
				break;

			default:
				throw new GeneratorException("bad type " + grammarNodeType);
			}
		}
	}

	private void linkGrammarNode(final EnumNodeType grammarNodeType,
			final StatementDef statementDef, final IGrammarNode grammarNode)
			throws GeneratorException {

		/* true if concat node, else must be alternative node */
		final boolean concat = EnumNodeType.CONCAT.equals(grammarNodeType);

		final IGrammarNode whiteSpaceGrammarNode =
		/**/getWhiteSpaceGrammarNode(statementDef);

		final List<OperandGrammarNode> operandList =
		/**/new LinkedList<OperandGrammarNode>();
		final List<StatementDef> list = statementDef.getOperandList();
		final boolean tokenSon = statementDef.isTokenSon();
		for (StatementDef subStatementDef : list) {
			String name = subStatementDef.getNonTerminalDefinedName();
			final Pair<StatementDef, IGrammarNode> pair = statementDefAnGrammarNodeMap
					.get(name);
			if (pair == null) {
				throw new GeneratorException("no definition for " + name);
			}
			IGrammarNode subGrammarNode = pair.getMemberB();
			/*
			 * non terminal special case
			 */
			while (EnumNodeType.NULL
					.equals(subGrammarNode.getGrammarNodeType())) {
				if (EnumStatementDefType.NON_TERMINAL.equals(subStatementDef
						.getStatementDefType())) {
					/*
					 * asserted statement type is a non terminal
					 */

					/*
					 * get the grammar node for the non terminal
					 */
					name = subStatementDef.getNonTerminalName();
					final Pair<StatementDef, IGrammarNode> pair2 =
					/**/statementDefAnGrammarNodeMap.get(name);
					if (pair2 == null) {
						throw new GeneratorException("no definition for "
								+ name);
					}
					subGrammarNode = pair2.getMemberB();
					subStatementDef = pair2.getMemberA();
				} else {
					throw new GeneratorException("bad type "
							+ subStatementDef.getStatementDefType());
				}
			}

			if (concat
					&& !tokenSon
					&& whiteSpaceGrammarNode != null
					&& (subGrammarNode.isTerminal() || subGrammarNode.isToken())) {
				/* additional white space son */
				addToOperandList(operandList, whiteSpaceGrammarNode, true);
			}
			/* add son */
			addToOperandList(operandList, subGrammarNode, false);
		}
		/* add white space at end of concat start rules */
		if (concat
				&& whiteSpaceGrammarNode != null
				&& (START_NON_TERMINAL.equals(grammarNode.getNonTerminalName()) || list
						.isEmpty())) {
			/* additional white space son */
			addToOperandList(operandList, whiteSpaceGrammarNode, true);
		}
		grammarNode.setOperandList(operandList);
	}

	/**
	 * 
	 * @param operandList
	 *            the operand list to update
	 * @param grammarNode
	 *            grammar node of the operand
	 * @param whiteSpace
	 *            true if operand is a white space
	 */
	private void addToOperandList(final List<OperandGrammarNode> operandList,
			final IGrammarNode grammarNode, final boolean whiteSpace) {
		if (EnumNodeType.NULL.equals(grammarNode.getGrammarNodeType())) {
			throw new IllegalArgumentException(
					"can not add null grammar node to operand list");
		}
		final OperandGrammarNode operandGrammarNode = new OperandGrammarNode();
		operandGrammarNode.setGrammarNode(grammarNode);
		operandGrammarNode.setWhiteSpace(whiteSpace);
		operandList.add(operandGrammarNode);
	}

	private IGrammarNode getWhiteSpaceGrammarNode(final StatementDef def) {
		String whiteSpaceNonTerminalName = def.getWhiteSpaceNonTerminalName();
		if (whiteSpaceNonTerminalName == null) {
			throw new IllegalStateException(
					"white space non terminal value must be set for " + def);
		}
		IGrammarNode whiteSpaceGrammarNode;
		if (EMPTY_STRING.equals(whiteSpaceNonTerminalName)) {
			whiteSpaceGrammarNode = null;// NOPMD
		} else {
			whiteSpaceGrammarNode = statementDefAnGrammarNodeMap.get(
					whiteSpaceNonTerminalName).getMemberB();
			while (EnumNodeType.NULL.equals(whiteSpaceGrammarNode
					.getGrammarNodeType())) {
				whiteSpaceNonTerminalName = ((GrammarNodeNull) whiteSpaceGrammarNode)
						.getSonNonTerminalName();
				whiteSpaceGrammarNode = statementDefAnGrammarNodeMap.get(
						whiteSpaceNonTerminalName).getMemberB();
			}
		}
		return whiteSpaceGrammarNode;
	}
}
