/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.generator.service;//NOPMD high coupling

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.util.List;
import java.util.Map;

import net.sf.parser4j.generator.entity.grammardefnode.IGrammarDefNode;
import net.sf.parser4j.generator.entity.parserdata.ParserDataForGeneratorInput;
import net.sf.parser4j.kernelgenerator.entity.GrammarTreeFactoryData;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.AbstractMularyGrammarNode;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.IGrammarNode;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.OperandGrammarNode;
import net.sf.parser4j.kernelgenerator.service.GeneratorException;
import net.sf.parser4j.kernelgenerator.service.GrammarNodeVisitException;
import net.sf.parser4j.kernelgenerator.service.KernelParserDataGenerator;
import net.sf.parser4j.kernelgenerator.service.grammarnode.AbstractGrammarNodeVisitor;
import net.sf.parser4j.parser.entity.ParseResult;
import net.sf.parser4j.parser.entity.data.NonTerminal;
import net.sf.parser4j.parser.entity.data.NonTerminalMap;
import net.sf.parser4j.parser.entity.data.Pair;
import net.sf.parser4j.parser.entity.data.ParserData;
import net.sf.parser4j.parser.entity.parsenode.IParseNode;
import net.sf.parser4j.parser.service.HasAmbiguityParserException;
import net.sf.parser4j.parser.service.IParserListener;
import net.sf.parser4j.parser.service.Parser;
import net.sf.parser4j.parser.service.ParserException;
import net.sf.parser4j.parser.service.ParserFileReader;
import net.sf.parser4j.parser.service.data.ParserDataIO;
import net.sf.parser4j.parser.service.parsenode.ParseTreeToXML;

import org.apache.log4j.Logger;

/**
 * generate parser data file from grammar definition file
 * 
 * @author luc peuvrier
 * 
 */
public final class ParserDataGenerator extends AbstractGrammarNodeVisitor {

	private static final Logger _log = Logger
			.getLogger(ParserDataGenerator.class);

	private static final ParserDataGenerator INSTANCE =
	/**/new ParserDataGenerator();

	private static final ParserDataForGeneratorInput parserDataForGeneratorInput =
	/**/ParserDataForGeneratorInput.getInstance();

	private static final GrammarTreeFactory grammarFactory =
	/**/GrammarTreeFactory.getInstance();

	private static final KernelParserDataGenerator kernelParserDataGenerator =
	/**/KernelParserDataGenerator.getInstance();

	private static final ParserDataIO parserDataIO =
	/**/ParserDataIO.getInstance();

	private static final ParserDataToHtml parserDataToHtml =
	/**/ParserDataToHtml.getInstance();

	private static final ParseTreeToXML parseNodeToXML = ParseTreeToXML
			.getInstance();

	private static final GrammarDefNodeToTextFile grammarDefNodeToTextFile =
	/**/GrammarDefNodeToTextFile.getInstance();

	private static final GrammarToTextFile grammarToTextFile =
	/**/GrammarToTextFile.getInstance();

	public static ParserDataGenerator getInstance() {
		return INSTANCE;
	}

	private transient String whiteSpaceNonTerminalName;

	private ParserDataGenerator() {
		super();
	}

	/**
	 * generate parser data file from grammar definition file
	 * 
	 * @param grammarInputFile
	 *            grammar definition UTF-8 text file
	 * @param htmlOutputFile
	 *            parser data in html form file, null if no output in html form
	 * @param parserDataOutputFile
	 *            parser data file to generate
	 * @param matchMgrPackageName
	 *            name of package containing the match manager classes
	 * @param grammarDefTreeTextFile
	 *            to output grammar definition tree in text file, null if no
	 *            output
	 * @param grammarTreeTextFile
	 *            to output grammar tree in text file, null if no output
	 * @return parsing result
	 * @throws GeneratorException
	 * 
	 */
	public ParseResult generate(final File grammarInputFile,
			final File htmlOutputFile, final File parserDataOutputFile,
			final String matchMgrPackageName,
			final File grammarDefTreeTextFile, final File grammarTreeTextFile,
			final IParserListener listener) throws GeneratorException {

		/* read parser data for generator input */
		final ParserData parserData4Gen = parserDataForGeneratorInput.read();
		return generate(parserData4Gen, grammarInputFile, htmlOutputFile,
				parserDataOutputFile, matchMgrPackageName,
				grammarDefTreeTextFile, grammarTreeTextFile, listener);
	}

	public ParseResult generate(final ParserData parserData4Gen,
			final File grammarInputFile, final File htmlOutputFile,
			final File parserDataOutputFile, final String matchMgrPackageName,
			final File grammarDefTreeTextFile, final File grammarTreeTextFile,
			final IParserListener listener) throws GeneratorException {
		/* generate with parser data for generator input */
		final Pair<ParseResult, ParserData> pair = generate(parserData4Gen,
				grammarInputFile, htmlOutputFile, matchMgrPackageName,
				grammarDefTreeTextFile, grammarTreeTextFile, listener);
		final ParserData parserData = pair.getMemberB();
		if (parserData != null && parserDataOutputFile != null) {
			outputParserData(parserData, parserDataOutputFile, htmlOutputFile);
		}
		return pair.getMemberA();
	}

	/**
	 * write parser data for parser usage and optionally in html form
	 * 
	 * @param parserData
	 *            parser data to output
	 * @param parserDataOutputFile
	 *            parser data file to generate
	 * @param htmlOutputFile
	 *            parser data in html form file, null if no output in html form
	 * @throws GeneratorException
	 */
	private void outputParserData(final ParserData parserData,
			final File parserDataOutputFile, final File htmlOutputFile)
			throws GeneratorException {
		try {
			parserDataIO.writeToFile(parserDataOutputFile, parserData);
		} catch (ParserException exception) {
			throw new GeneratorException(exception);
		}
		if (htmlOutputFile != null) {
			try {
				parserDataToHtml.generateHtml(parserData, htmlOutputFile);
			} catch (IOException exception) {
				throw new GeneratorException(
						"failed html output of parser data in "
								+ htmlOutputFile, exception);
			}
		}
	}

	/**
	 * 
	 * @param parserData4Gen
	 *            parser data to use to parse grammar definition
	 * @param grammarInputFile
	 *            grammar definition UTF-8 text file
	 * @param matchMgrPackageName
	 *            name of package containing the match manager classes
	 * @param grammarDefTreeTextFile
	 *            to output grammar definition tree in text file, null if no
	 *            output
	 * @param grammarTreeTextFile
	 *            to output grammar tree in text file, null if no output
	 * @param listener
	 * @return parsing result
	 * @throws GeneratorException
	 */
	public Pair<ParseResult, ParserData> generate(
			final ParserData parserData4Gen, final File grammarInputFile,
			final File htmlOutputFile, final String matchMgrPackageName,
			final File grammarDefTreeTextFile, final File grammarTreeTextFile,
			final IParserListener listener) throws GeneratorException {

		/* create reader of generator input */
		InputStream inputStream;
		try {
			inputStream = new FileInputStream(grammarInputFile);
		} catch (FileNotFoundException exception) {
			throw new GeneratorException("creating reader of "
					+ grammarInputFile, exception);
		}
		Reader reader;
		try {
			reader = new InputStreamReader(inputStream, "UTF-8");
		} catch (UnsupportedEncodingException exception) {
			throw new GeneratorException("creating reader of "
					+ grammarInputFile, exception);
		}
		final ParserFileReader parserFileReader =
		/**/new ParserFileReader(reader, grammarInputFile.getPath());
		final Pair<ParseResult, ParserData> generate = generate(parserData4Gen,
				parserFileReader, matchMgrPackageName, grammarDefTreeTextFile,
				grammarTreeTextFile, null, listener);
		return generate;
	}

	/**
	 * close reader after parsing
	 * 
	 * @param parserData4Gen
	 *            parser data to use to parse grammar definition
	 * @param reader
	 *            reader of grammar definition UTF-8 text file
	 * @param matchMgrPackageName
	 *            name of package containing the match manager classes
	 * @param grammarDefTreeTextFile
	 *            to output grammar definition tree in text file, null if no
	 *            output
	 * @param grammarTreeTextFile
	 *            to output grammar tree in text file, null if no output
	 * @param whiteSpaceNonTerminalName
	 *            name of global white space non terminal, null if none
	 * @param listener
	 * @return parsing result
	 * @throws GeneratorException
	 */
	public Pair<ParseResult, ParserData> generate(
			final ParserData parserData4Gen, final ParserFileReader reader,
			final String matchMgrPackageName,
			final File grammarDefTreeTextFile, final File grammarTreeTextFile,
			final String whiteSpaceNonTerminalName,
			final IParserListener listener) throws GeneratorException {
		/* create parser of generator input */
		Parser parser;
		try {
			parser = new Parser(parserData4Gen, listener);
		} catch (ParserException exception) {
			throw new GeneratorException(
					"creating parser for grammar definition", exception);
		}
		/* parse the generator input */
		final ParseResult parseResult;
		final ParserData parserData;
		try {
			/* parsing of the generator input */
			parseResult = parser
					.parse(reader, false/* syntaxAcceptAmbiguity */, false/* lexAcceptAmbiguity */);
			final IParseNode rootParseNode = parseResult.getResultParseNode();
			if (parseResult.isInError()) {
				parserData = null;
			} else {
				/*
				 * have a parsing result: compute the grammar tree from the
				 * parsing result
				 */
				final IGrammarDefNode rootData = (IGrammarDefNode) rootParseNode
						.getUniqData();
				if (grammarDefTreeTextFile != null) {
					try {
						grammarDefNodeToTextFile.generateTextFile(rootData,
								grammarDefTreeTextFile);
					} catch (GrammarDefNodeVisitException exception) {
						_log.error("generating " + grammarDefTreeTextFile,
								exception);
					}
				}
				final GrammarTreeFactoryData grammarTreeFactoryData =
				/**/grammarFactory.createGrammar(rootData);
				final NonTerminalMap nonTerminalMap = grammarTreeFactoryData
						.getNonTerminalMap();
				final IGrammarNode rootGrammarNode = grammarTreeFactoryData
						.getRootGrammarNode();
				if (whiteSpaceNonTerminalName != null) {
					this.whiteSpaceNonTerminalName = whiteSpaceNonTerminalName;
					try {
						initialize();
						rootGrammarNode.accept(this);
						close();
					} catch (GrammarNodeVisitException exception) {
						throw new GeneratorException(exception);
					}
				}

				if (grammarTreeTextFile != null) {
					try {
						grammarToTextFile.generateTextFile(rootGrammarNode,
								grammarTreeTextFile);
					} catch (GeneratorException exception) {// NOPMD
						_log.error("generating " + grammarTreeTextFile,
								exception);
					}
				}
				/* generate parser data */
				parserData = kernelParserDataGenerator.generate(
						rootGrammarNode, nonTerminalMap, matchMgrPackageName);
			}
		} catch (HasAmbiguityParserException exception) {
			/*
			 * output information on encountered ambiguity during parsing
			 */
			final Map<Integer, NonTerminal> nonTerminalByIdentifierMap =
			/**/parserData4Gen.getNonTerminalByIdentifierMap();
			try {
				new File("runtime").mkdirs();
				parseNodeToXML.toXml("runtime/parsetree", exception,
						nonTerminalByIdentifierMap);
				throw new GeneratorException(
						"have ambiguity: see 'runtime/parsetree <[0-9]+> .xml'",
						exception);
			} catch (Exception exception2) {
				_log.error("while output to xml for ambiguity exception",
						exception2);
				throw new GeneratorException(// NOPMD no problems, exception is
						// logged
						"have ambiguity: failed generate 'runtime/parsetree <[0-9]+> .xml', see log");
			}
		} catch (ParserException exception) {
			throw new GeneratorException(exception);
		}
		return new Pair<ParseResult, ParserData>(parseResult, parserData);
	}

	@Override
	protected void beginVisitImpl(final IGrammarNode grammarNode)
			throws GrammarNodeVisitException {
		visitImpl(grammarNode);
	}

	@Override
	protected void endVisitImpl(final IGrammarNode grammarNode)
			throws GrammarNodeVisitException {
		// no implementation
	}

	@Override
	protected void visitImpl(final IGrammarNode grammarNode)
			throws GrammarNodeVisitException {
		if (grammarNode instanceof AbstractMularyGrammarNode) {
			final AbstractMularyGrammarNode mulary = (AbstractMularyGrammarNode) grammarNode;
			final List<OperandGrammarNode> operandList = mulary
					.getOperandList();
			for (OperandGrammarNode operandGrammarNode : operandList) {
				final String nonTerminalName = operandGrammarNode
						.getNonTerminalName();
				if (whiteSpaceNonTerminalName.equals(nonTerminalName)) {
					operandGrammarNode.setWhiteSpace(true);
				}
			}
		}
	}
}