/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.generator.service;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.parser4j.parser.entity.data.FromNonTerminalEntry;
import net.sf.parser4j.parser.entity.data.FromTerminalEntry;
import net.sf.parser4j.parser.entity.data.GotoByNonTerminalEntry;
import net.sf.parser4j.parser.entity.data.GotoByTerminalEntry;
import net.sf.parser4j.parser.entity.data.MatchEntry;
import net.sf.parser4j.parser.entity.data.MatchEntrySet;
import net.sf.parser4j.parser.entity.data.NonTerminal;
import net.sf.parser4j.parser.entity.data.Pair;
import net.sf.parser4j.parser.entity.data.ParserData;
import net.sf.parser4j.parser.entity.data.ParserDataRecognitionPoint;
import net.sf.parser4j.parser.entity.data.StateRecognitionPointList;
import net.sf.parser4j.parser.entity.data.TerminalCharRange;
import net.sf.parser4j.parser.entity.data.ToStateAndWhiteSpaceFlag;

/**
 * 
 * @author luc peuvrier
 * 
 */
public final class ParserDataToHtml {

	private static final ParserDataToHtml INSTANCE = new ParserDataToHtml();

	private static final String EMPTY_STR = "";

	private final String[] asciiHtmlCodes = new String[128 - 32 - 1];// NOPMD

	public static ParserDataToHtml getInstance() {
		return INSTANCE;
	}

	private ParserDataToHtml() {
		super();
		for (char charValue = ' '; charValue < 0x7f; charValue++) {
			asciiHtmlCodes[charValue - ' '] = String.valueOf(charValue);
		}
		asciiHtmlCodes['"' - ' '] = "&quot;";
		asciiHtmlCodes['&' - ' '] = "&amp;";
		asciiHtmlCodes['<' - ' '] = "&lt;";
		asciiHtmlCodes['>' - ' '] = "&gt;";
	}

	public void generateHtml(final ParserData parserData, final File file)
			throws IOException {
		final FileOutputStream fileOutputStream = new FileOutputStream(file);
		OutputStreamWriter outputStreamWriter;
		try {
			outputStreamWriter = new OutputStreamWriter(fileOutputStream,
					"UTF8");
		} catch (UnsupportedEncodingException exception) {
			try {
				fileOutputStream.close();
			} catch (IOException exception2) {// NOPMD
				// ignore error
			}
			throw exception;
		}
		final BufferedWriter bufferedWriter = new BufferedWriter(
				outputStreamWriter);
		try {
			generateHtml(parserData, bufferedWriter);
		} catch (IOException exception) {
			try {
				bufferedWriter.close();
			} catch (IOException exception2) {// NOPMD
				// ignore error
			}
			throw exception;
		}
		bufferedWriter.close();
	}

	private void generateHtml(final ParserData parserData,
			final BufferedWriter bufferedWriter) throws IOException {

		writeHeader(bufferedWriter);

		final Map<Integer, NonTerminal> nonTerminalNameMap = parserData
				.getNonTerminalByIdentifierMap();
		final GotoByNonTerminalEntry[] gotoByNonTerminalEntries = parserData
				.getGotoByNonTerminalTable();
		final GotoByNonTerminalEntry[] tokenGotoByNonTerminalEntries = parserData
				.getTokenGotoByNonTerminalTable();
		final GotoByTerminalEntry[] gotoByTerminalEntries = parserData
				.getGotoByTerminalTable();
		final GotoByTerminalEntry[] tokenGotoByTerminalEntries = parserData
				.getTokenGotoByTerminalTable();
		final boolean[] tokenRecognition = parserData.getTokenRecognition();
		final MatchEntrySet[] matchEntrySets = parserData.getMatchTable();
		final MatchEntrySet[] tokenMatchEntrySets = parserData
				.getTokenMatchTable();
		final StateRecognitionPointList[] stateRecognitionPointLists = parserData
				.getByStateRecognitionPointList();
		final int numberOfState = parserData.getNumberOfState();

		/*
		 * create from state by terminal and non terminal
		 */
		final FromNonTerminalEntry[] fromNonTerminalEntries =
		/**/new FromNonTerminalEntry[numberOfState];
		final FromTerminalEntry[] fromTerminalEntries =
		/**/new FromTerminalEntry[numberOfState];
		createFromState(numberOfState, gotoByNonTerminalEntries,
				tokenGotoByNonTerminalEntries, gotoByTerminalEntries,
				tokenGotoByTerminalEntries, fromNonTerminalEntries,
				fromTerminalEntries);

		/*
		 * write state
		 */
		for (int stateNumber = 0; stateNumber < numberOfState; stateNumber++) {
			writeState(tokenRecognition[stateNumber], bufferedWriter,
					nonTerminalNameMap, gotoByNonTerminalEntries[stateNumber],
					tokenGotoByNonTerminalEntries[stateNumber],
					gotoByTerminalEntries[stateNumber],
					tokenGotoByTerminalEntries[stateNumber],
					matchEntrySets[stateNumber],
					tokenMatchEntrySets[stateNumber], stateNumber,
					fromNonTerminalEntries[stateNumber],
					fromTerminalEntries[stateNumber],
					stateRecognitionPointLists[stateNumber]);
		}

		writeTailer(bufferedWriter);

	}

	private void createFromState(final int numberOfState,
			final GotoByNonTerminalEntry[] gotoByNonTerminalEntries,
			final GotoByNonTerminalEntry[] tokenGotoByNonTerminalEntries,
			final GotoByTerminalEntry[] gotoByTerminalEntries,
			final GotoByTerminalEntry[] tokenGotoByTerminalEntries,
			final FromNonTerminalEntry[] fromNonTerminalEntries,
			final FromTerminalEntry[] fromTerminalEntries) {

		for (int stateNumber = 0; stateNumber < numberOfState; stateNumber++) {
			fromNonTerminalEntries[stateNumber] =
			/**/new FromNonTerminalEntry();// NOPMD
			fromTerminalEntries[stateNumber] =
			/**/new FromTerminalEntry();// NOPMD
		}

		for (int stateNumber = 0; stateNumber < numberOfState; stateNumber++) {

			GotoByNonTerminalEntry gotoByNonTerminalEntry =
			/**/gotoByNonTerminalEntries[stateNumber];
			Map<Integer, ToStateAndWhiteSpaceFlag> gotoByNonTerminalMap =
			/**/gotoByNonTerminalEntry.getGotoMap();
			updateFromNonTerminalEntries(stateNumber, gotoByNonTerminalMap,
					fromNonTerminalEntries);

			gotoByNonTerminalEntry =
			/**/tokenGotoByNonTerminalEntries[stateNumber];
			gotoByNonTerminalMap =
			/**/gotoByNonTerminalEntry.getGotoMap();
			updateFromNonTerminalEntries(stateNumber, gotoByNonTerminalMap,
					fromNonTerminalEntries);

			GotoByTerminalEntry gotoByTerminalEntry =
			/**/gotoByTerminalEntries[stateNumber];
			Map<TerminalCharRange, Integer> gotoByTerminalMap =
			/**/gotoByTerminalEntry.getGotoMap();
			updateFromTerminalEntries(stateNumber, gotoByTerminalMap,
					fromTerminalEntries);

			gotoByTerminalEntry =
			/**/tokenGotoByTerminalEntries[stateNumber];
			gotoByTerminalMap =
			/**/gotoByTerminalEntry.getGotoMap();
			updateFromTerminalEntries(stateNumber, gotoByTerminalMap,
					fromTerminalEntries);
		}
	}

	private void updateFromTerminalEntries(final int stateNumber,
			final Map<TerminalCharRange, Integer> gotoByTerminalMap,
			final FromTerminalEntry[] fromTerminalEntries) {
		for (Map.Entry<TerminalCharRange, Integer> entry : gotoByTerminalMap
				.entrySet()) {
			final TerminalCharRange terminalCharRange = entry.getKey();
			final int toState = entry.getValue();
			fromTerminalEntries[toState]
					.addFrom(stateNumber, terminalCharRange);
		}
	}

	private void updateFromNonTerminalEntries(final int stateNumber,
			final Map<Integer, ToStateAndWhiteSpaceFlag> gotoByNonTerminalMap,
			final FromNonTerminalEntry[] fromNonTerminalEntries) {
		for (Map.Entry<Integer, ToStateAndWhiteSpaceFlag> entry : gotoByNonTerminalMap
				.entrySet()) {
			final int grammarSymbol = entry.getKey();
			final ToStateAndWhiteSpaceFlag value = entry.getValue();
			final int toState = value.getToState();
			fromNonTerminalEntries[toState].addFrom(stateNumber, grammarSymbol);
		}
	}

	private void writeHeader(final BufferedWriter bufferedWriter)
			throws IOException {
		bufferedWriter
				.write(
				/**/"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
						+
						/**/"<html>\n"
						+
						/**/"<head>\n"
						+
						/**/"  <meta content=\"text/html;charset=UTF-8\" http-equiv=\"Content-Type\">\n"
						+
						/**/"  <title>parser data</title>\n"
						+
						/**/"  <meta content=\"description\" name=\"parser data\">\n"
						+
						/**/"</head>\n" +
						/**/"<body>\n" +
						/**/"<span style=\"font-family: monospace;\">\n");
	}

	private void writeState(final boolean tokenRecognition,
			final BufferedWriter bufferedWriter,
			final Map<Integer, NonTerminal> nonTerminalNameMap,
			final GotoByNonTerminalEntry gotoByNonTerminalEntry,
			final GotoByNonTerminalEntry tokenGotoByNonTerminalEntry,
			final GotoByTerminalEntry gotoByTerminalEntry,
			final GotoByTerminalEntry tokenGotoByTerminalEntry,
			final MatchEntrySet matchEntrySet,
			final MatchEntrySet tokenMatchEntrySet, final int stateNumber,
			final FromNonTerminalEntry fromNonTerminalEntry,
			final FromTerminalEntry fromTerminalEntry,
			final StateRecognitionPointList stateRecognitionPointList)
			throws IOException {
		final StringBuilder stringBuilder = new StringBuilder();
		// <a name="state0"></a>state #0<br>
		stringBuilder.append("<a name=\"state");
		stringBuilder.append(stateNumber);
		stringBuilder
				.append("\"></a><span style=\"font-weight: bold; text-decoration: underline;\">state #");
		stringBuilder.append(stateNumber);
		if (tokenRecognition) {
			stringBuilder.append(" (token)");
		}
		stringBuilder.append("</span><br><br>\n");

		stringBuilder.append("<div style=\"margin-left: 40px;\">\n");

		stringBuilder.append("goto by non terminal:<br>\n");
		stringBuilder.append("<ul>\n");// NOPMD
		Map<Integer, ToStateAndWhiteSpaceFlag> gotoByNonTerminalMap =
		/**/gotoByNonTerminalEntry.getGotoMap();
		writeGotoByNonTerminal(nonTerminalNameMap, stringBuilder,
				gotoByNonTerminalMap);
		stringBuilder.append("</ul>\n");// NOPMD

		stringBuilder.append("token goto by non terminal:<br>\n");
		stringBuilder.append("<ul>\n");// NOPMD
		gotoByNonTerminalMap =
		/**/tokenGotoByNonTerminalEntry.getGotoMap();
		writeGotoByNonTerminal(nonTerminalNameMap, stringBuilder,
				gotoByNonTerminalMap);
		stringBuilder.append("</ul>\n");// NOPMD

		stringBuilder.append("goto by terminal:<br>\n");
		stringBuilder.append("<ul>\n");
		Map<TerminalCharRange, Integer> gotoByTerminalMap = gotoByTerminalEntry
				.getGotoMap();
		writeGotoByTerminal(stringBuilder, gotoByTerminalMap);
		stringBuilder.append("</ul>\n");

		stringBuilder.append("token goto by terminal:<br>\n");
		stringBuilder.append("<ul>\n");
		gotoByTerminalMap = tokenGotoByTerminalEntry.getGotoMap();
		writeGotoByTerminal(stringBuilder, gotoByTerminalMap);
		stringBuilder.append("</ul>\n");

		stringBuilder.append("non token match:<br>\n");
		writeMatch(matchEntrySet.getNonTerminalMatchSet(), nonTerminalNameMap,
				stringBuilder);

		stringBuilder.append("token match:<br>\n");
		writeMatch(tokenMatchEntrySet.getNonTerminalMatchSet(),
				nonTerminalNameMap, stringBuilder);

		stringBuilder.append("from<br>\n");

		stringBuilder.append("<ul>\n");

		// <li><a href="#state0">state #0</a> by terminal name </li>

		final List<Pair<Integer, Integer>> listFromNonTerminal =
		/**/fromNonTerminalEntry.getList();
		for (Pair<Integer, Integer> pair : listFromNonTerminal) {
			final int fromStateNumber = pair.getMemberA();
			final int grammarSymbol = pair.getMemberB();
			stringBuilder.append("<li><a href=\"#state");
			stringBuilder.append(fromStateNumber);
			stringBuilder.append("\">state #");// NOPMD
			stringBuilder.append(fromStateNumber);
			stringBuilder.append("</a> by terminal ");
			stringBuilder.append(nonTerminalNameMap.get(grammarSymbol)
					.getName());
			stringBuilder.append(" </li>\n");
		}

		final List<Pair<Integer, TerminalCharRange>> listFromTerminal =
		/**/fromTerminalEntry.getList();
		for (Pair<Integer, TerminalCharRange> pair : listFromTerminal) {
			final int fromStateNumber = pair.getMemberA();
			final TerminalCharRange terminalCharRange = pair.getMemberB();
			stringBuilder.append("<li><a href=\"#state");
			stringBuilder.append(fromStateNumber);
			stringBuilder.append("\">state #");
			stringBuilder.append(fromStateNumber);
			stringBuilder.append("</a> by non terminal ");
			stringBuilder.append(stringToHtml(terminalCharRange.toString()));
			stringBuilder.append(" </li>\n");
		}

		stringBuilder.append("</ul>\n");

		stringBuilder.append("rule recognition:<br>\n");
		stringBuilder.append("<ul>\n");
		writeRecognition(nonTerminalNameMap, stringBuilder,
				stateRecognitionPointList);
		stringBuilder.append("</ul>\n");

		stringBuilder.append("</div>\n");

		bufferedWriter.write(stringBuilder.toString());
	}

	private void writeMatch(final Set<MatchEntry> set,
			final Map<Integer, NonTerminal> nonTerminalNameMap,
			final StringBuilder stringBuilder) {
		stringBuilder.append("<ul>\n");
		// <li>ws</li>
		for (MatchEntry entry : set) {
			final int identifier = entry.getMatchedNonTerminalIdentifier();
			final NonTerminal nonTerminal = nonTerminalNameMap.get(identifier);
			stringBuilder.append("<li>\n");
			stringBuilder.append(nonTerminal.getName());
			if (nonTerminal.isToken()) {
				stringBuilder.append(" (token)");
			}
			stringBuilder.append("</li>\n");
		}
		stringBuilder.append("</ul>\n");
	}

	private void writeGotoByTerminal(final StringBuilder stringBuilder,
			final Map<TerminalCharRange, Integer> gotoByTerminalMap) {
		// <li>by [] goto <a href="#state0">state #0</a></li>
		for (Map.Entry<TerminalCharRange, Integer> entry : gotoByTerminalMap
				.entrySet()) {
			final TerminalCharRange terminalCharRange = entry.getKey();
			final int toState = entry.getValue();
			stringBuilder.append("<li>by ");
			stringBuilder.append(stringToHtml(terminalCharRange.toString()));
			stringBuilder.append(" goto <a href=\"#state");
			stringBuilder.append(toState);
			stringBuilder.append("\">state #");
			stringBuilder.append(toState);
			stringBuilder.append("</a></li>\n");
		}
	}

	private void writeGotoByNonTerminal(
			final Map<Integer, NonTerminal> nonTerminalNameMap,
			final StringBuilder stringBuilder,
			final Map<Integer, ToStateAndWhiteSpaceFlag> gotoByNonTerminalMap) {
		// <li>by ws goto <a href="#state0">state #0</a></li>
		for (Map.Entry<Integer, ToStateAndWhiteSpaceFlag> entry : gotoByNonTerminalMap
				.entrySet()) {
			final int grammarSymbol = entry.getKey();
			final ToStateAndWhiteSpaceFlag value = entry.getValue();
			final int toState = value.getToState();
			final boolean whiteSpace = value.isWhiteSpace();
			final NonTerminal nonTerminal = nonTerminalNameMap
					.get(grammarSymbol);
			stringBuilder.append("<li>by ");
			stringBuilder.append(nonTerminal.getName());
			if (nonTerminal.isToken()) {
				stringBuilder.append(" (token)");
			}
			if (whiteSpace) {
				stringBuilder.append(" (as white space)");
			}
			stringBuilder.append(" goto <a href=\"#state");
			stringBuilder.append(toState);
			stringBuilder.append("\">state #");// NOPMD
			stringBuilder.append(toState);
			stringBuilder.append("</a></li>\n");
		}
	}

	private void writeRecognition(
			final Map<Integer, NonTerminal> nonTerminalNameMap,
			final StringBuilder stringBuilder,
			final StateRecognitionPointList stateRecognitionPointList) {

		final List<ParserDataRecognitionPoint> list =
		/**/stateRecognitionPointList.getList();
		stringBuilder.append("<ul>\n");
		for (ParserDataRecognitionPoint recognitionPoint : list) {
			stringBuilder.append("<li>");
			final int identifier = recognitionPoint
					.getDefinedNonTerminalIdentifier();
			String name = nonTerminalNameMap.get(identifier).getName();
			final boolean token = recognitionPoint.isToken();
			if (token) {
				stringBuilder.append('[');
			}
			stringBuilder.append(name);
			if (token) {
				stringBuilder.append(']');
			}
			stringBuilder.append(" :");
			final int recognitionPointIndex = recognitionPoint
					.getRecognitionPointIndex();
			final int[] rightPart = recognitionPoint.getRightPart();
			final boolean concat = recognitionPoint.isConcat();
			int index = 0;
			for (int rightIdentifier : rightPart) {
				stringBuilder.append(' ');
				if (!concat || index == recognitionPointIndex) {
					stringBuilder.append('.');
				}
				if (rightIdentifier == -1) {
					stringBuilder.append(stringToHtml("<terminal>"));
				} else {
					name = nonTerminalNameMap.get(rightIdentifier).getName();
					stringBuilder.append(name);
					if (!concat && index != rightPart.length - 1) {
						stringBuilder.append(" |");
					}
				}
				index++;
			}
			if (index == recognitionPointIndex) {
				stringBuilder.append(" .");
			}
			stringBuilder.append("</li>");
		}
		stringBuilder.append("</ul>\n");
	}

	private void writeTailer(final BufferedWriter bufferedWriter)
			throws IOException {
		bufferedWriter.write(
		/**/"</span>\n" +
		/**/"</body>\n" +
		/**/"</html>\n");
	}

	private String stringToHtml(final String string) {
		final StringBuilder stringBuilder = new StringBuilder(
				string.length() * 2);
		for (int index = 0; index < string.length(); index++) {
			stringBuilder.append(charToHtml(string.charAt(index)));
		}
		return stringBuilder.toString();
	}

	private String charToHtml(final char charValue) {
		final String charToHtml;
		if (charValue >= ' ' && charValue < 0x7f) {
			charToHtml = asciiHtmlCodes[charValue - ' '];
		} else {
			charToHtml = EMPTY_STR;
		}
		return charToHtml;
	}
}
