/*
 * Copyright 2009 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.generator.service;

import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import net.sf.parser4j.generator.entity.grammardefnode.GrammarDef;
import net.sf.parser4j.generator.entity.grammardefnode.IGrammarDefNode;
import net.sf.parser4j.generator.entity.grammardefnode.RuleDef;
import net.sf.parser4j.generator.entity.grammardefnode.RulesDef;
import net.sf.parser4j.generator.entity.grammardefnode.RulesGroupListDef;
import net.sf.parser4j.generator.entity.grammardefnode.StatementDef;
import net.sf.parser4j.generator.service.grammardefnode.IGrammarDefVisitor;
import net.sf.parser4j.kernelgenerator.service.GeneratorException;

/**
 * Update statement definition token son state using
 * {@link StatementDef#setTokenSon(boolean)}<br>
 * Also create set of not referenced non terminal definition<br>
 * 
 * @author luc peuvrier
 * 
 */
public class StatementDefTokenSonUpdater implements IGrammarDefVisitor {

	private static final String START_NON_TERMINAL = "Start";

	private static final String EMPTY_STRING = "";

	private transient Map<String, StatementDef> statementDefByNonTerminalNameMap;

	private transient boolean tokenSon;

	private transient Set<String> visitedSet;

	private transient Set<String> notReferenced;

	public void update(final IGrammarDefNode rootData)
			throws GeneratorException {
		final GrammarDef grammarDef = (GrammarDef) rootData;
		statementDefByNonTerminalNameMap = grammarDef
				.getStatementDefByNonTerminalNameMap();
		final Set<String> whiteSpaceNameSet = grammarDef.getWhiteSpaceNameSet();
		notReferenced = new TreeSet<String>();
		notReferenced.addAll(statementDefByNonTerminalNameMap.keySet());
		StatementDef statementDef = getStartStatementDef();
		visitedSet = new TreeSet<String>();
		tokenSon = false;
		/* do the visits */
		try {
			statementDef.accept(this);
			for (String whiteSpaceName : whiteSpaceNameSet) {
				if (!EMPTY_STRING.equals(whiteSpaceName)) {
					statementDef = statementDefByNonTerminalNameMap
							.get(whiteSpaceName);
					if (statementDef == null) {
						throw new GeneratorException(whiteSpaceName
								+ " non terminal must be defined");
					}
					statementDef.accept(this);
				}
			}
		} catch (GrammarDefNodeVisitException exception) {
			throw new GeneratorException(exception);
		}
		visitedSet = null;// NOPMD free memory
		// notReferenced.removeAll(whiteSpaceNameSet);
	}

	private StatementDef getStartStatementDef() throws GeneratorException {
		final StatementDef startStatementDef = statementDefByNonTerminalNameMap
				.get(START_NON_TERMINAL);
		if (startStatementDef == null) {
			throw new GeneratorException("Start non terminal must be defined");
		}
		if (startStatementDef.isToken()) {
			throw new GeneratorException(
					"Start non terminal can not be a token");
		}
		return startStatementDef;
	}

	/**
	 * @return the notReferenced
	 */
	public Set<String> getNotReferenced() {
		return notReferenced;
	}

	@Override
	public void beginVisit(final GrammarDef def)
			throws GrammarDefNodeVisitException {
		unsupported();
	}

	@Override
	public void endVisit(final GrammarDef def)
			throws GrammarDefNodeVisitException {
		unsupported();
	}

	@Override
	public void beginVisit(final RulesGroupListDef def)
			throws GrammarDefNodeVisitException {
		unsupported();
	}

	@Override
	public void endVisit(final RulesGroupListDef def)
			throws GrammarDefNodeVisitException {
		unsupported();
	}

	@Override
	public void beginVisit(final RulesDef def)
			throws GrammarDefNodeVisitException {
		unsupported();
	}

	@Override
	public void endVisit(final RulesDef def)
			throws GrammarDefNodeVisitException {
		unsupported();
	}

	@Override
	public void beginVisit(final RuleDef def)
			throws GrammarDefNodeVisitException {
		unsupported();
	}

	@Override
	public void endVisit(final RuleDef def) throws GrammarDefNodeVisitException {
		unsupported();
	}

	private void unsupported() {
		throw new UnsupportedOperationException("must not be visited");
	}

	@Override
	public boolean beginVisit(final StatementDef def)
			throws GrammarDefNodeVisitException {
		final boolean visit;
		final String nonTerminalDefinedName = def.getNonTerminalDefinedName();
		if (visitedSet.contains(nonTerminalDefinedName)) {
			visit = false;
		} else {
			visitedSet.add(nonTerminalDefinedName);
			try {
				def.setTokenSon(tokenSon);
			} catch (GeneratorException exception) {
				throw new GrammarDefNodeVisitException(exception);
			}
			if (def.isNonTerminalDefType()) {
				final String nonTerminalName = def.getNonTerminalName();
				visitStatementDef(nonTerminalName);
				final String whiteSpaceNonTerminalName = def
						.getWhiteSpaceNonTerminalName();
				if (!EMPTY_STRING.equals(whiteSpaceNonTerminalName)) {
					visitStatementDef(whiteSpaceNonTerminalName);
				}
			}
			visit = true;
		}
		return visit;
	}

	private void visitStatementDef(final String nonTerminalName)
			throws GrammarDefNodeVisitException {
		notReferenced.remove(nonTerminalName);
		final StatementDef statementDef = statementDefByNonTerminalNameMap
				.get(nonTerminalName);
		if (statementDef == null) {
			throw new GrammarDefNodeVisitException(
					"missing definition for non terminal " + nonTerminalName);
		}
		final boolean currentToken = tokenSon;
		if (statementDef.isToken()) {
			tokenSon = true;
		}
		statementDef.accept(this);
		tokenSon = currentToken;
	}

	@Override
	public void endVisit(final StatementDef def)
			throws GrammarDefNodeVisitException {
		visitedSet.remove(def.getNonTerminalDefinedName());
	}
}
