/*
 * Copyright 2009 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.generator.service;

import java.util.Set;

import net.sf.parser4j.generator.entity.grammardefnode.EnumStatementDefType;
import net.sf.parser4j.generator.entity.grammardefnode.StatementDef;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.GrammarNodeNull;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.IGrammarNode;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeAlternative;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeConcat;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeEmpty;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeString;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeTerminalChar;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeTerminalCharRange;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeTerminalCharSet;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeTerminalSetOfTerminal;
import net.sf.parser4j.kernelgenerator.service.GeneratorException;
import net.sf.parser4j.parser.entity.data.NonTerminalMap;

/**
 * 
 * @author luc peuvrier
 * 
 */
public final class StatementToGrammarNode {

	private final static StatementToGrammarNode INSTANCE =
	/**/new StatementToGrammarNode();

	public static StatementToGrammarNode getInstance() {
		return INSTANCE;
	}

	private StatementToGrammarNode() {
		super();
	}

	/**
	 * create {@link IGrammarNode} from {@link StatementDef}<br>
	 * for {@link StatementDef} of type
	 * {@link EnumStatementDefType#NON_TERMINAL} witch for match manager class
	 * name is not defined will create a null {@link IGrammarNode}
	 * 
	 * @param statementDef
	 *            source
	 * @param nonTerminalMap
	 *            to update with non terminal defined by {@link StatementDef}
	 * @return grammar node from {@link StatementDef}
	 * @throws GeneratorException
	 */
	public IGrammarNode create(final StatementDef statementDef,
			final NonTerminalMap nonTerminalMap) throws GeneratorException {
		final IGrammarNode grammarNode;
		final EnumStatementDefType type = statementDef.getStatementDefType();
		switch (type) {
		case ALTERNATIVE:
			grammarNode = new GrammarNodeAlternative();
			break;
		case CONCAT:
			grammarNode = new GrammarNodeConcat();
			break;
		case EMPTY:
			grammarNode = new GrammarNodeEmpty();
			break;
		case NON_TERMINAL:
			/*
			 * by default no grammar node associated to non terminal statement
			 * definition
			 */
			grammarNode = new GrammarNodeNull(statementDef.getNonTerminalName());
			break;
		case TERMINAL_CHAR:
			final GrammarNodeTerminalChar grammarNodeTerminalChar =
			/**/new GrammarNodeTerminalChar();
			final char terminalValue = statementDef.getTerminalCharValue();
			grammarNodeTerminalChar.setTerminalValue(terminalValue);
			grammarNode = grammarNodeTerminalChar;
			break;
		case TERMINAL_CHAR_RANGE:
			final GrammarNodeTerminalCharRange grammarNodeTerminalCharRange =
			/**/new GrammarNodeTerminalCharRange();
			final char firstTerminalValue = statementDef
					.getFirstTerminalValue();
			grammarNodeTerminalCharRange
					.setFirstTerminalValue(firstTerminalValue);
			final char lastTerminalValue = statementDef.getLastTerminalValue();
			grammarNodeTerminalCharRange
					.setLastTerminalValue(lastTerminalValue);
			grammarNode = grammarNodeTerminalCharRange;
			break;
		case TERMINAL_CHAR_SET:
			final GrammarNodeTerminalCharSet grammarNodeTerminalCharSet =
			/**/new GrammarNodeTerminalCharSet();
			final Set<Integer> charValueSet = statementDef.getTerminalSet();
			grammarNodeTerminalCharSet.setCharValueSet(charValueSet);
			grammarNode = grammarNodeTerminalCharSet;
			break;
		case TERMINAL_SET_OF_TERMINAL:
			final GrammarNodeTerminalSetOfTerminal
			/**/grammarNodeTerminalSetOfTerminal =
			/*  */new GrammarNodeTerminalSetOfTerminal();
			grammarNode = grammarNodeTerminalSetOfTerminal;
			break;
		case TERMINAL_STRING:
			grammarNode = new GrammarNodeString();
			break;
		default:
			throw new IllegalArgumentException("bad type " + type);
		}
		grammarNodeSetup(statementDef, nonTerminalMap, grammarNode);
		return grammarNode;
	}

	/**
	 * Setup of token flag, non terminal identifier, and match manager class
	 * name of grammar node according to statement definition state<br>
	 * If does not add statement definition non terminal to terminal map, the
	 * statement definition terminal must exist in terminal map ( non terminal
	 * identifier is know for non terminal name)<br>
	 * 
	 * @param statementDef
	 *            the definition used to setup the grammar node
	 * @param nonTerminalMap
	 *            non terminal identifier and name map
	 * @param grammarNode
	 *            the grammar node to setup
	 * @throws GeneratorException
	 */
	private void grammarNodeSetup(final StatementDef statementDef,
			final NonTerminalMap nonTerminalMap, final IGrammarNode grammarNode)
			throws GeneratorException {
		final String nonTerminalName = statementDef.getNonTerminalDefinedName();
		final boolean token = statementDef.isToken();
		final String matchClassName = statementDef.getMatchClassName();
		grammarNodeSetup(grammarNode, nonTerminalMap, nonTerminalName, token,
				matchClassName);
	}

	private void grammarNodeSetup(final IGrammarNode grammarNode,
			final NonTerminalMap nonTerminalMap, final String nonTerminalName,
			final boolean token, final String matchClassName)
			throws GeneratorException {
		grammarNode.setNonTerminalName(nonTerminalName);
		final int nonTerminalIdentifier;
		nonTerminalIdentifier = nonTerminalMap.createNonTerminalIdentifier(
				nonTerminalName, matchClassName, token);
		grammarNode.setToken(token);
		grammarNode.setNonTerminalIdentifier(nonTerminalIdentifier);
		grammarNode.setMatchClassSimpleName(matchClassName);
	}
}
