/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.kernelgenerator.entity;

import java.util.Iterator;
import java.util.List;

import net.sf.parser4j.kernelgenerator.entity.grammarnode.IGrammarNode;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.OperandGrammarNode;
import net.sf.parser4j.parser.entity.data.ParserDataRecognitionPoint;
import net.sf.parser4j.parser.entity.data.TerminalCharRange;

/**
 * recognition point: point of recognition in {@link IGrammarNode} ( a basic
 * rule )
 * 
 * @author luc peuvrier
 * 
 */
public class RecognitionPoint implements Comparable<RecognitionPoint> {

	private final IGrammarNode grammarNode;// NOPMD

	private final int recognitionPointIndex;// NOPMD

	private boolean token;// NOPMD

	public RecognitionPoint(final IGrammarNode grammarNode,
			final int recognitionPointIndex, final boolean token) {
		super();
		this.grammarNode = grammarNode;
		this.recognitionPointIndex = recognitionPointIndex;
		this.token = token;
	}

	public IGrammarNode getGrammarNode() {
		return grammarNode;
	}

	public int getRecognitionPointIndex() {
		return recognitionPointIndex;
	}

	/**
	 * 
	 * @return true is this is for a terminal grammar node
	 */
	public boolean isTerminal() {
		return grammarNode.isTerminal();
	}

	/**
	 * 
	 * @return true is this is for a string terminal grammar node
	 */
	public boolean isStringTerminal() {
		return grammarNode.isStringTerminal();
	}

	/**
	 * 
	 * @return list of expected grammar node for this recognition point index
	 * @throws IllegalArgumentException
	 *             bad recognition point index value
	 */
	public List<OperandGrammarNode> expectedNonTerminal() {
		return grammarNode.expectedNonTerminal(recognitionPointIndex);
	}

	/**
	 * 
	 * @return iterator on terminal values
	 */
	public Iterator<TerminalCharRange> terminalValuesIterator() {
		return grammarNode.terminalValuesIterator();
	}

	/**
	 * 
	 * @return non terminal value
	 */
	public int getNonTerminalIdentifier() {
		return grammarNode.getNonTerminalIdentifier();
	}

	/**
	 * 
	 * @return non terminal name
	 */
	public String getNonTerminalName() {
		return grammarNode.getNonTerminalName();
	}

	public int getNodeType() {
		return grammarNode.getGrammarNodeType().ordinal();
	}

	public void setToken(final boolean token) {
		this.token = token;
	}

	public boolean isToken() {
		return token;
	}

	public boolean isTokenGrammarNode() {
		return grammarNode.isToken();
	}

	/**
	 * 
	 * @return true if match for this recognition point index value
	 * @throws IllegalArgumentException
	 *             bad recognition point index value
	 */
	public boolean match() {
		return grammarNode.match(recognitionPointIndex);
	}

	public RecognitionPoint nextRecognitionPoint() {
		return new RecognitionPoint(grammarNode, recognitionPointIndex + 1,
				token);
	}

	public ParserDataRecognitionPoint computeParserDataRecognitionPoint() {
		return new ParserDataRecognitionPoint(grammarNode.getGrammarNodeType(),
				grammarNode.getNonTerminalIdentifier(), grammarNode
						.getRightPart(), recognitionPointIndex, token);
	}

	public int compareTo(final RecognitionPoint recognitionPoint) {
		int compareTo = grammarNode
				.compareTo(recognitionPoint.getGrammarNode());
		if (compareTo == 0) {
			final int otherRecognitionPointIndex = recognitionPoint
					.getRecognitionPointIndex();
			if (recognitionPointIndex > otherRecognitionPointIndex) {
				compareTo = 1;
			} else if (recognitionPointIndex < otherRecognitionPointIndex) {
				compareTo = -1;
			}
		}
		return compareTo;
	}

	@Override
	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result
				+ ((grammarNode == null) ? 0 : grammarNode.hashCode());
		result = PRIME * result + recognitionPointIndex;
		return result;
	}

	@Override
	@SuppressWarnings("PMD")
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final RecognitionPoint other = (RecognitionPoint) obj;
		if (grammarNode == null) {
			if (other.grammarNode != null)
				return false;
		} else if (!grammarNode.equals(other.grammarNode))
			return false;
		if (recognitionPointIndex != other.recognitionPointIndex)
			return false;
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		final StringBuilder builder = new StringBuilder();
		builder.append(grammarNode.toString());
		builder.append("\nrecognize point index=");
		builder.append(recognitionPointIndex);
		builder.append("\nis token=");
		builder.append(token);
		return builder.toString();
	}

}
