/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.kernelgenerator.entity;

import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.OperandGrammarNode;
import net.sf.parser4j.parser.entity.data.NonTerminalMap;
import net.sf.parser4j.parser.entity.data.TerminalCharRange;

/**
 * state is a set of recognition point<br>
 * 
 * @author luc peuvrier
 * 
 */
public class State implements Comparable<State> {// NOPMD

	private static final Logger _log = Logger.getLogger(State.class);

	/** identifier of this state */
	private int stateIdentifier;

	/** recognition point set of this state for non token */
	private final Set<RecognitionPoint> recognitionPointSet = // NOPMD
	/**/new TreeSet<RecognitionPoint>();

	/** recognition point set of this state for token */
	private final Set<RecognitionPoint> tokenRecognitionPointSet = // NOPMD
	/**/new TreeSet<RecognitionPoint>();

	/** recognition point by terminal from this state for non token */
	private final RecognitionPointByTerminalMap recognitionPointByTerminal = // NOPMD
	/**/new RecognitionPointByTerminalMap();

	/** recognition point by terminal from this state for token */
	private final RecognitionPointByTerminalMap tokenRecognitionPointByTerminal = // NOPMD
	/**/new RecognitionPointByTerminalMap();

	/** recognition point by non terminal from this state for non token */
	private final Map<Integer, RecognitionPointSetAndWhiteSpaceFlag>
	/**/recognitionPointByNonTerminal = // NOPMD
	/*  */new TreeMap<Integer, RecognitionPointSetAndWhiteSpaceFlag>();

	/** recognition point by non terminal from this state for token */
	private final Map<Integer, RecognitionPointSetAndWhiteSpaceFlag>
	/**/tokenRecognitionPointByNonTerminal = // NOPMD
	/*  */new TreeMap<Integer, RecognitionPointSetAndWhiteSpaceFlag>();

	/** goto by terminal map for non token */
	private final GotoByTerminalMap gotoByTerminalMap = // NOPMD
	/**/new GotoByTerminalMap(this);

	/** goto by terminal map for token */
	private final GotoByTerminalMap tokenGotoByTerminalMap = // NOPMD
	/**/new GotoByTerminalMap(this);

	/** goto by non terminal map for non token */
	private final Map<Integer, StateAndWhiteSpaceFlag> gotoByNonTerminalMap = // NOPMD
	/**/new TreeMap<Integer, StateAndWhiteSpaceFlag>();

	/** goto by non terminal map for token */
	private final Map<Integer, StateAndWhiteSpaceFlag> tokenGotoByNonTerminalMap = // NOPMD
	/**/new TreeMap<Integer, StateAndWhiteSpaceFlag>();

	/** matched non terminal list for non token */
	private final List<RecognitionPoint> matchList = // NOPMD
	/**/new LinkedList<RecognitionPoint>();

	/** matched non terminal list for token */
	private final List<RecognitionPoint> tokenMatchList = // NOPMD
	/**/new LinkedList<RecognitionPoint>();

	private final NonTerminalMap nonTerminalMap;// NOPMD

	public State(final NonTerminalMap nonTerminalMap) {
		super();
		this.nonTerminalMap = nonTerminalMap;
	}

	public void setStateIdentifier(final int stateIdentifier) {
		this.stateIdentifier = stateIdentifier;
	}

	public int getStateIdentifier() {
		return stateIdentifier;
	}

	/**
	 * add a recognition point to this state
	 * 
	 * @param rootRecognitionPoint
	 *            root recognition point for this state
	 */
	public void addRecognitionPoint(final RecognitionPoint rootRecognitionPoint) {
		final Deque<RecognitionPoint> toAdd = new LinkedList<RecognitionPoint>();
		/* set the first to add : the root */
		toAdd.addLast(rootRecognitionPoint);
		RecognitionPoint recognitionPoint = toAdd.pollFirst();
		/* for all to add */
		while (recognitionPoint != null) {
			if (addRecognitionPointToSet(recognitionPoint)) {
				if (recognitionPoint.match()) {
					/* recognition point match */
					if (recognitionPoint.isToken()) {
						tokenMatchList.add(recognitionPoint);
					} else {
						matchList.add(recognitionPoint);
					}
				} else if (recognitionPoint.isTerminal()
						&& !recognitionPoint.isStringTerminal()) {
					/*
					 * recognition point is terminal
					 */
					final RecognitionPoint nextRecognitionPoint = recognitionPoint
							.nextRecognitionPoint();
					final Iterator<TerminalCharRange> iterator = recognitionPoint
							.terminalValuesIterator();
					while (iterator.hasNext()) {
						final TerminalCharRange terminalCharRange = iterator
								.next();
						addToRecognitionPointByTerminal(terminalCharRange,
								nextRecognitionPoint);
					}
				} else {
					/*
					 * recognition point is non terminal
					 */
					/* compute new recognition point to add */
					final List<OperandGrammarNode> expectedList = recognitionPoint
							.expectedNonTerminal();
					final boolean token = recognitionPoint.isToken();
					for (OperandGrammarNode expectedGrammarNode : expectedList) {
						/*
						 * add new recognition point to add to this state
						 */
						final RecognitionPoint toAddRecognitionPoint =
						/**/newRecognitionPoint(expectedGrammarNode, token);
						toAdd.addLast(toAddRecognitionPoint);

						/*
						 * 
						 */
						final int grammarSymbol = expectedGrammarNode
								.getNonTerminalIdentifier();
						final RecognitionPoint nextRecognitionPoint =
						/**/recognitionPoint.nextRecognitionPoint();
						final boolean whiteSpace = expectedGrammarNode
								.isWhiteSpace();
						addToRecognitionPointByNonTerminal(grammarSymbol,
								nextRecognitionPoint, whiteSpace);
					}
				}
			}
			recognitionPoint = toAdd.pollFirst();
		}
	}

	private RecognitionPoint newRecognitionPoint(
			final OperandGrammarNode expectedGrammarNode, final boolean token) {
		return new RecognitionPoint(
		/**/expectedGrammarNode.getGrammarNode(), 0, token
				|| expectedGrammarNode.isToken());
	}

	private boolean addRecognitionPointToSet(
			final RecognitionPoint recognitionPoint) {
		final boolean token = recognitionPoint.isToken();
		final boolean added;
		if (token) {
			if (tokenRecognitionPointSet.contains(recognitionPoint)) {
				added = false;
			} else {
				tokenRecognitionPointSet.add(recognitionPoint);
				added = true;
			}
		} else {
			if (recognitionPointSet.contains(recognitionPoint)) {
				added = false;
			} else {
				recognitionPointSet.add(recognitionPoint);
				added = true;
			}
		}
		return added;
	}

	private void addToRecognitionPointByTerminal(
			final TerminalCharRange terminalCharRange,
			final RecognitionPoint recognitionPoint) {
		final boolean token = recognitionPoint.isToken();
		if (token) {
			tokenRecognitionPointByTerminal.add(terminalCharRange,
					recognitionPoint);
		} else {
			recognitionPointByTerminal.add(terminalCharRange, recognitionPoint);
		}
	}

	private void addToRecognitionPointByNonTerminal(
			final int nonTerminalIdentifier,
			final RecognitionPoint recognitionPoint, final boolean whiteSpace) {
		if (_log.isDebugEnabled()) {
			_log.debug(nonTerminalMap.getNonTerminalByIdentifierMap().get(
					nonTerminalIdentifier).getName()
					+ "\n"
					+ recognitionPoint
					+ "\nis white space="
					+ whiteSpace);
		}
		final boolean token = recognitionPoint.isToken();
		RecognitionPointSetAndWhiteSpaceFlag recognitionPointSetAndWhiteSpaceFlag;
		if (token) {
			recognitionPointSetAndWhiteSpaceFlag =
			/**/tokenRecognitionPointByNonTerminal.get(nonTerminalIdentifier);
			if (recognitionPointSetAndWhiteSpaceFlag == null) {
				// FIXME par whitSpace do not care ?
				recognitionPointSetAndWhiteSpaceFlag =
				/**/new RecognitionPointSetAndWhiteSpaceFlag(whiteSpace);
				tokenRecognitionPointByNonTerminal.put(nonTerminalIdentifier,
						recognitionPointSetAndWhiteSpaceFlag);
			}
		} else {
			recognitionPointSetAndWhiteSpaceFlag =
			/**/recognitionPointByNonTerminal.get(nonTerminalIdentifier);
			if (recognitionPointSetAndWhiteSpaceFlag == null) {
				recognitionPointSetAndWhiteSpaceFlag =
				/**/new RecognitionPointSetAndWhiteSpaceFlag(whiteSpace);
				recognitionPointByNonTerminal.put(nonTerminalIdentifier,
						recognitionPointSetAndWhiteSpaceFlag);
			} else {
				recognitionPointSetAndWhiteSpaceFlag.setWhiteSpace(whiteSpace);
			}
		}
		final Set<RecognitionPoint> set = recognitionPointSetAndWhiteSpaceFlag
				.getSet();
		set.add(recognitionPoint);
	}

	public Iterator<RecognitionPointForTerminalRange>
	/**/recognitionPointByTerminalIterator() {
		return recognitionPointByTerminal.iterator();
	}

	public Iterator<RecognitionPointForTerminalRange>
	/**/tokenRecognitionPointByTerminalIterator() {
		return tokenRecognitionPointByTerminal.iterator();
	}

	public Iterator<Entry<Integer, RecognitionPointSetAndWhiteSpaceFlag>>
	/**/recognitionPointByNonTerminalIterator() {
		return recognitionPointByNonTerminal.entrySet().iterator();
	}

	public Iterator<Entry<Integer, RecognitionPointSetAndWhiteSpaceFlag>>
	/**/tokenRecognitionPointByNonTerminalIterator() {
		return tokenRecognitionPointByNonTerminal.entrySet().iterator();
	}

	public Set<RecognitionPoint> getRecognitionPointSet() {
		return recognitionPointSet;
	}

	public Set<RecognitionPoint> getTokenRecognitionPointSet() {
		return tokenRecognitionPointSet;
	}

	public void addGotoByTerminal(final TerminalCharRange terminalCharRange,
			final State state) {
		gotoByTerminalMap.addGoto(terminalCharRange, state);
	}

	public void addTokenGotoByTerminal(
			final TerminalCharRange terminalCharRange, final State state) {
		tokenGotoByTerminalMap.addGoto(terminalCharRange, state);
	}

	public void addGotoByNonTerminal(final int nonTerminal, final State state,
			final boolean whiteSpace) {
		if (gotoByNonTerminalMap.put(nonTerminal, new StateAndWhiteSpaceFlag(
				state, whiteSpace)) != null) {
			throw new IllegalArgumentException("entry for non terminal "
					+ nonTerminal + " already exist");
		}
	}

	public void addTokenGotoByNonTerminal(final int nonTerminal,
			final State state) {
		if (tokenGotoByNonTerminalMap.put(nonTerminal,
				new StateAndWhiteSpaceFlag(state, false)) != null) {
			throw new IllegalArgumentException("entry for non terminal "
					+ nonTerminal + " already exist");
		}
	}

	public Iterator<StateByTerminalRange> gotoByTerminalIterator() {
		return gotoByTerminalMap.iterator();
	}

	public Iterator<StateByTerminalRange> tokenGotoByTerminalIterator() {
		return tokenGotoByTerminalMap.iterator();
	}

	public Iterator<Entry<Integer, StateAndWhiteSpaceFlag>> gotoByNonTerminalIterator() {
		return gotoByNonTerminalMap.entrySet().iterator();
	}

	public Iterator<Entry<Integer, StateAndWhiteSpaceFlag>> tokenGotoByNonTerminalIterator() {
		return tokenGotoByNonTerminalMap.entrySet().iterator();
	}

	public Iterator<RecognitionPoint> matchListIterator() {
		return matchList.iterator();
	}

	public Iterator<RecognitionPoint> tokenMatchListIterator() {
		return tokenMatchList.iterator();
	}

	public boolean isTokenRecognition() {
		return !tokenRecognitionPointSet.isEmpty();
	}

	public int compareTo(final State state) {
		int compareTo = 0;
		Iterator<RecognitionPoint> iterator = recognitionPointSet.iterator();
		Iterator<RecognitionPoint> otherIterator = state
				.getRecognitionPointSet().iterator();
		while (compareTo == 0 && iterator.hasNext() && otherIterator.hasNext()) {
			compareTo = iterator.next().compareTo(otherIterator.next());
		}

		if (compareTo == 0) {
			if (iterator.hasNext() && !otherIterator.hasNext()) {
				compareTo = 1;
			} else if (!iterator.hasNext() && otherIterator.hasNext()) {
				compareTo = -1;
			}
		}

		if (compareTo == 0) {
			iterator = tokenRecognitionPointSet.iterator();
			otherIterator = state.getTokenRecognitionPointSet().iterator();
			while (compareTo == 0 && iterator.hasNext()
					&& otherIterator.hasNext()) {
				compareTo = iterator.next().compareTo(otherIterator.next());
			}
		}

		if (compareTo == 0) {
			if (iterator.hasNext() && !otherIterator.hasNext()) {
				compareTo = 1;
			} else if (!iterator.hasNext() && otherIterator.hasNext()) {
				compareTo = -1;
			}
		}

		return compareTo;
	}

	@Override
	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		result = PRIME
				* result
				+ ((recognitionPointSet == null) ? 0 : recognitionPointSet
						.hashCode());
		result = PRIME
				* result
				+ ((tokenRecognitionPointSet == null) ? 0
						: tokenRecognitionPointSet.hashCode());
		return result;
	}

	@Override
	@SuppressWarnings("PMD")
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final State other = (State) obj;
		if (recognitionPointSet == null) {
			if (other.recognitionPointSet != null)
				return false;
		} else if (!recognitionPointSet.equals(other.recognitionPointSet))
			return false;
		if (tokenRecognitionPointSet == null) {
			if (other.tokenRecognitionPointSet != null)
				return false;
		} else if (!tokenRecognitionPointSet
				.equals(other.tokenRecognitionPointSet))
			return false;
		return true;
	}

	@Override
	public String toString() {
		final StringBuilder builder = new StringBuilder();
		builder.append("state #");
		builder.append(stateIdentifier);
		builder.append("\nnon token:\n");
		for (RecognitionPoint recognitionPoint : recognitionPointSet) {
			builder.append(recognitionPoint.getNonTerminalName());
			builder.append(' ');
			builder.append(recognitionPoint.getRecognitionPointIndex());
			builder.append('\n');
		}
		builder.append("\ntoken:\n");
		for (RecognitionPoint recognitionPoint : tokenRecognitionPointSet) {
			builder.append(recognitionPoint.getNonTerminalName());
			builder.append(' ');
			builder.append(recognitionPoint.getRecognitionPointIndex());
			builder.append('\n');
		}

		builder.append("\nnon token goto by terminal\n");
		Iterator<StateByTerminalRange> iteratorByTerminal = gotoByTerminalMap
				.iterator();
		gotoByTerminalToString(builder, iteratorByTerminal);

		builder.append("\ntoken goto by terminal\n");
		iteratorByTerminal = tokenGotoByTerminalMap.iterator();
		gotoByTerminalToString(builder, iteratorByTerminal);

		builder.append("\nnon token goto by non terminal\n");
		Iterator<Entry<Integer, StateAndWhiteSpaceFlag>> iteratorByNonTerminal = gotoByNonTerminalMap
				.entrySet().iterator();
		gotoByNonTerminalToString(builder, iteratorByNonTerminal);

		builder.append("\ntoken goto by non terminal\n");
		iteratorByNonTerminal = tokenGotoByNonTerminalMap.entrySet().iterator();
		gotoByNonTerminalToString(builder, iteratorByNonTerminal);

		builder.append("\nnon token match\n");
		for (RecognitionPoint recognitionPoint : matchList) {
			final int nonTerminalIdentifier = recognitionPoint
					.getNonTerminalIdentifier();
			builder.append(nonTerminalMap
					.nonTerminalName(nonTerminalIdentifier));
			builder.append('\n');
		}

		builder.append("\ntoken match\n");
		for (RecognitionPoint recognitionPoint : tokenMatchList) {
			final int nonTerminalIdentifier = recognitionPoint
					.getNonTerminalIdentifier();
			builder.append(nonTerminalMap
					.nonTerminalName(nonTerminalIdentifier));
			builder.append('\n');
		}

		return builder.toString();
	}

	private void gotoByNonTerminalToString(
			final StringBuilder builder,
			final Iterator<Entry<Integer, StateAndWhiteSpaceFlag>> iteratorByNonTerminal) {
		while (iteratorByNonTerminal.hasNext()) {
			final Entry<Integer, StateAndWhiteSpaceFlag> entry = iteratorByNonTerminal
					.next();
			builder.append(nonTerminalMap.nonTerminalName(entry.getKey()));
			builder.append(" goto ");
			final StateAndWhiteSpaceFlag value = entry.getValue();
			final State state = value.getState();
			builder.append(state.getStateIdentifier());
			if (value.isWhiteSpace()) {
				builder.append(" (as white space)");
			}
			builder.append('\n');
		}
	}

	private void gotoByTerminalToString(final StringBuilder builder,
			final Iterator<StateByTerminalRange> iteratorByTerminal) {
		while (iteratorByTerminal.hasNext()) {
			final StateByTerminalRange stateByTerminalRange = iteratorByTerminal
					.next();
			builder.append(stateByTerminalRange.toString());
			builder.append('\n');
		}
	}
}
