/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.kernelgenerator.service;

import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.Map.Entry;

import net.sf.parser4j.kernelgenerator.entity.RecognitionPoint;
import net.sf.parser4j.kernelgenerator.entity.RecognitionPointForTerminalRange;
import net.sf.parser4j.kernelgenerator.entity.RecognitionPointSetAndWhiteSpaceFlag;
import net.sf.parser4j.kernelgenerator.entity.State;
import net.sf.parser4j.kernelgenerator.entity.StateAndWhiteSpaceFlag;
import net.sf.parser4j.kernelgenerator.entity.StateByTerminalRange;
import net.sf.parser4j.kernelgenerator.entity.StateMap;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.IGrammarNode;
import net.sf.parser4j.parser.entity.data.NonTerminal;
import net.sf.parser4j.parser.entity.data.NonTerminalMap;
import net.sf.parser4j.parser.entity.data.ParserData;
import net.sf.parser4j.parser.entity.data.ParserDataRecognitionPoint;
import net.sf.parser4j.parser.entity.data.TerminalCharRange;
import net.sf.parser4j.parser.entity.data.ToStateAndWhiteSpaceFlag;

import org.apache.log4j.Logger;

/**
 * the {@link ParserData} generator from grammar tree
 * 
 * @author luc peuvrier
 * 
 */
public final class KernelParserDataGenerator {

	private static final Logger _log = Logger
			.getLogger(KernelParserDataGenerator.class);

	private static final KernelParserDataGenerator INSTANCE =
	/**/new KernelParserDataGenerator();

	private transient NonTerminalMap nonTerminalMap;

	public static KernelParserDataGenerator getInstance() {
		return INSTANCE;
	}

	private KernelParserDataGenerator() {
		super();
	}

	/**
	 * 
	 * @param rootGrammarNode
	 * @param nonTerminalMap
	 * @param matchMgrPackageName
	 *            name of package containing the match manager classes
	 * @return
	 */
	public ParserData generate(final IGrammarNode rootGrammarNode,
			final NonTerminalMap nonTerminalMap,
			final String matchMgrPackageName) {
		this.nonTerminalMap = nonTerminalMap;
		final StateMap stateMap = new StateMap();
		final State initialState = createInitialState(rootGrammarNode, stateMap);
		createAllOtherState(stateMap, initialState);
		return createParserData(stateMap, rootGrammarNode, matchMgrPackageName);
	}

	/**
	 * creation of initial state
	 * 
	 * @param rootGrammarNode
	 * @param stateMap
	 * @return
	 */
	private State createInitialState(final IGrammarNode rootGrammarNode,
			final StateMap stateMap) {
		final State state = new State(nonTerminalMap);
		// FIXME start can not be a token
		final RecognitionPoint rootRecognitionPoint = new RecognitionPoint(
				rootGrammarNode, 0, rootGrammarNode.isToken());
		state.addRecognitionPoint(rootRecognitionPoint);
		stateMap.addState(state);
		if (state.getStateIdentifier() != 0) {
			throw new IllegalStateException(
					"initial state identifier must be 0");
		}
		if (_log.isDebugEnabled()) {
			_log.debug("initial state #" + state.getStateIdentifier()
					+ " created");
		}
		return state;
	}

	private void createAllOtherState(final StateMap stateMap,
			final State initialState) {
		final Deque<State> stateQue = new LinkedList<State>();
		stateQue.addLast(initialState);
		State state = stateQue.pollFirst();
		while (state != null) {

			Iterator<Entry<Integer, RecognitionPointSetAndWhiteSpaceFlag>>
			/**/iteratorByNonTerminal =
			/*  */state.recognitionPointByNonTerminalIterator();
			addGotoByNonTerminal(false, stateMap, stateQue, state,
					iteratorByNonTerminal);

			iteratorByNonTerminal = state
					.tokenRecognitionPointByNonTerminalIterator();
			addGotoByNonTerminal(true, stateMap, stateQue, state,
					iteratorByNonTerminal);

			Iterator<RecognitionPointForTerminalRange> iteratorByTerminal =
			/**/state.recognitionPointByTerminalIterator();
			addGotoByTerminal(false, stateMap, stateQue, state,
					iteratorByTerminal);

			iteratorByTerminal =
			/**/state.tokenRecognitionPointByTerminalIterator();
			addGotoByTerminal(true, stateMap, stateQue, state,
					iteratorByTerminal);

			if (_log.isDebugEnabled()) {
				_log.debug("goto added " + state.toString());
			}

			state = stateQue.pollFirst();
		}
	}

	private void addGotoByTerminal(final boolean token,
			final StateMap stateMap, final Deque<State> stateQue,
			final State state,
			final Iterator<RecognitionPointForTerminalRange> iteratorByTerminal) {
		while (iteratorByTerminal.hasNext()) {
			final RecognitionPointForTerminalRange range = iteratorByTerminal
					.next();
			final Set<RecognitionPoint> set = range.getSetOfRecognitionPoint();
			final State currentState = computeGotoState(stateMap, stateQue, set);
			final TerminalCharRange terminalCharRange = range
					.getTerminalCharRange();
			if (token) {
				state.addTokenGotoByTerminal(terminalCharRange, currentState);
			} else {
				state.addGotoByTerminal(terminalCharRange, currentState);
			}
		}
	}

	private void addGotoByNonTerminal(
			final boolean token,
			final StateMap stateMap,
			final Deque<State> stateQue,
			final State state,
			final Iterator<Entry<Integer, RecognitionPointSetAndWhiteSpaceFlag>> iteratorByNonTerminal) {

		while (iteratorByNonTerminal.hasNext()) {
			final Entry<Integer, RecognitionPointSetAndWhiteSpaceFlag> entry =
			/**/iteratorByNonTerminal.next();
			final RecognitionPointSetAndWhiteSpaceFlag value = entry.getValue();
			final Set<RecognitionPoint> set = value.getSet();
			final State currentState = computeGotoState(stateMap, stateQue, set);
			final int nonTerminal = entry.getKey();
			if (token) {
				state.addTokenGotoByNonTerminal(nonTerminal, currentState);
			} else {
				state.addGotoByNonTerminal(nonTerminal, currentState, value
						.isWhiteSpace());
			}
		}
	}

	/**
	 * find or create state for a recognition point set
	 * 
	 * @param stateMap
	 *            existing state set
	 * @param stateQue
	 *            updated with created state if state created
	 * @param set
	 *            recognition point set of the state to find or create
	 * @return existing or created state
	 */
	private State computeGotoState(final StateMap stateMap,
			final Deque<State> stateQue, final Set<RecognitionPoint> set) {
		if (_log.isDebugEnabled()) {
			_log.debug("new state number of recognition point " + set.size());
		}
		final State newState = new State(nonTerminalMap);// NOPMD
		for (RecognitionPoint recognitionPoint : set) {
			newState.addRecognitionPoint(recognitionPoint);
		}
		State currentState = stateMap.existingState(newState);
		if (currentState == null) {
			stateMap.addState(newState);
			currentState = newState;
			stateQue.addLast(newState);
			if (_log.isDebugEnabled()) {
				_log.debug("new state  #" + currentState.getStateIdentifier()
						+ " created");
			}
		}
		return currentState;
	}

	/**
	 * create parser data
	 * 
	 * @param stateMap
	 * @param rootGrammarNode
	 * @param matchMgrPackageName
	 *            name of package containing the match manager classes
	 * @return parser data
	 */
	private ParserData createParserData(final StateMap stateMap,
			final IGrammarNode rootGrammarNode, final String matchMgrPackageName) {
		final int numberOfState = stateMap.numberOfState();
		final int numberOfNonterminal = nonTerminalMap.numberOfNonterminal();
		final ParserData parserData = new ParserData(numberOfState,
				numberOfNonterminal);
		final Map<Integer, NonTerminal> nonTerminalByIdentifierMap = nonTerminalMap
				.getNonTerminalByIdentifierMap();
		if (_log.isDebugEnabled()) {
			logNonTerminalByIdentifierMap(nonTerminalByIdentifierMap);
		}
		parserData.setNonTerminalByIdentifierMap(nonTerminalByIdentifierMap);

		for (int stateIdentifier = 0; stateIdentifier < numberOfState; stateIdentifier++) {
			final State state = stateMap.stateByIdentifier(stateIdentifier);
			final int fromStateIdentifier = state.getStateIdentifier();

			/*
			 * set state token recognition
			 */
			if (state.isTokenRecognition()) {
				parserData.setTokenRecognition(fromStateIdentifier);
			}

			/*
			 * create goto table for non terminal
			 */
			Iterator<Entry<Integer, StateAndWhiteSpaceFlag>> gotoByNonTerminalMapIterator =
			/**/state.gotoByNonTerminalIterator();
			createGotoTableByNonTerminal(parserData, fromStateIdentifier,
					gotoByNonTerminalMapIterator, false);
			gotoByNonTerminalMapIterator =
			/**/state.tokenGotoByNonTerminalIterator();
			createGotoTableByNonTerminal(parserData, fromStateIdentifier,
					gotoByNonTerminalMapIterator, true);
			/*
			 * create goto table for terminal
			 */
			Iterator<StateByTerminalRange> gotoByTerminalIterator = state
					.gotoByTerminalIterator();
			createGotoTableByTerminal(parserData, fromStateIdentifier,
					gotoByTerminalIterator, false);
			gotoByTerminalIterator = state.tokenGotoByTerminalIterator();
			createGotoTableByTerminal(parserData, fromStateIdentifier,
					gotoByTerminalIterator, true);

			/*
			 * create match table
			 */
			Iterator<RecognitionPoint> matchListIterator = state
					.matchListIterator();
			while (matchListIterator.hasNext()) {
				final RecognitionPoint next = matchListIterator.next();
				final int nonTerminalIdentifier = next
						.getNonTerminalIdentifier();
				final int length = next.getRecognitionPointIndex();
				final int nodeType = next.getNodeType();
				final boolean token = next.isTokenGrammarNode();
				parserData.addMatch(fromStateIdentifier, nonTerminalIdentifier,
						length, nodeType, token);
			}

			matchListIterator = state.tokenMatchListIterator();
			while (matchListIterator.hasNext()) {
				final RecognitionPoint next = matchListIterator.next();
				final int nonTerminalIdentifier = next
						.getNonTerminalIdentifier();
				final int length = next.getRecognitionPointIndex();
				final int nodeType = next.getNodeType();
				final boolean token = next.isTokenGrammarNode();
				parserData.addTokenMatch(fromStateIdentifier,
						nonTerminalIdentifier, length, nodeType, token);
			}

			/*
			 * add recognition point
			 */
			final Set<RecognitionPoint> all =
			/**/new TreeSet<RecognitionPoint>();// NOPMD
			all.addAll(state.getRecognitionPointSet());
			all.addAll(state.getTokenRecognitionPointSet());
			final List<ParserDataRecognitionPoint> list =
			/**/new LinkedList<ParserDataRecognitionPoint>();// NOPMD
			for (RecognitionPoint recognitionPoint : all) {
				final ParserDataRecognitionPoint parserDataRecognitionPoint =
				/**/recognitionPoint.computeParserDataRecognitionPoint();
				list.add(parserDataRecognitionPoint);
			}
			parserData.setRecognitionPoint(fromStateIdentifier, list);
		}

		/*
		 * create match manager class table
		 */
		for (int nonTerminalIndex = 0; nonTerminalIndex < numberOfNonterminal; nonTerminalIndex++) {
			final int nonTerminalIdentifier = nonTerminalIndex + 0x10000;
			final String nonTerminalMatchClassSimpleName = nonTerminalMap
					.nonTerminalMatchClassSimpleName(nonTerminalIdentifier);
			final String matchClassName;
			if (nonTerminalMatchClassSimpleName == null) {
				matchClassName = null;
			} else {
				matchClassName = matchMgrPackageName + "."
						+ nonTerminalMatchClassSimpleName;
			}
			parserData.setNonTerminalMatchClassName(nonTerminalIdentifier,
					matchClassName);
		}

		/*
		 * add root non terminal identifier
		 */
		final int rootNonTerminalIdentifier = rootGrammarNode
				.getNonTerminalIdentifier();
		parserData.setRootNonTerminalIdentifier(rootNonTerminalIdentifier);
		return parserData;
	}

	private void createGotoTableByTerminal(final ParserData parserData,
			final int fromStateIdentifier,
			final Iterator<StateByTerminalRange> gotoByTerminalIterator,
			final boolean token) {
		while (gotoByTerminalIterator.hasNext()) {
			final StateByTerminalRange gotoByTerminalEntry = gotoByTerminalIterator
					.next();
			final TerminalCharRange terminalCharRange = gotoByTerminalEntry
					.getTerminalCharRange();
			final State targetState = gotoByTerminalEntry.getState();
			parserData.addGotoByTerminal(token, fromStateIdentifier,
					terminalCharRange, targetState.getStateIdentifier());
		}
	}

	private void createGotoTableByNonTerminal(
			final ParserData parserData,
			final int fromStateIdentifier,
			final Iterator<Entry<Integer, StateAndWhiteSpaceFlag>> gotoByNonTerminalMapIterator,
			final boolean token) {
		while (gotoByNonTerminalMapIterator.hasNext()) {
			final Entry<Integer, StateAndWhiteSpaceFlag> entry = gotoByNonTerminalMapIterator
					.next();
			final int nonTerminalIdentifier = entry.getKey();
			final StateAndWhiteSpaceFlag value = entry.getValue();
			final State targetState = value.getState();
			final boolean whiteSpace = value.isWhiteSpace();
			final ToStateAndWhiteSpaceFlag toStateAndWhiteSpaceFlag =
			/**/new ToStateAndWhiteSpaceFlag(// NOPMD
					/*    */targetState.getStateIdentifier(), whiteSpace);
			parserData.addGotoByNonTerminal(token, fromStateIdentifier,
					nonTerminalIdentifier, toStateAndWhiteSpaceFlag);
		}
	}

	private void logNonTerminalByIdentifierMap(
			final Map<Integer, NonTerminal> nonTerminalByIdentifierMap) {
		final StringBuilder stringBuilder = new StringBuilder();
		stringBuilder.append("non terminal identifier map\n");
		for (Map.Entry<Integer, NonTerminal> entry : nonTerminalByIdentifierMap
				.entrySet()) {
			stringBuilder.append(entry.getKey());
			stringBuilder.append(", ");
			final NonTerminal value = entry.getValue();
			stringBuilder.append(value.getName());
			stringBuilder.append(", is token ");
			stringBuilder.append(value.isToken());
			stringBuilder.append('\n');
		}
		_log.debug(stringBuilder.toString());
	}
}
