/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parserdatatohtml;

import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.HeadlessException;
import java.io.File;
import java.io.IOException;

import javax.swing.JFrame;
import javax.swing.JLayeredPane;
import javax.swing.JProgressBar;
import javax.swing.JTextArea;

import net.sf.parser4j.commonfortool.view.filechooser.HtmlSaveFileChooser;
import net.sf.parser4j.commonfortool.view.filechooser.OpenFileChooser;
import net.sf.parser4j.generator.service.ParserDataToHtml;
import net.sf.parser4j.parser.entity.data.ParserData;
import net.sf.parser4j.parser.service.ParserException;
import net.sf.parser4j.parser.service.data.ParserDataIO;

/**
 * 
 * @author luc peuvrier
 * 
 */
public final class MainParserDataToHtml {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private transient JFrame frame;

	private transient JProgressBar progressBar;

	private transient JTextArea textArea;

	public static void main(final String[] args) {
		final MainParserDataToHtml main = new MainParserDataToHtml();
		main.run();
	}

	private MainParserDataToHtml() throws HeadlessException {
		super();
	}

	private void run() {
		createAndShowGUI();
		javax.swing.SwingUtilities.invokeLater(new CreateAndSelectFile());
	}

	private class CreateAndSelectFile implements Runnable {

		public void run() {
			final File parserDataFile = chooseParserDataFile();
			if (parserDataFile != null) {
				textArea.setVisible(false);
				progressBar.setVisible(true);
				frame.pack();
				new Thread(new ReadParserData(parserDataFile)).start();
			}
		}
	}

	private class ReadParserData implements Runnable {

		private transient final File parserDataFile;

		public ReadParserData(final File parserDataFile) {
			this.parserDataFile = parserDataFile;
		}

		@Override
		public void run() {
			final ParserDataIO parserDataIO = ParserDataIO.getInstance();
			ParserData parserData = null;
			ParserException exception = null;
			try {
				parserData = parserDataIO.readFromFile(parserDataFile);
			} catch (ParserException exception2) {
				exception = exception2;
			}
			javax.swing.SwingUtilities.invokeLater(new SelectHtmlFile(
					parserData, exception));
		}
	}

	private class SelectHtmlFile implements Runnable {

		private final transient ParserData parserData;

		private final transient ParserException exception;

		public SelectHtmlFile(final ParserData parserData,
				final ParserException exception) {
			super();
			this.parserData = parserData;
			this.exception = exception;
		}

		@Override
		public void run() {
			progressBar.setVisible(false);
			if (parserData == null) {
				textArea.setText(exception.toString());
				textArea.setCaretPosition(0);
				textArea.setVisible(true);
				frame.pack();
			} else {
				frame.pack();
				final File htmlFile = chooseHtmlFile();
				if (htmlFile != null) {
					textArea.setVisible(false);
					progressBar.setVisible(true);
					new Thread(new SaveToHtml(parserData, htmlFile)).start();
				}
			}
		}
	}

	private class SaveToHtml implements Runnable {

		private final transient ParserData parserData;

		private final transient File htmlFile;

		public SaveToHtml(final ParserData parserData, final File htmlFile) {
			super();
			this.parserData = parserData;
			this.htmlFile = htmlFile;
		}

		@Override
		public void run() {
			final ParserDataToHtml parserDataToHtml = ParserDataToHtml
					.getInstance();
			IOException exception = null;
			try {
				parserDataToHtml.generateHtml(parserData, htmlFile);
			} catch (IOException exception2) {
				exception = exception2;
			}
			javax.swing.SwingUtilities.invokeLater(new End(exception));
		}
	}

	private class End implements Runnable {

		private final transient IOException exception;

		public End(final IOException exception) {
			this.exception = exception;
		}

		@Override
		public void run() {
			progressBar.setVisible(false);
			textArea.setVisible(true);
			if (exception == null) {
				textArea.setText("done");
			} else {
				textArea.setText(exception.toString());
			}
			textArea.setCaretPosition(0);
			frame.pack();
		}
	}

	private void createAndShowGUI() {
		frame = new JFrame("Parser data to html");
		frame.setMenuBar(null);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		final JLayeredPane layeredPane = new JLayeredPane();
		layeredPane.setLayout(new FlowLayout());
		progressBar = new JProgressBar();
		final Dimension dim = new Dimension(200, 50);
		progressBar.setPreferredSize(dim);
		progressBar.setMaximumSize(dim);
		progressBar.setMinimumSize(dim);
		progressBar.setEnabled(true);
		progressBar.setIndeterminate(true);
		progressBar.setVisible(false);
		layeredPane.add(progressBar, 0);

		textArea = new JTextArea(5, 40);
		textArea.setEditable(false);
		textArea.setEnabled(true);
		textArea.setVisible(false);
		layeredPane.add(textArea, 1);

		frame.setContentPane(layeredPane);

		frame.pack();
		frame.setVisible(true);
	}

	private File chooseParserDataFile() {
		final File parserDataFile;
		final OpenFileChooser fileChooser = new OpenFileChooser(frame,
				"open parser data", ".");
		if (fileChooser.approve()) {
			parserDataFile = fileChooser.getSelectedFile();
		} else {
			parserDataFile = null;
			textArea.setText("open parser data file canceled");
			textArea.setCaretPosition(0);
			textArea.setVisible(true);
		}
		return parserDataFile;
	}

	private File chooseHtmlFile() {
		final File htmlFile;
		final HtmlSaveFileChooser fileChooser = new HtmlSaveFileChooser(frame,
				"save parser data to html file", ".");
		if (fileChooser.approve()) {
			htmlFile = fileChooser.getSelectedFile();
		} else {
			htmlFile = null;
			textArea.setText("select file for save parser data canceled");
			textArea.setCaretPosition(0);
			textArea.setVisible(true);
		}
		return htmlFile;
	}
}
