/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parser.entity;

import java.util.Set;
import java.util.TreeSet;

import net.sf.parser4j.parser.entity.data.TerminalCharRange;
import net.sf.parser4j.parser.entity.parsenode.IParseNode;
import net.sf.parser4j.parser.entity.parsenode.data.IParseNodeData;
import net.sf.parser4j.parser.entity.parsenode.status.IParseNodeInErrorStatus;
import net.sf.parser4j.parser.entity.parsestate.ParseStack;
import net.sf.parser4j.parser.service.DebugUtil;
import net.sf.parser4j.parser.service.ParserException;
import net.sf.parser4j.parser.service.parsenode.ParsingToStringUtil;

/**
 * parse result
 * 
 * @author luc peuvrier
 * 
 */
public class ParseResult {

	private static final ParsingToStringUtil parsingToStringUtil =
	/**/ParsingToStringUtil.getInstance();

	/** true if error during parsing */
	private final boolean inError;// NOPMD

	/** true if syntax parsing error */
	private boolean parseError;

	/** true if reduce error while parsing */
	private boolean reduceError;

	/** file name where is syntax error */
	private String fileName; // NOPMD

	/** line number in input text file for syntax error */
	private int lineNumber = -1;// NOPMD

	/** column number in input text file for syntax error */
	private int columnNumber = -1;// NOPMD

	/** last terminal value readed */
	private int lastTerminalValueReaded = -1;// NOPMD

	/** expected terminal not found in input text file for syntax error */
	private final Set<TerminalCharRange> expectedTerminal = // NOPMD
	/**/new TreeSet<TerminalCharRange>();

	/** expected non terminal not found in input text file for syntax error */
	private final Set<String> expectedNonTerminal = // NOPMD
	/**/new TreeSet<String>();

	/** the reduce error exception cause of reduce error */
	private ParserException reduceException;// NOPMD

	/** the error status for error cause of reduce error */
	private IParseNodeInErrorStatus parseNodeInErrorStatus;// NOPMD

	/** file name where is reduce error */
	private String reduceErrorFileName; // NOPMD

	/** line number in input text file for reduce error */
	private int reduceErrorLineNumber;// NOPMD

	/** column number in input text file for reduce error */
	private int reduceErrorColumnNumber;// NOPMD

	/** the resulting parse tree root node if no syntax error */
	private IParseNode resultParseNode;

	private ParseStack[] parseStacks;

	private ParseStack[] tokenParseStacks;// NOPMD

	private boolean[] tokenRecognition;// NOPMD

	/**
	 * set error state at creation
	 * 
	 * @param inError
	 *            true if error during parsing
	 */
	public ParseResult(final boolean inError) {
		this.inError = inError;
	}

	/**
	 * set the resulting parse tree root node
	 * 
	 * @param resultParseNode
	 */
	public void setResultParseNode(final IParseNode resultParseNode) {
		this.resultParseNode = resultParseNode;
	}

	/**
	 * @param lastTerminalValueReaded
	 *            the last terminal value readed to set
	 */
	public void setLastTerminalValueReaded(final int lastTerminalValueReaded) {
		this.lastTerminalValueReaded = lastTerminalValueReaded;
	}

	/**
	 * set line number and column number in input text file for syntax error
	 * 
	 * @param lineNumber
	 *            line in file where is syntax error
	 * @param columnNumber
	 *            column in file where is syntax error
	 * @param fileName
	 *            file name where is syntax error
	 */
	public void setInTextPosition(final int lineNumber, final int columnNumber,
			final String fileName) {
		this.fileName = fileName;
		this.lineNumber = lineNumber;
		this.columnNumber = columnNumber;
	}

	/**
	 * set reduce error state
	 * 
	 * @param reduceError
	 *            true if reduce error while parsing
	 */
	public void setReduceError(final boolean reduceError) {
		this.reduceError = reduceError;
	}

	/**
	 * set syntax error state
	 * 
	 * @param parseError
	 *            true if syntax parsing error
	 */
	public void setParseError(final boolean parseError) {
		this.parseError = parseError;
	}

	/**
	 * set the reduce error informations
	 * 
	 * @param reduceException
	 *            the reduce error cause for reduce error
	 * @param reduceErrorLineNumber
	 *            line number in input text file for reduce error
	 * @param reduceErrorColumnNumber
	 *            column number in input text file for reduce error
	 * @param reduceErrorFileName
	 *            file name where is reduce error
	 */
	public void setReduceErrorInfo(final ParserException reduceException,
			final int reduceErrorLineNumber, final int reduceErrorColumnNumber,
			final String reduceErrorFileName) {
		this.reduceErrorFileName = reduceErrorFileName;
		this.reduceException = reduceException;
		this.reduceErrorLineNumber = reduceErrorLineNumber;
		this.reduceErrorColumnNumber = reduceErrorColumnNumber;
	}

	/**
	 * set the reduce error informations
	 * 
	 * @param parseNodeInErrorStatus
	 *            the error status for error cause of reduce error
	 * @param lastTerminalValueReaded
	 *            last terminal value readed
	 * @param reduceErrorFileName
	 *            file name where is reduce error
	 */
	public void setReduceErrorInfo(
			final IParseNodeInErrorStatus parseNodeInErrorStatus,
			final String reduceErrorFileName) {
		this.reduceErrorFileName = reduceErrorFileName;
		this.parseNodeInErrorStatus = parseNodeInErrorStatus;
	}

	/**
	 * 
	 * @return true if error during parsing
	 */
	public boolean isInError() {
		return inError;
	}

	/**
	 * 
	 * @return the resulting parse tree root node if no syntax error
	 */
	public IParseNode getResultParseNode() {
		return resultParseNode;
	}

	/**
	 * @return
	 * @see net.sf.parser4j.parser.entity.parsenode.IParseNode#getUniqData()
	 */
	public IParseNodeData getUniqData() {
		return resultParseNode.getUniqData();
	}

	/**
	 * 
	 * @return true if syntax parsing error
	 */
	public boolean isParseError() {
		return parseError;
	}

	/**
	 * 
	 * @return true if reduce error while parsing
	 */
	public boolean isReduceError() {
		return reduceError;
	}

	/**
	 * 
	 * @return file name where is syntax error
	 */
	public String getFileName() {
		return fileName;
	}

	/**
	 * 
	 * @return column number in input text file for syntax error
	 */
	public int getColumnNumber() {
		return columnNumber;
	}

	/**
	 * 
	 * @return line number in input text file for syntax error
	 */
	public int getLineNumber() {
		return lineNumber;
	}

	/**
	 * 
	 * @return expected non terminal not found in input text file for syntax
	 *         error
	 */
	public Set<String> getExpectedNonTerminal() {
		return expectedNonTerminal;
	}

	/**
	 * 
	 * @return expected terminal not found in input text file for syntax error
	 */
	public Set<TerminalCharRange> getExpectedTerminal() {
		return expectedTerminal;
	}

	/**
	 * 
	 * @return the reduce error cause for reduce error
	 */
	public ParserException getReduceException() {
		return reduceException;
	}

	/**
	 * @return the error status for error cause of reduce error
	 */
	public IParseNodeInErrorStatus getParseNodeInErrorStatus() {
		return parseNodeInErrorStatus;
	}

	/**
	 * 
	 * @return file name where is reduce error
	 */
	public String getReduceErrorFileName() {
		return reduceErrorFileName;
	}

	/**
	 * 
	 * @return line number in input text file for reduce error
	 */
	public int getReduceErrorLineNumber() {
		return reduceErrorLineNumber;
	}

	/**
	 * 
	 * @return column number in input text file for reduce error
	 */
	public int getReduceErrorColumnNumber() {
		return reduceErrorColumnNumber;
	}

	@SuppressWarnings("PMD")
	public void setParseStacks(final ParseStack[] parseStacks,
			final ParseStack[] tokenParseStacks,
			final boolean[] tokenRecognition) {
		this.parseStacks = parseStacks;
		this.tokenParseStacks = tokenParseStacks;
		this.tokenRecognition = tokenRecognition;
	}

	/**
	 * @return the parseStacks
	 */
	public ParseStack[] getParseStacks() {
		return parseStacks;// NOPMD
	}

	/**
	 * @return the tokenParseStacks
	 */
	public ParseStack[] getTokenParseStacks() {
		return tokenParseStacks;// NOPMD
	}

	/**
	 * @return the tokenRecognition
	 */
	public boolean[] getTokenRecognition() {
		return tokenRecognition;// NOPMD
	}

	@Override
	public String toString() {
		final StringBuilder stringBuilder = new StringBuilder();
		if (inError) {
			stringBuilder.append("has parsing error\n");
		} else {
			stringBuilder.append("does not have parsing error\n");
		}
		if (parseError) {
			stringBuilder.append("parse failed: file ");
			stringBuilder.append(fileName);
			stringBuilder.append(", line ");
			stringBuilder.append(lineNumber);
			stringBuilder.append(" column ");
			stringBuilder.append(columnNumber);
			stringBuilder.append(", last character readed ");
			if (lastTerminalValueReaded == -1) {
				stringBuilder.append("EOF");
			} else if (lastTerminalValueReaded < ' '
					|| lastTerminalValueReaded >= 0x7f) {
				stringBuilder.append(String.format("0X%2X",
						new Object[] { lastTerminalValueReaded }));
			} else {
				stringBuilder.append('"');
				stringBuilder.append((char) lastTerminalValueReaded);
				stringBuilder.append('"');
			}
			stringBuilder.append("\nexpected terminal:\n");
			for (TerminalCharRange terminalCharRange : expectedTerminal) {
				stringBuilder.append('\t');
				stringBuilder.append(terminalCharRange.toString());
				stringBuilder.append('\n');
			}
			stringBuilder.append("expected non terminal:\n");
			for (String nonTerminal : expectedNonTerminal) {
				stringBuilder.append('\t');
				stringBuilder.append(nonTerminal);
				stringBuilder.append('\n');
			}
		} else {
			stringBuilder.append("parse succeed\n");
		}
		if (reduceError) {
			stringBuilder.append("reduce error: file ");
			stringBuilder.append(reduceErrorFileName);
			stringBuilder.append('\n');
			if (reduceException != null) {
				stringBuilder.append("reduce exception: line ");
				stringBuilder.append(reduceErrorLineNumber);
				stringBuilder.append(", column ");
				stringBuilder.append(reduceErrorColumnNumber);
				final DebugUtil debugUtil = DebugUtil.getInstance();
				stringBuilder.append(debugUtil.stackTrace(reduceException));
			}
			if (parseNodeInErrorStatus != null
					&& parseNodeInErrorStatus.isInError()) {
				stringBuilder.append(parseNodeInErrorStatus.toString());
			}
		} else {
			stringBuilder.append("no reduce error\n");
		}

		if (parseStacks != null) {
			stringBuilder.append("syntax parse stacks\n");
			stringBuilder.append(parsingToStringUtil.parseStacksToString(
					parseStacks, tokenRecognition, false/*
														 * true/ addParseTree
														 */));
		}

		if (tokenParseStacks != null) {
			stringBuilder.append("lexical parse stacks\n");
			stringBuilder.append(parsingToStringUtil.parseStacksToString(
					tokenParseStacks, tokenRecognition, false/*
															 * true/
															 * addParseTree
															 */));
		}
		return stringBuilder.toString();
	}
}
