/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parser.entity.data;

import java.io.Serializable;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.parser4j.parser.service.ExStringBuilder;

/**
 * 
 * @author luc peuvrier
 * 
 */
public class ParserData implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 2384785354559121747L;

	private final int numberOfState;// NOPMD

	private final int numberOfNonterminal;// NOPMD

	/** for each state is defined a goto state by terminal for non token */
	private GotoByTerminalEntry[] gotoByTerminalTable;// NOPMD

	/** for each state is defined a goto state by terminal for token */
	private GotoByTerminalEntry[] tokenGotoByTerminalTable;// NOPMD

	/** for each state is defined a goto state by non terminal for non token */
	private GotoByNonTerminalEntry[] gotoByNonTerminalTable;// NOPMD

	/** for each state is defined a goto state by non terminal for token */
	private GotoByNonTerminalEntry[] tokenGotoByNonTerminalTable;// NOPMD

	/**
	 * for each state define non terminal matched and associated informations
	 * for non token
	 */
	private MatchEntrySet[] matchTable;// NOPMD

	/**
	 * for each state define non terminal matched and associated informations
	 * for token
	 */
	private MatchEntrySet[] tokenMatchTable;// NOPMD

	/** for each state define if it is token recognition */
	private boolean[] tokenRecognition;// NOPMD

	/** for each non terminal define match management class name */
	private String[] matchClassNameTable;// NOPMD

	/** root non terminal identifier */
	private int rootNonTerminalIdentifier;

	/** non terminal name by non terminal identifier */
	private Map<Integer, NonTerminal> nonTerminalByIdentifierMap; // NOPMD

	private StateRecognitionPointList[] byStateRecognitionPointList;// NOPMD

	/**
	 * number of state and number of non terminal are knew at construction
	 * 
	 * @param numberOfState
	 */
	public ParserData(final int numberOfState, final int numberOfNonterminal) {
		this.numberOfState = numberOfState;
		this.numberOfNonterminal = numberOfNonterminal;
		gotoByTerminalTable = new GotoByTerminalEntry[numberOfState];
		tokenGotoByTerminalTable = new GotoByTerminalEntry[numberOfState];
		gotoByNonTerminalTable = new GotoByNonTerminalEntry[numberOfState];
		tokenGotoByNonTerminalTable = new GotoByNonTerminalEntry[numberOfState];
		matchTable = new MatchEntrySet[numberOfState];
		tokenMatchTable = new MatchEntrySet[numberOfState];
		tokenRecognition = new boolean[numberOfState];
		byStateRecognitionPointList = new StateRecognitionPointList[numberOfState];
		for (int stateIdentifier = 0; stateIdentifier < numberOfState; stateIdentifier++) {
			gotoByTerminalTable[stateIdentifier] =
			/**/new GotoByTerminalEntry();// NOPMD
			tokenGotoByTerminalTable[stateIdentifier] =
			/**/new GotoByTerminalEntry();// NOPMD
			gotoByNonTerminalTable[stateIdentifier] =
			/**/new GotoByNonTerminalEntry();// NOPMD
			tokenGotoByNonTerminalTable[stateIdentifier] =
			/**/new GotoByNonTerminalEntry();// NOPMD
			matchTable[stateIdentifier] = new MatchEntrySet(false);// NOPMD
			tokenMatchTable[stateIdentifier] = new MatchEntrySet(true);// NOPMD
			tokenRecognition[stateIdentifier] = false;
		}
		matchClassNameTable = new String[numberOfNonterminal];
	}

	public int getNumberOfNonterminal() {
		return numberOfNonterminal;
	}

	public int getNumberOfState() {
		return numberOfState;
	}

	/**
	 * add a goto from state to state by terminal
	 * 
	 * @param token
	 * @param fromStateIdentifier
	 * @param terminalCharRange
	 * @param toStateIdentifier
	 */
	public void addGotoByTerminal(final boolean token,
			final int fromStateIdentifier,
			final TerminalCharRange terminalCharRange,
			final int toStateIdentifier) {
		if (token) {
			tokenGotoByTerminalTable[fromStateIdentifier].addGoto(
					terminalCharRange, toStateIdentifier);
		} else {
			gotoByTerminalTable[fromStateIdentifier].addGoto(terminalCharRange,
					toStateIdentifier);
		}
	}

	// /**
	// * add a goto from state to state by non terminal
	// *
	// * @param token
	// * @param fromStateIdentifier
	// * @param nonTerminalIdentifier
	// * @param nonTerminalAsWhiteSpace
	// * @param toStateIdentifier
	// */
	// public void addGotoByNonTerminal(final boolean token,
	// final int fromStateIdentifier, final int nonTerminalIdentifier,
	// final boolean nonTerminalAsWhiteSpace, final int toStateIdentifier) {
	// if (token) {
	// tokenGotoByNonTerminalTable[fromStateIdentifier].addGoto(
	// nonTerminalIdentifier, nonTerminalAsWhiteSpace,
	// toStateIdentifier);
	// } else {
	// gotoByNonTerminalTable[fromStateIdentifier].addGoto(
	// nonTerminalIdentifier, nonTerminalAsWhiteSpace,
	// toStateIdentifier);
	// }
	// }

	/**
	 * add a goto from state to state by non terminal
	 * 
	 * @param token
	 * @param fromStateIdentifier
	 * @param nonTerminalIdentifier
	 * @param toStateAndWhiteSpaceFlag
	 */
	public void addGotoByNonTerminal(final boolean token,
			final int fromStateIdentifier, final int nonTerminalIdentifier,
			final ToStateAndWhiteSpaceFlag toStateAndWhiteSpaceFlag) {
		if (token) {
			tokenGotoByNonTerminalTable[fromStateIdentifier].addGoto(
					nonTerminalIdentifier, toStateAndWhiteSpaceFlag);
		} else {
			gotoByNonTerminalTable[fromStateIdentifier].addGoto(
					nonTerminalIdentifier, toStateAndWhiteSpaceFlag);
		}
	}

	/**
	 * add non terminal match to a state for non token
	 * 
	 * @param fromStateIdentifier
	 * @param nonTerminalIdentifier
	 * @param length
	 * @param nodeType
	 * @param token
	 */
	public void addMatch(final int fromStateIdentifier,
			final int nonTerminalIdentifier, final int length,
			final int nodeType, final boolean token) {
		matchTable[fromStateIdentifier].addMatch(nonTerminalIdentifier, length,
				nodeType, token);
	}

	/**
	 * add non terminal match to a state for token
	 * 
	 * @param fromStateIdentifier
	 * @param nonTerminalIdentifier
	 * @param length
	 * @param nodeType
	 * @param token
	 */
	public void addTokenMatch(final int fromStateIdentifier,
			final int nonTerminalIdentifier, final int length,
			final int nodeType, final boolean token) {
		tokenMatchTable[fromStateIdentifier].addMatch(nonTerminalIdentifier,
				length, nodeType, token);
	}

	/**
	 * 
	 * @param stateIdentifier
	 */
	public void setTokenRecognition(final int stateIdentifier) {
		tokenRecognition[stateIdentifier] = true;
	}

	public boolean[] getTokenRecognition() {
		return tokenRecognition;// NOPMD
	}

	/**
	 * 
	 * @param nonTerminalIdentifier
	 * @param matchClassName
	 *            can be null
	 */
	public void setNonTerminalMatchClassName(final int nonTerminalIdentifier,
			final String matchClassName) {
		matchClassNameTable[nonTerminalIdentifier - 0x10000] = matchClassName;
	}

	public boolean match(final int stateIdentifier) {
		return matchTable[stateIdentifier].match();
	}

	public GotoByTerminalEntry[] getGotoByTerminalTable() {
		return gotoByTerminalTable;// NOPMD
	}

	public GotoByTerminalEntry[] getTokenGotoByTerminalTable() {
		return tokenGotoByTerminalTable;// NOPMD
	}

	public GotoByNonTerminalEntry[] getGotoByNonTerminalTable() {
		return gotoByNonTerminalTable;// NOPMD
	}

	public GotoByNonTerminalEntry[] getTokenGotoByNonTerminalTable() {
		return tokenGotoByNonTerminalTable;// NOPMD
	}

	/**
	 * 
	 * @return match management class name for non terminal
	 */
	public String[] getMatchClassNameTable() {
		return matchClassNameTable;// NOPMD
	}

	public MatchEntrySet[] getMatchTable() {
		return matchTable;// NOPMD
	}

	public MatchEntrySet[] getTokenMatchTable() {
		return tokenMatchTable;// NOPMD
	}

	public int getRootNonTerminalIdentifier() {
		return rootNonTerminalIdentifier;
	}

	public void setRootNonTerminalIdentifier(final int rootNonTerminalIdentifier) {
		this.rootNonTerminalIdentifier = rootNonTerminalIdentifier;
	}

	public void setNonTerminalByIdentifierMap(
			final Map<Integer, NonTerminal> nonTerminalByIdentifierMap) {
		this.nonTerminalByIdentifierMap = nonTerminalByIdentifierMap;
	}

	public Map<Integer, NonTerminal> getNonTerminalByIdentifierMap() {
		return nonTerminalByIdentifierMap;
	}

	public void setRecognitionPoint(final int stateIdentifier,
			final List<ParserDataRecognitionPoint> list) {
		byStateRecognitionPointList[stateIdentifier] = new StateRecognitionPointList(
				list);
	}

	public List<ParserDataRecognitionPoint> getRecognitionPoint(
			final int stateIdentifier) {
		return byStateRecognitionPointList[stateIdentifier].getList();
	}

	/**
	 * @return the byStateRecognitionPointList
	 */
	public StateRecognitionPointList[] getByStateRecognitionPointList() {
		return byStateRecognitionPointList;// NOPMD
	}

	@Override
	public String toString() {
		final ExStringBuilder builder = new ExStringBuilder();
		for (int stateIdentifier = 0; stateIdentifier < numberOfState; stateIdentifier++) {
			builder.append("\nstate #");
			builder.append(stateIdentifier);

			builder.append("\n\ngoto by terminal:\n");
			final GotoByTerminalEntry gotoByTerminalEntry =
			/**/gotoByTerminalTable[stateIdentifier];
			builder.appendTabulate(1, gotoByTerminalEntry.toString());

			builder.append("\ngoto by non terminal:\n");
			final Map<Integer, ToStateAndWhiteSpaceFlag> gotoMap = gotoByNonTerminalTable[stateIdentifier]
					.getGotoMap();
			for (Map.Entry<Integer, ToStateAndWhiteSpaceFlag> entry : gotoMap
					.entrySet()) {
				builder.append("\tby ");
				final NonTerminal nonTerminal = nonTerminalByIdentifierMap
						.get(entry.getKey());
				builder.append(nonTerminal.getName());
				if (nonTerminal.isToken()) {
					builder.append(" (token)");
				}
				final ToStateAndWhiteSpaceFlag value = entry.getValue();
				if (value.isWhiteSpace()) {
					builder.append(" (as white space)");
				}
				builder.append(" goto ");
				builder.append(value.getToState());
				builder.append('\n');
			}

			builder.append("\nmatch:\n");
			final Set<MatchEntry> set = matchTable[stateIdentifier]
					.getNonTerminalMatchSet();
			for (MatchEntry matchEntry : set) {
				final int identifier = matchEntry
						.getMatchedNonTerminalIdentifier();
				builder.append('\t');
				final NonTerminal nonTerminal = nonTerminalByIdentifierMap
						.get(identifier);
				builder.append(nonTerminal.getName());
				if (nonTerminal.isToken()) {
					builder.append(" (token)");
				}
				builder.append('\n');
			}

			builder.append('\n');
			final StateRecognitionPointList stateRecognitionPointList =
			/**/byStateRecognitionPointList[stateIdentifier];
			final List<ParserDataRecognitionPoint> list =
			/**/stateRecognitionPointList.getList();
			for (ParserDataRecognitionPoint recognitionPoint : list) {
				final int identifier = recognitionPoint
						.getDefinedNonTerminalIdentifier();
				String name = nonTerminalByIdentifierMap.get(identifier)
						.getName();
				final boolean token = recognitionPoint.isToken();
				if (token) {
					builder.append('[');
				}
				builder.append(name);
				if (token) {
					builder.append(']');
				}
				builder.append(':');
				final int recognitionPointIndex = recognitionPoint
						.getRecognitionPointIndex();
				final int[] rightPart = recognitionPoint.getRightPart();
				final boolean concat = recognitionPoint.isConcat();
				int index = 0;
				for (int rightIdentifier : rightPart) {
					builder.append(' ');
					if (!concat || index == recognitionPointIndex) {
						builder.append('.');
					}
					if (rightIdentifier == -1) {
						builder.append("<terminal>");
					} else {
						name = nonTerminalByIdentifierMap.get(rightIdentifier)
								.getName();
						builder.append(name);
						if (!concat && index != rightPart.length - 1) {
							builder.append(" |");
						}
					}
					index++;
				}
				if (index == recognitionPointIndex) {
					builder.append(" .");
				}
				builder.append('\n');
			}
		}
		return builder.toString();
	}
}
