/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parser.entity.parsenode;

import java.util.Arrays;

import net.sf.parser4j.parser.entity.EnumNodeType;
import net.sf.parser4j.parser.entity.parsenode.data.IParseNodeData;
import net.sf.parser4j.parser.entity.parsenode.data.NullParseNodeData;
import net.sf.parser4j.parser.service.ParserException;
import net.sf.parser4j.parser.service.parsenode.IParseNodeVisitor;
import net.sf.parser4j.parser.service.parsenode.ParsingToStringUtil;
import net.sf.parser4j.parser.service.parsenode.TooMuchAlternativeParserException;

/**
 * common to all non terminal parse node
 * 
 * @author luc peuvrier
 * 
 */
public abstract class AbstractNonTerminalParseNode extends AbstractParseNode {

	private final static ParsingToStringUtil parsingToStringUtil =
	/**/ParsingToStringUtil.getInstance();

	/** number of alternative */
	private int numberOfAlternative;// NOPMD

	/** sons of this parse node by alternative */
	protected IParseNode[][] sonParseNodesByAlternative;// NOPMD

	/** data associated to this parse node by alternative */
	private IParseNodeData[] dataByAlternative;// NOPMD

	public AbstractNonTerminalParseNode(final EnumNodeType parseNodeType) {
		super(parseNodeType);
	}

	/**
	 * construct setting this parse node type and sons
	 * 
	 * @param parseNodeType
	 *            this parse node type
	 * @param parseNodes
	 *            this parse node sons parse node
	 */
	public AbstractNonTerminalParseNode(final EnumNodeType parseNodeType,
			final IParseNode[] parseNodes) {
		super(parseNodeType);
		sonParseNodesByAlternative = new IParseNode[1][];
		sonParseNodesByAlternative[0] = parseNodes;
		dataByAlternative = new IParseNodeData[1];
		dataByAlternative[0] = NullParseNodeData.getInstance();
		numberOfAlternative = 1;
	}

	public int getNumberOfAlternative() {
		return numberOfAlternative;
	}

	public IParseNode[][] getSonParseNodesByAlternative() {
		return sonParseNodesByAlternative;// NOPMD
	}

	public boolean setAlternativeParseNode(
			final IParseNode alternativeParseNode,
			final int maxNumberOfAlternative) throws ParserException,
			TooMuchAlternativeParserException {
		final boolean added;
		if (alternativeParseNode == this) {
			added = false;
		} else {
			added = true;
			final int numberOfNewAlternative = alternativeParseNode
					.getNumberOfAlternative();
			final IParseNode[][] newAlternativeSons = alternativeParseNode
					.getSonParseNodesByAlternative();
			final IParseNodeData[] newAlternativeData = alternativeParseNode
					.getDataByAlternative();
			final int newNumberOfAlternative = numberOfAlternative
					+ numberOfNewAlternative;

			if (_log.isDebugEnabled()) {
				_log.debug("for "
						+ ParsingToStringUtil.getInstance().parseNodeToString(
								alternativeParseNode)
						+ "\nnumber of alternative: old=" + numberOfAlternative
						+ ", new=" + newNumberOfAlternative
						+ ", new alternative=" + numberOfNewAlternative);
			}

			if (newNumberOfAlternative <= maxNumberOfAlternative) {
				sonParseNodesByAlternative = Arrays.copyOf(
						sonParseNodesByAlternative, newNumberOfAlternative);
				dataByAlternative = Arrays.copyOf(dataByAlternative,
						newNumberOfAlternative);
				int destIndex = numberOfAlternative;
				for (int srcIndex = 0; srcIndex < numberOfNewAlternative; srcIndex++) {
					sonParseNodesByAlternative[destIndex] = newAlternativeSons[srcIndex];
					dataByAlternative[destIndex] = newAlternativeData[srcIndex];
					destIndex++;
				}
				numberOfAlternative = newNumberOfAlternative;
			} else {
				final String thisParseNode = parsingToStringUtil
						.parseTreeToString(this);
				final String altParseNode = parsingToStringUtil
						.parseTreeToString(alternativeParseNode);
				final String message = "to add to\n" + thisParseNode
						+ "\nalternative\n" + altParseNode;
				throw new TooMuchAlternativeParserException(message);
			}
		}
		return added;
	}

	public IParseNodeData getUniqData() {
		if (numberOfAlternative != 1) {
			throw new IllegalStateException("have multiple alternative");
		}
		return dataByAlternative[0];
	}

	public IParseNodeData[] getDataByAlternative() {
		return dataByAlternative;// NOPMD
	}

	public void setData(final IParseNodeData data) {
		if (data == null) {
			throw new IllegalArgumentException("null argument");
		}
		if (numberOfAlternative != 1) {
			throw new IllegalStateException("have multiple alternative");
		}
		this.dataByAlternative[0] = data;
	}

	protected void setSonParseNodes(final IParseNode[] array) {
		sonParseNodesByAlternative[0] = array;
	}

	public void accept(final IParseNodeVisitor visitor) {
		if (visitor.beginVisit(this)) {
			if (visitor.visitFirstAlternativeOnly()) {
				final IParseNode[] sonParseNodes = sonParseNodesByAlternative[0];
				visitor.beginAlternativeVisit(this, 0);
				for (IParseNode parseNode : sonParseNodes) {
					parseNode.accept(visitor);
				}
				visitor.endAlternativeVisit(this, 0);
			} else {
				int alternativeCount = 0;
				for (IParseNode[] sonParseNodes : sonParseNodesByAlternative) {
					visitor.beginAlternativeVisit(this, alternativeCount);
					for (IParseNode parseNode : sonParseNodes) {
						parseNode.accept(visitor);
					}
					visitor.endAlternativeVisit(this, alternativeCount);
					alternativeCount++;
				}
			}
			visitor.endVisit(this);
		}
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		for (int index = 0; index < numberOfAlternative; index++) {
			result = prime * result
					+ Arrays.hashCode(sonParseNodesByAlternative[index]);
		}
		return result;
	}

	@Override
	@SuppressWarnings("PMD")
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (!super.equals(obj))
			return false;
		if (!(obj instanceof AbstractNonTerminalParseNode))
			return false;
		AbstractNonTerminalParseNode other = (AbstractNonTerminalParseNode) obj;
		if (numberOfAlternative != other.numberOfAlternative)
			return false;
		for (int index = 0; index < numberOfAlternative; index++) {
			if (!Arrays.equals(sonParseNodesByAlternative[index],
					other.sonParseNodesByAlternative[index]))
				return false;
		}
		return true;
	}
}
