/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parser.entity.parsenode;

import net.sf.parser4j.parser.entity.EnumNodeType;
import net.sf.parser4j.parser.entity.parsenode.status.IParseNodeInErrorStatus;
import net.sf.parser4j.parser.entity.parsenode.status.IParseNodeStatus;
import net.sf.parser4j.parser.entity.parsenode.status.impl.NoErrorParseNodeStatus;
import net.sf.parser4j.parser.service.match.IMatchMgr;
import net.sf.parser4j.parser.service.parsenode.ParsingToStringUtil;

import org.apache.log4j.Logger;

/**
 * common to all parse node
 * 
 * @author luc peuvrier
 * 
 */
public abstract class AbstractParseNode implements IParseNode {

	protected transient Logger _log = Logger.getLogger(getClass());// NOPMD

	/** use to give an identifier number to parse node */
	private static int nextParseNodeIdentifier = 0;

	/** this parse node identifier number */
	private final int myIdentifier;// NOPMD

	private final EnumNodeType parseNodeType;// NOPMD

	/** the match manager that update this parse node, null if none */
	private IMatchMgr matchMgr;

	/** grammar symbol identifier of this parse node */
	private int grammarSymbolIdentifier;

	protected int beginLineNumber;// NOPMD

	protected int beginColumnNumber;// NOPMD

	protected int endLineNumber;// NOPMD

	protected int endColumnNumber;// NOPMD

	private boolean whiteSpace;

	/** the parse node status: ok by default at construction */
	private IParseNodeStatus status = NoErrorParseNodeStatus.getInstance();// NOPMD

	protected AbstractParseNode(final EnumNodeType parseNodeType) {
		super();
		this.parseNodeType = parseNodeType;
		myIdentifier = nextParseNodeIdentifier++;
	}

	public int getMyIdentifier() {
		return myIdentifier;
	}

	public EnumNodeType getParseNodeType() {
		return parseNodeType;
	}

	public int getGrammarSymbolIdentifier() {
		return grammarSymbolIdentifier;
	}

	@Override
	public boolean isWhiteSpace() {
		return whiteSpace;
	}

	@Override
	public void setWhiteSpace(final boolean whiteSpace) {
		this.whiteSpace = whiteSpace;
	}

	public IMatchMgr getMatchMgr() {
		return matchMgr;
	}

	public void setMatchMgr(final IMatchMgr matchMgr) {
		this.matchMgr = matchMgr;
	}

	public boolean isNonTerminal() {
		return parseNodeType.isNonTerminal();
	}

	public boolean isTerminal() {
		return parseNodeType.isTerminal();
	}

	public void setGrammarSymbolIdentifier(final int grammarSymbolIdentifier) {
		this.grammarSymbolIdentifier = grammarSymbolIdentifier;
	}

	@Override
	public void setLineAndColumnNumber(final int beginLineNumber,
			final int beginColumnNumber, final int endLineNumber,
			final int endColumnNumber) {
		this.beginLineNumber = beginLineNumber;
		this.beginColumnNumber = beginColumnNumber;
		this.endLineNumber = endLineNumber;
		this.endColumnNumber = endColumnNumber;
	}

	public int getBeginLineNumber() {
		return beginLineNumber;
	}

	public int getEndLineNumber() {
		return endLineNumber;
	}

	public int getBeginColumnNumber() {
		return beginColumnNumber;
	}

	public int getEndColumnNumber() {
		return endColumnNumber;
	}

	/**
	 * @return
	 * @see net.sf.parser4j.parser.entity.parsenode.status.IParseNodeStatus#isInError()
	 */
	public boolean isInError() {
		return status.isInError();
	}

	public IParseNodeInErrorStatus getErrorStatus() {
		return (IParseNodeInErrorStatus) status;
	}

	public void setErrorStatus(final IParseNodeInErrorStatus status) {
		this.status = status;
	}

	@Override
	public String toString() {
		throw new UnsupportedOperationException(getClass() + ", use "
				+ ParsingToStringUtil.class);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + grammarSymbolIdentifier;
		result = prime * result
				+ ((parseNodeType == null) ? 0 : parseNodeType.hashCode());
		return result;
	}

	@Override
	@SuppressWarnings("PMD")
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		AbstractParseNode other = (AbstractParseNode) obj;
		if (grammarSymbolIdentifier != other.grammarSymbolIdentifier)
			return false;
		if (parseNodeType == null) {
			if (other.parseNodeType != null)
				return false;
		} else if (!parseNodeType.equals(other.parseNodeType))
			return false;
		return true;
	}
}
