/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parser.entity.parsenode;

import net.sf.parser4j.parser.entity.EnumNodeType;
import net.sf.parser4j.parser.entity.parsenode.data.IParseNodeData;
import net.sf.parser4j.parser.entity.parsenode.status.IParseNodeInErrorStatus;
import net.sf.parser4j.parser.service.ParserException;
import net.sf.parser4j.parser.service.match.IMatchMgr;
import net.sf.parser4j.parser.service.parsenode.IParseNodeVisitor;
import net.sf.parser4j.parser.service.parsenode.TooMuchAlternativeParserException;

/**
 * all parse node interface
 * 
 * @author luc peuvrier
 * 
 */
public interface IParseNode {

	/**
	 * 
	 * @return this parse node identifier
	 */
	int getMyIdentifier();

	/**
	 * 
	 * @return the parse node type
	 */
	EnumNodeType getParseNodeType();

	/**
	 * set grammar symbol identifier of this parse node
	 * 
	 * @param grammarSymbolIdentifier
	 */
	void setGrammarSymbolIdentifier(int grammarSymbolIdentifier);

	/**
	 * set begin/end line/column number
	 * 
	 * @param beginLineNumber
	 * @param beginColumnNumber
	 * @param endLineNumber
	 * @param endColumnNumber
	 */
	void setLineAndColumnNumber(int beginLineNumber, int beginColumnNumber,
			int endLineNumber, int endColumnNumber);

	/**
	 * 
	 * @return begin line number
	 */
	int getBeginLineNumber();

	/**
	 * 
	 * @return end line number
	 */
	int getEndLineNumber();

	/**
	 * 
	 * @return begin column number
	 */
	int getBeginColumnNumber();

	/**
	 * 
	 * @return end column number
	 */
	int getEndColumnNumber();

	/**
	 * 
	 * @return number of alternative of this parse node
	 */
	int getNumberOfAlternative();

	/**
	 * 
	 * @return non terminal identifier of this parse node
	 */
	int getGrammarSymbolIdentifier();

	/**
	 * 
	 * @return true if parse node is in error
	 */
	boolean isInError();

	/**
	 * 
	 * @return the error status
	 */
	IParseNodeInErrorStatus getErrorStatus();

	/**
	 * set error status
	 * 
	 * @param status
	 *            the error status
	 */
	void setErrorStatus(final IParseNodeInErrorStatus status);

	/**
	 * 
	 * @return true if this parse node is a non terminal
	 */
	boolean isNonTerminal();

	/**
	 * 
	 * @return true if this parse node is a terminal
	 */
	boolean isTerminal();

	/**
	 * 
	 * @param whiteSpace
	 *            true if the matched non terminal of this parse node is a white
	 *            space
	 */
	void setWhiteSpace(boolean whiteSpace);

	/**
	 * 
	 * @return true if the matched non terminal of this parse node is a white
	 *         space
	 */
	boolean isWhiteSpace();

	/**
	 * set alternative parse node, special case for terminal ( see
	 * implementation )
	 * 
	 * @param alternativeParseNode
	 * @param maxNumberOfAlternative
	 * @return true if alternative added, false if not added when alternative
	 *         parse node is same instance than this
	 * @throws ParserException
	 * @throws TooMuchAlternativeParserException
	 */
	boolean setAlternativeParseNode(IParseNode alternativeParseNode,
			int maxNumberOfAlternative) throws ParserException,
			TooMuchAlternativeParserException;

	/**
	 * @param data
	 *            data associated to this parse node first alternative
	 */
	void setData(IParseNodeData data);

	/**
	 * 
	 * @return uniq data associated to this parse node
	 */
	IParseNodeData getUniqData();

	/**
	 * 
	 * @return data associated to this parse node by alternative
	 */
	IParseNodeData[] getDataByAlternative();

	/**
	 * 
	 * @return sons parse nodes by alternative
	 */
	IParseNode[][] getSonParseNodesByAlternative();

	/**
	 * 
	 * @param matchMgr
	 *            the match manager that update this parse node, null if none
	 */
	void setMatchMgr(IMatchMgr matchMgr);

	/**
	 * 
	 * @return the match manager that update this parse node, null if none
	 */
	IMatchMgr getMatchMgr();

	/**
	 * to visit this parse node
	 * 
	 * @param visitor
	 *            the parse node visitor
	 */
	void accept(IParseNodeVisitor visitor);
}
