/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parser.entity.parsestate;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.log4j.Logger;

import net.sf.parser4j.parser.entity.parsenode.IParseNode;
import net.sf.parser4j.parser.service.HasAmbiguityParserException;
import net.sf.parser4j.parser.service.ParserException;
import net.sf.parser4j.parser.service.parsenode.ParsingToStringUtil;
import net.sf.parser4j.parser.service.parsenode.TooMuchAlternativeParserException;

/**
 * 
 * @author luc peuvrier
 * 
 */
public abstract class AbstractParseState implements IParseState {

	protected final Logger _log = Logger.getLogger(getClass());// NOPMD

	protected static final ParsingToStringUtil parsingToStringUtil =
	/**/ParsingToStringUtil.getInstance();

	/** true if accept ambiguity */
	protected final boolean acceptAmbiguity;// NOPMD

	/** set of parse stack. use map to obtains existing equals element */
	protected final Map<ParseStack, ParseStack> parseStackSet = // NOPMD
	/**/new TreeMap<ParseStack, ParseStack>();

	private boolean alternativeForAmbiguity = false;// NOPMD

	/** for each state define if it is token recognition */
	protected final boolean[] tokenRecognition;// NOPMD

	protected int maxNumberOfAlternative;// NOPMD

	public AbstractParseState(final boolean acceptAmbiguity,
			final int maxNumberOfAlternative, final boolean[] tokenRecognition) {// NOPMD
		super();
		this.acceptAmbiguity = acceptAmbiguity;
		this.maxNumberOfAlternative = maxNumberOfAlternative;
		this.tokenRecognition = tokenRecognition;
	}

	public boolean add(final ParseStack parseStack) throws ParserException,
			HasAmbiguityParserException {
		final boolean added;
		final ParseStack existingParseStack = parseStackSet.get(parseStack);
		if (existingParseStack == null) {
			parseStackSet.put(parseStack, parseStack);
			added = true;
			if (_log.isDebugEnabled()) {
				final String toString = parsingToStringUtil.parseStackToString(
						parseStack, tokenRecognition, false);
				_log.debug("added\n" + toString);
			}
		} else if (existingParseStack == parseStack) {// NOPMD
			throw new ParserException("add same");
		} else {
			if (_log.isDebugEnabled()) {
				final String toString = parsingToStringUtil.parseStackToString(
						existingParseStack, tokenRecognition, false);
				_log.debug("set alternative for\n" + toString);
			}
			added = false;
			alternativeForAmbiguity = true;
			final IParseNode[] existing = existingParseStack.getParseNode();
			final IParseNode[] alternative = parseStack.getParseNode();
			final boolean[] alternativeAddedByStackElt = new boolean[existing.length];
			int index = 0;
			boolean hasAmbiguity = false;
			for (IParseNode existingParseNode : existing) {
				try {
					final boolean alternativeAdded = existingParseNode
							.setAlternativeParseNode(alternative[index],
									maxNumberOfAlternative);
					alternativeAddedByStackElt[index] = alternativeAdded;
					if (alternativeAdded) {
						hasAmbiguity = true;
					}
				} catch (TooMuchAlternativeParserException exception) {
					final String existingToString = parsingToStringUtil
							.parseStackToString(existingParseStack,
									tokenRecognition, false);
					final String message = "for element #" + index + " of\n"
							+ existingToString;
					throw new ParserException(message, exception);
				}
				index++;
			}
			if (hasAmbiguity && !acceptAmbiguity) {
				final List<IParseNode> parseNodeList = new LinkedList<IParseNode>();
				final StringBuilder builder = new StringBuilder();
				builder.append("alternative on stack below entry:");
				index = 0;
				for (boolean a : alternativeAddedByStackElt) {
					if (a) {
						builder.append(" #");
						builder.append(index);
						parseNodeList.add(existing[index]);
					}
					index++;
				}
				builder.append('\n');
				builder.append(parsingToStringUtil.parseStackToString(
						parseStack, tokenRecognition, false));
				throw new HasAmbiguityParserException(builder.toString(),
						parseNodeList);
			}
		}
		return added;
	}

	public void clear() {
		parseStackSet.clear();
	}

	public boolean remove(final ParseStack parseStack) {
		return parseStackSet.remove(parseStack) != null;
	}

	public IParseNode getMatchParseNode() {
		if (parseStackSet.size() != 1) {
			throw new IllegalStateException("bad size " + parseStackSet.size()
					+ ", 1 parse stack expected");
		}
		return parseStackSet.keySet().iterator().next().getMatchAllParseNode();
	}

	@Override
	public Iterator<ParseStack> iterator() {
		return parseStackSet.keySet().iterator();
	}

	public boolean isEmpty() {
		return parseStackSet.isEmpty();
	}

	public int size() {
		return parseStackSet.size();
	}

	public ParseStack[] toArray() {
		final ParseStack[] parseStacks = new ParseStack[parseStackSet.size()];
		parseStackSet.keySet().toArray(parseStacks);
		return parseStacks;
	}

	public boolean hasAlternativeForAmbiguity() {
		return alternativeForAmbiguity;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see net.sf.parser4j.parser.entity.parsestate.IParseState#copy()
	 */
	@Override
	public IParseState copy() {
		return new ParseState(this);
	}
}
