/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parser.entity.parsestate;

import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;

import net.sf.parser4j.parser.entity.parsenode.IParseNode;
import net.sf.parser4j.parser.entity.parsenode.status.IParseNodeInErrorStatus;

/**
 * parse stack<br>
 * 
 * @author luc peuvrier
 * 
 */
public class ParseStack implements Comparable<ParseStack> {

	private Deque<ParseStackElement> parseStateNodeStack;// NOPMD

	private int stateNumber;// NOPMD

	/** true if all input matched */
	private boolean matchAll;

	/** match all parse node if {@link #matchAll} is true, else null */
	private IParseNode matchAllParseNode; // NOPMD

	/**
	 * error status from son node, null if none : computed at call of
	 * {@link #getMatchedParseNode(int)}
	 */
	private transient IParseNodeInErrorStatus parseNodeInErrorStatus;

	/**
	 * computed at call of {@link #getMatchedParseNode(int)}
	 */
	private transient int[] notWhiteIndexes;

	public ParseStack() {
		super();
		parseStateNodeStack = new LinkedList<ParseStackElement>();
		stateNumber = 0;
	}

	public ParseStack(final int stateNumber) {
		super();
		parseStateNodeStack = new LinkedList<ParseStackElement>();
		this.stateNumber = stateNumber;
	}

	public ParseStack(final Deque<ParseStackElement> parseStateNodeStack,
			final int stateNumber) {
		super();
		this.parseStateNodeStack = parseStateNodeStack;
		this.stateNumber = stateNumber;
	}

	public int getStateNumber() {
		return stateNumber;
	}

	public Deque<ParseStackElement> getParseStateNodeStack() {
		return parseStateNodeStack;
	}

	public void shift(final IParseNode parseNode, final int toState) {
		final ParseStackElement parseStackElement = new ParseStackElement(
				parseNode, stateNumber);
		parseStateNodeStack.addLast(parseStackElement);
		stateNumber = toState;
	}

	public IParseNode[] getMatchedParseNode(final int matchedLength) {
		final Iterator<ParseStackElement> iterator = parseStateNodeStack
				.descendingIterator();
		final Deque<IParseNode> deque = new LinkedList<IParseNode>();
		final Deque<Integer> notWhiteIndexList = new LinkedList<Integer>();
		for (int index = 0; index < matchedLength; index++) {
			final IParseNode parseNode = iterator.next().getParseNode();
			deque.addFirst(parseNode);
			if (parseNode.isInError()) {
				parseNodeInErrorStatus = parseNode.getErrorStatus();
			}
			if (!parseNode.isWhiteSpace()) {
				notWhiteIndexList.addFirst(matchedLength - index - 1);
			}
		}
		final IParseNode[] parseNodes = new IParseNode[matchedLength];
		deque.toArray(parseNodes);
		notWhiteIndexes = new int[notWhiteIndexList.size()];
		int index = 0;
		for (int notWhiteIndex : notWhiteIndexList) {
			notWhiteIndexes[index++] = notWhiteIndex;
		}
		return parseNodes;
	}

	/**
	 * 
	 * @return error status from son node, null if none
	 */
	public IParseNodeInErrorStatus getParseNodeInErrorStatus() {
		return parseNodeInErrorStatus;
	}

	public int[] getNotWhiteIndexes() {
		return notWhiteIndexes;// NOPMD
	}

	public IParseNode[] getParseNode() {
		final IParseNode[] parseNodes = new IParseNode[parseStateNodeStack
				.size()];
		int index = 0;
		for (ParseStackElement parseStackElement : parseStateNodeStack) {
			parseNodes[index] = parseStackElement.getParseNode();
			index++;
		}
		return parseNodes;
	}

	public ParseStack copy() {
		final Deque<ParseStackElement> parseStackCopy =
		/**/new LinkedList<ParseStackElement>(parseStateNodeStack);
		return new ParseStack(parseStackCopy, stateNumber);
	}

	public void reduce(final int matchedLength) {
		if (matchedLength != 0) {
			for (int count = 0; count < matchedLength - 1; count++) {
				parseStateNodeStack.pollLast();
			}
			stateNumber = parseStateNodeStack.pollLast().getStateNumber();
		}
	}

	/**
	 * set this stack match all the input
	 * 
	 * @param matchAllParseNode
	 *            match all parse node
	 * 
	 */
	public void setMatchAll(final IParseNode matchAllParseNode) {
		matchAll = true;
		this.matchAllParseNode = matchAllParseNode;
	}

	/**
	 * @return true if all the input matched
	 */
	public boolean isMatchAll() {
		return matchAll;
	}

	/**
	 * 
	 * @return match all parse node if {@link #matchAll} is true, else null
	 */
	public IParseNode getMatchAllParseNode() {
		return matchAllParseNode;
	}

	public int compareTo(final ParseStack parseStack) {

		int compareTo = parseStateNodeStack.size()
				- parseStack.parseStateNodeStack.size();
		if (compareTo == 0) {
			final Iterator<ParseStackElement> iterator = parseStateNodeStack
					.iterator();
			final Iterator<ParseStackElement> iteratorOnOther =
			/**/parseStack.parseStateNodeStack.iterator();
			while (compareTo == 0 && iterator.hasNext()
					&& iteratorOnOther.hasNext()) {
				final ParseStackElement next = iterator.next();
				final ParseStackElement otherNext = iteratorOnOther.next();
				compareTo = next.getStateNumber() - otherNext.getStateNumber();
				if (compareTo == 0) {
					final IParseNode nextParseNode = next.getParseNode();
					final IParseNode otherNextParseNode = otherNext
							.getParseNode();
					compareTo = nextParseNode.getGrammarSymbolIdentifier()
							- otherNextParseNode.getGrammarSymbolIdentifier();
				}
			}
			if (compareTo == 0) {
				compareTo = getStateNumber() - parseStack.getStateNumber();
			}
		}
		return compareTo;
	}

	@Override
	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		for (ParseStackElement parseStackElement : parseStateNodeStack) {
			result = PRIME * result + parseStackElement.getStateNumber();
			result = PRIME * result
					+ parseStackElement.getParseNode().hashCode();
		}
		result = PRIME * result + getStateNumber();
		return result;
	}

	@Override
	@SuppressWarnings("PMD")
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final ParseStack other = (ParseStack) obj;
		if (compareTo(other) != 0)
			return false;
		return true;
	}
}
