/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parser.service;

import net.sf.parser4j.parser.entity.ParseSession;
import net.sf.parser4j.parser.entity.parsenode.IParseNode;
import net.sf.parser4j.parser.entity.parsenode.data.IParseNodeData;
import net.sf.parser4j.parser.service.match.IMatchMgr;
import net.sf.parser4j.parser.service.parsenode.ParsingToStringUtil;

import org.apache.log4j.Logger;

/**
 * 
 * @author luc peuvrier
 * 
 */
public class MatchMgrInvoker {

	private final static Logger _log = Logger.getLogger(MatchMgrInvoker.class);

	private transient final ParsingToStringUtil parsingToStringUtil =
	/**/ParsingToStringUtil.getInstance();

	/** on match manager by non terminal */
	private final IMatchMgr[] matchMgrs;// NOPMD

	private transient boolean skipReduceAfterError;

	private transient ParserException reduceException;

	/**
	 * set match manager by identifier
	 * 
	 * @param matchClassNameTable
	 *            match management class name for non terminal
	 * @param nonTerminalByIdentifierMap
	 * @throws ParserException
	 */
	public MatchMgrInvoker(final String[] matchClassNameTable)
			throws ParserException {
		final int length = matchClassNameTable.length;
		matchMgrs = new IMatchMgr[length];
		try {
			for (int nonTerminalIndex = 0; nonTerminalIndex < length; nonTerminalIndex++) {
				final String className = matchClassNameTable[nonTerminalIndex];
				final IMatchMgr matchMgr;
				if (className == null) {
					matchMgr = null;
				} else {
					final Class<?> matchMgrClass = Class.forName(className);
					matchMgr = (IMatchMgr) matchMgrClass.newInstance();
				}
				matchMgrs[nonTerminalIndex] = matchMgr;
			}
		} catch (ClassNotFoundException exception) {
			throw new ParserException("missing match manager class", exception);
		} catch (InstantiationException exception) {
			throw new ParserException("failed instantiate match manager class",
					exception);
		} catch (IllegalAccessException exception) {
			throw new ParserException("failed instantiate match manager class",
					exception);
		}
	}

	public void initialize() {
		skipReduceAfterError = false;
	}

	public boolean isSkipReduceAfterError() {
		return skipReduceAfterError;
	}

	public ParserException getReduceException() {
		return reduceException;
	}

	/**
	 * 
	 * @param matchedNonterminalIdentifier
	 * @param parseNodes
	 * @param notWhiteIndexes
	 * @param parseSession
	 * @return true if reduction confirmed
	 * @throws ParserException
	 */
	public boolean confirmReduction(final int matchedNonterminalIdentifier,
			final IParseNode[] parseNodes, final int[] notWhiteIndexes,
			final ParseSession parseSession) throws ParserException {
		final IMatchMgr matchMgr =
		/**/matchMgrs[matchedNonterminalIdentifier - 0x10000];
		if (_log.isDebugEnabled()) {
			_log.debug("match manager is " + matchMgr);
		}
		/* no match manager is equivalent to reduction confirmation */
		boolean confirmReduction;
		if (matchMgr == null) {
			confirmReduction = true;
		} else {
			try {
				confirmReduction = matchMgr.confirmReduction(parseSession,
						parseNodes, notWhiteIndexes);
			} catch (ParserException exception) {// NOPMD
				throw exception;
			} catch (Exception exception) {
				confirmReduction = false;
				final String title = "doing confirm reduce";
				matchMgrError(matchedNonterminalIdentifier, parseNodes,
						notWhiteIndexes, exception, title);
			}
		}

		if (_log.isDebugEnabled()) {
			_log.debug(confirmReduction ? "reduction confirmed"
					: "reduction cancelled");
		}
		return confirmReduction;
	}

	/**
	 * for reduction action invoke match manager for matched non terminal
	 * 
	 * @param matchedNonterminalIdentifier
	 * @param fatherParseNode
	 * @param parseNodes
	 *            parse nodes subject to reduction
	 * @param notWhiteIndexes
	 * @throws ParserException
	 */
	public void reduceAction(final int matchedNonterminalIdentifier,
			final IParseNode fatherParseNode, final IParseNode[] parseNodes,
			final int[] notWhiteIndexes, final ParseSession parseSession)
			throws ParserException {
		try {
			if (_log.isDebugEnabled()) {
				final StringBuilder builder = new StringBuilder();
				parseNodesToString(builder, matchedNonterminalIdentifier,
						parseNodes, notWhiteIndexes);
				_log.debug(builder.toString());
			}
			final IMatchMgr matchMgr =
			/**/matchMgrs[matchedNonterminalIdentifier - 0x10000];
			if (_log.isDebugEnabled()) {
				_log.debug("match manager is " + matchMgr);
			}
			fatherParseNode.setMatchMgr(matchMgr);
			if (matchMgr != null && !skipReduceAfterError) {
				try {
					matchMgr.reduceAction(parseSession, fatherParseNode,
							parseNodes, notWhiteIndexes);
				} catch (ParserException exception) {// NOPMD
					throw exception;
				} catch (Exception exception) {
					final String title = "doing reduce action";
					matchMgrError(matchedNonterminalIdentifier, parseNodes,
							notWhiteIndexes, exception, title);
				}
			}
			if (_log.isDebugEnabled()) {
				final StringBuilder builder = new StringBuilder();
				builder.append("reduced to:\n");
				parseNodeToString(builder, fatherParseNode);
				_log.debug(builder.toString());
			}
		} catch (RuntimeException exception) {
			final StringBuilder builder = new StringBuilder();
			builder.append("runtime exception\n");
			parseNodesToString(builder, matchedNonterminalIdentifier,
					parseNodes, notWhiteIndexes);
			_log.error(builder.toString(), exception);
			throw exception;
		}
	}

	private void matchMgrError(final int matchedNonterminalIdentifier,
			final IParseNode[] parseNodes, final int[] notWhiteIndexes,
			final Exception exception, final String title) {
		final StringBuilder builder = new StringBuilder();
		builder.append(title);
		builder.append('\n');
		parseNodesToString(builder, matchedNonterminalIdentifier, parseNodes,
				notWhiteIndexes);
		final String message = builder.toString();
		_log.error(message, exception);
		reduceException = new ParserException(message, exception);
		skipReduceAfterError = true;
	}

	private void parseNodesToString(final StringBuilder builder,
			final int matchedNonterminalIdentifier,
			final IParseNode[] parseNodes, final int[] notWhiteIndexes) {
		builder.append(parsingToStringUtil
				.grammarSymbolIdentifierToString(matchedNonterminalIdentifier));
		builder.append(" <-");
		for (IParseNode parseNode : parseNodes) {
			builder.append(' ');
			builder.append(parsingToStringUtil
					.grammarSymbolIdentifierToString(parseNode
							.getGrammarSymbolIdentifier()));
		}
		builder.append("\nnot white space ");
		for (int nwi : notWhiteIndexes) {
			builder.append(nwi);
			builder.append(' ');
		}
		builder.append('\n');
		for (IParseNode parseNode : parseNodes) {
			parseNodeToString(builder, parseNode);
			builder.append('\n');
		}
	}

	private void parseNodeToString(final StringBuilder builder,
			final IParseNode parseNode) {
		builder.append(parsingToStringUtil
				.grammarSymbolIdentifierToString(parseNode
						.getGrammarSymbolIdentifier()));
		builder.append(": ");
		builder.append(parseNode.getParseNodeType());
		builder.append(", ");
		builder.append(parseNode.getClass().getSimpleName());
		if (parseNode.isNonTerminal()) {
			builder.append(", match mgr ");
			final IMatchMgr matchMgr = parseNode.getMatchMgr();
			if (matchMgr == null) {
				builder.append("null");
			} else {
				builder.append(matchMgr.getClass().getSimpleName());
			}
			builder.append(",");
			final IParseNodeData[] dataByAlternative = parseNode
					.getDataByAlternative();
			int index = 0;
			for (IParseNodeData parseNodeData : dataByAlternative) {
				builder.append(" data #");
				builder.append(index);
				builder.append(": ");
				if (parseNodeData == null) {
					builder.append("null");
				} else {
					builder.append(parseNodeData.getClass().getSimpleName());
				}
				index++;
			}
			if (parseNode.isWhiteSpace()) {
				builder.append(", is white space");
			} else {
				builder.append(", is not white space");
			}
		}
	}
}
