/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parser.service;

import java.io.IOException;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import net.sf.parser4j.parser.entity.ParseResult;
import net.sf.parser4j.parser.entity.ParseSession;
import net.sf.parser4j.parser.entity.data.GotoByNonTerminalEntry;
import net.sf.parser4j.parser.entity.data.GotoByTerminalEntry;
import net.sf.parser4j.parser.entity.data.MatchEntry;
import net.sf.parser4j.parser.entity.data.MatchEntrySet;
import net.sf.parser4j.parser.entity.data.NonTerminal;
import net.sf.parser4j.parser.entity.data.ParserData;
import net.sf.parser4j.parser.entity.data.TerminalCharRange;
import net.sf.parser4j.parser.entity.data.ToStateAndWhiteSpaceFlag;
import net.sf.parser4j.parser.entity.parsenode.CharacterParseNode;
import net.sf.parser4j.parser.entity.parsenode.IParseNode;
import net.sf.parser4j.parser.entity.parsenode.status.IParseNodeInErrorStatus;
import net.sf.parser4j.parser.entity.parsestate.IParseState;
import net.sf.parser4j.parser.entity.parsestate.LexParseState;
import net.sf.parser4j.parser.entity.parsestate.ParseStack;
import net.sf.parser4j.parser.entity.parsestate.ParseStackElement;
import net.sf.parser4j.parser.entity.parsestate.SyntaxParseState;
import net.sf.parser4j.parser.service.parsenode.ParseNodeFactory;
import net.sf.parser4j.parser.service.parsenode.ParsingToStringUtil;

import org.apache.log4j.Logger;

/**
 * The parser framework !<br>
 * 
 * @author luc peuvrier
 * 
 */
public class Parser {

	private static final Logger _log = Logger.getLogger(Parser.class);

	private static final ParserStackLogger parserStackLogger =
	/**/ParserStackLogger.getInstance();

	private final static ParseNodeFactory parseNodeFactory =
	/**/ParseNodeFactory.getInstance();

	/** special ending state number: 0 by Start to -1 when match all input */
	public static final int ENDING_STATE_NUMBER = -1;

	public static final ToStateAndWhiteSpaceFlag ENDING_STATE =
	/**/new ToStateAndWhiteSpaceFlag(ENDING_STATE_NUMBER, false);

	private final MatchMgrInvoker matchClassInvoker;// NOPMD

	private final Map<Integer, NonTerminal> nonTerminalByIdentifierMap;// NOPMD

	/** for each state is defined a goto state by terminal for non token */
	private final GotoByTerminalEntry[] gotoByTerminalTable;// NOPMD

	/** for each state is defined a goto state by terminal for token */
	private final GotoByTerminalEntry[] tokenGotoByTerminalTable;// NOPMD

	/** for each state is defined a goto state by non terminal for non token */
	private final GotoByNonTerminalEntry[] gotoByNonTerminalTable;// NOPMD

	/** for each state is defined a goto state by non terminal for token */
	private final GotoByNonTerminalEntry[] tokenGotoByNonTerminalTable;// NOPMD

	/** for each state define non terminal matched for non token */
	private final MatchEntrySet[] matchTable;// NOPMD

	/** for each state define non terminal matched for token */
	private final MatchEntrySet[] tokenMatchTable;// NOPMD

	/** for each state define if it is token recognition */
	private final boolean[] tokenRecognition;// NOPMD

	/** root non terminal identifier */
	private final int rootNonTerminalIdentifier;// NOPMD

	private final IParserListener listener;// NOPMD

	private transient ParseResult parseResult;// NOPMD

	/**
	 * true if last {@link #parse(ParserFileReader, boolean, boolean)} result
	 * have alternative for ambiguity
	 */
	private transient boolean alternativeForAmbiguity;

	/**
	 * last syntax parse stacks to shift from with input. computed by
	 * {@link #parse(ParserInput, Set)} and used by
	 * {@link #computeErrorInfo(Set, ParserInput)}<br>
	 * 
	 * it is the previous {@link #syntaxParseState} state before trying shift by
	 * grammar symbol
	 */
	private transient ParseStack[] parseStacks;

	/**
	 * last lexical parse stacks to shift from with input. computed by
	 * {@link #parse(ParserInput, Set)} and used by
	 * {@link #computeErrorInfo(Set, ParserInput)}<br>
	 * 
	 * it is the previous {@link #lexParseState} state before trying shift by
	 * grammar symbol
	 */
	private transient ParseStack[] tokenParseStacks;

	private transient final ParseSession parseSession;

	private transient int currentColumnNumber;

	private transient int currentLineNumber;

	private transient SyntaxParseState syntaxParseState;

	private transient LexParseState lexParseState;

	private transient ParserInput parserInput;

	private transient List<IParseNode> matchedTokenList;

	private transient boolean skipReduceAfterError;

	private transient ParserException reduceException;

	private transient int reduceErrorLineNumber;

	private transient int reduceErrorColumnNumber;

	private transient int maxNumberOfAlternative;

	public Parser(final ParserData parserData, final IParserListener listener)
			throws ParserException {
		super();
		this.listener = listener;
		final String[] matchClassNameTable = parserData
				.getMatchClassNameTable();
		matchClassInvoker = new MatchMgrInvoker(matchClassNameTable);
		gotoByTerminalTable = parserData.getGotoByTerminalTable();
		tokenGotoByTerminalTable = parserData.getTokenGotoByTerminalTable();
		gotoByNonTerminalTable = parserData.getGotoByNonTerminalTable();
		tokenGotoByNonTerminalTable = parserData
				.getTokenGotoByNonTerminalTable();
		matchTable = parserData.getMatchTable();
		tokenMatchTable = parserData.getTokenMatchTable();
		rootNonTerminalIdentifier = parserData.getRootNonTerminalIdentifier();
		tokenRecognition = parserData.getTokenRecognition();
		parserStackLogger.setTokenRecognition(tokenRecognition);
		parseSession = new ParseSession();
		nonTerminalByIdentifierMap = parserData.getNonTerminalByIdentifierMap();
		parserStackLogger
				.setNonTerminalByIdentifierMap(nonTerminalByIdentifierMap);
		maxNumberOfAlternative = 10;
	}

	/**
	 * @param maxNumberOfAlternative
	 *            the maxNumberOfAlternative to set
	 */
	public void setMaxNumberOfAlternative(final int maxNumberOfAlternative) {
		this.maxNumberOfAlternative = maxNumberOfAlternative;
	}

	/**
	 * @return the parseSession
	 */
	public ParseSession getParseSession() {
		return parseSession;
	}

	/**
	 * do parsing<b> close reader after parsing<br>
	 * 
	 * @param parserFileReader
	 *            reader of input to parse
	 * @param syntaxAcceptAmbiguity
	 * @param lexAcceptAmbiguity
	 * @return parsing result
	 * @throws ParserException
	 *             for errors
	 * @throws HasAmbiguityParserException
	 *             reject ambiguity
	 */
	public ParseResult parse(final ParserFileReader parserFileReader,
			final boolean syntaxAcceptAmbiguity,
			final boolean lexAcceptAmbiguity) throws ParserException,
			HasAmbiguityParserException {
		matchClassInvoker.initialize();
		skipReduceAfterError = false;
		parserInput = new ParserInput(parserFileReader);
		parseSession.setParserInput(parserInput);
		currentColumnNumber = parserInput.getColumnNumber();
		currentLineNumber = parserInput.getLineNumber();
		syntaxParseState = new SyntaxParseState(syntaxAcceptAmbiguity,
				maxNumberOfAlternative, tokenRecognition);
		final ParseStack initialParseStack = new ParseStack();
		syntaxParseState.add(initialParseStack);
		checkForMatch(true, initialParseStack);
		boolean syntaxMode = true;
		parseSession.setSyntaxMode(syntaxMode);
		boolean hasTokenRecognition = syntaxParseState.hasTokenRecognition();
		if (_log.isDebugEnabled()) {
			_log.debug("hasTokenRecognition " + hasTokenRecognition);
		}

		int terminalValue;
		do {
			if (syntaxMode && hasTokenRecognition) {
				parserInput.mark();
				matchedTokenList = null;// NOPMD
				syntaxMode = false;
				parseSession.setSyntaxMode(syntaxMode);
				final List<ParseStack> list = syntaxParseState
						.poolTokenRecognitionParseStackList();
				lexParseState = newLexParseState(lexAcceptAmbiguity, list);
				final ParseStack[] parseStacks = lexParseState.toArray();
				for (ParseStack lexParseStack : parseStacks) {
					checkForMatch(syntaxMode, lexParseStack);
				}
				if (_log.isDebugEnabled()) {
					_log
							.debug("have token match "
									+ (matchedTokenList != null));
				}
			}
			if (syntaxMode) {

				lexParseState = null;// NOPMD
				do {
					if (parserStackLogger.isDebugEnabled()) {
						parserStackLogger.logParseState("syntax: before parse",
								syntaxParseState);
					}
					terminalValue = parse(true/* syntaxMode */);
					hasTokenRecognition = syntaxParseState
							.hasTokenRecognition();
					if (_log.isDebugEnabled()) {
						_log
								.debug("hasTokenRecognition "
										+ hasTokenRecognition);
					}
				} while (terminalValue != -1 && !syntaxParseState.isEmpty()
						&& !hasTokenRecognition);

			} else {// lexical analysis mode

				do {
					if (parserStackLogger.isDebugEnabled()) {
						parserStackLogger.logParseState("lex: before parse",
								lexParseState);
					}
					terminalValue = parse(false/* syntaxMode */);
				} while (terminalValue != -1 && !lexParseState.isEmpty());

				syntaxMode = true;
				parseSession.setSyntaxMode(syntaxMode);

				if (matchedTokenList == null) {
					if (_log.isDebugEnabled()) {
						_log.debug("no token match");
					}
					hasTokenRecognition = false;
				} else {
					if (parserStackLogger.isDebugEnabled()) {
						parserStackLogger.logParseState(
								"syntax: before shift by token",
								syntaxParseState);
					}
					shiftByToken();

					hasTokenRecognition = syntaxParseState
							.hasTokenRecognition();
				}
				matchedTokenList = null;// NOPMD
				parserInput.reset();
				terminalValue = 0;// dummy value not equals to -1
			}
		} while (terminalValue != -1 && !syntaxParseState.isEmpty());

		if (_log.isDebugEnabled()) {
			_log.debug("parse end: end of input=" + (terminalValue == -1)
					+ ", syntax parse state empty="
					+ syntaxParseState.isEmpty());
		}
		if (parserStackLogger.isDebugEnabled()) {
			parserStackLogger.logParseState("syntax: end of parsing",
					syntaxParseState);
			parserStackLogger.logParseState("lex: end of parsing",
					lexParseState);
		}

		final IParseNode resultParseNode = computeResult();
		alternativeForAmbiguity = syntaxParseState.hasAlternativeForAmbiguity();
		if (resultParseNode == null) {
			parseResult = new ParseResult(true);
			parseResult.setLastTerminalValueReaded(terminalValue);
			parseResult.setResultParseNode(resultParseNode);
			parseResult.setParseError(true);
			parseResult.setReduceError(false);
			computeErrorInfo();
		} else {
			final boolean reduceError = skipReduceAfterError
					|| resultParseNode.isInError();
			parseResult = new ParseResult(reduceError);
			parseResult.setLastTerminalValueReaded(terminalValue);
			parseResult.setResultParseNode(resultParseNode);
			parseResult.setReduceError(reduceError);
			parseResult.setParseError(false);
			if (skipReduceAfterError) {
				_log.error("reduce error at " + reduceErrorLineNumber + ","
						+ reduceErrorColumnNumber, reduceException);
				parseResult.setReduceErrorInfo(reduceException,
						reduceErrorLineNumber, reduceErrorColumnNumber,
						parserInput.getFileName());
			} else if (resultParseNode.isInError()) {
				final IParseNodeInErrorStatus errorStatus = resultParseNode
						.getErrorStatus();
				_log.error("reduce error at "
						+ errorStatus.getBeginLineNumber() + ","
						+ errorStatus.getBeginColumnNumber() + " thru "
						+ errorStatus.getEndLineNumber() + ","
						+ errorStatus.getEndColumnNumber() + "\n"
						+ errorStatus.toString());
				parseResult.setReduceErrorInfo(errorStatus, parserInput
						.getFileName());
			}
		}
		parserInput.close();
		return parseResult;
	}

	private LexParseState newLexParseState(final boolean lexAcceptAmbiguity,
			final List<ParseStack> list) throws ParserException,
			HasAmbiguityParserException {
		return new LexParseState(
		/**/lexAcceptAmbiguity, maxNumberOfAlternative, tokenRecognition, list);
	}

	private String mode(final boolean syntaxMode) {
		return syntaxMode ? "syntax" : "lex";
	}

	public boolean hasAlternativeForAmbiguity() {
		return alternativeForAmbiguity;
	}

	/**
	 * compute parsing result
	 * 
	 * @return parsing result
	 */
	private IParseNode computeResult() {
		IParseNode resultParseNode = null;
		for (ParseStack parseStack : syntaxParseState) {
			if (parseStack.isMatchAll()) {
				if (resultParseNode == null) {
					resultParseNode = parseStack.getMatchAllParseNode();
				} else {
					throw new IllegalStateException(
							"can not have more than one stack matching grammar");
				}
			}
		}
		return resultParseNode;
	}

	/**
	 * compute error information in {@link #parseResult} from
	 * {@link #parseStacks},{@link #tokenParseStacks} and {@link #parserInput}
	 * 
	 * 
	 */
	private void computeErrorInfo() {

		parseResult.setParseStacks(parseStacks, tokenParseStacks,
				tokenRecognition);

		/* collect all ending state number */
		final Set<Integer> endingState = new TreeSet<Integer>();

		for (ParseStack parseStack : parseStacks) {
			final int stateNumber = parseStack.getStateNumber();
			if (stateNumber != ENDING_STATE_NUMBER) {
				endingState.add(stateNumber);
			}
		}
		if (lexParseState != null) {
			for (ParseStack parseStack : tokenParseStacks) {
				final int stateNumber = parseStack.getStateNumber();
				if (stateNumber != ENDING_STATE_NUMBER) {
					endingState.add(stateNumber);
				}
			}
		}

		final Set<TerminalCharRange> expectedTerminal = parseResult
				.getExpectedTerminal();
		final Set<String> expectedNonTerminal = parseResult
				.getExpectedNonTerminal();
		/* compute all expected grammar symbol */
		for (int stateNumber : endingState) {
			expectedTerminal
					.addAll(gotoByTerminalTable[stateNumber].expected());
			expectedTerminal.addAll(tokenGotoByTerminalTable[stateNumber]
					.expected());
			Set<Integer> set = gotoByNonTerminalTable[stateNumber].expected();
			for (int nonTerminalIdentifier : set) {
				expectedNonTerminal
						.add(grammarSymbolToString(nonTerminalIdentifier));
			}
			set = tokenGotoByNonTerminalTable[stateNumber].expected();
			for (int nonTerminalIdentifier : set) {
				expectedNonTerminal
						.add(grammarSymbolToString(nonTerminalIdentifier));
			}
		}
		parseResult.setInTextPosition(parserInput.getLineNumber(), parserInput
				.getColumnNumber() - 1, parserInput.getFileName());

	}

	/**
	 * was for debug<br>
	 * 
	 * @param parseStack
	 * @param expectedStateNumberList
	 *            ending state number list
	 * @return true if parseStack end with state number list
	 */
	@SuppressWarnings("unused")
	private boolean parseStackEndWith(final ParseStack parseStack,
			final int[] expectedStateNumberList) {
		int index = expectedStateNumberList.length - 1;
		boolean found = true;
		if (parseStack.getStateNumber() != expectedStateNumberList[index--]) {
			found = false;
		}
		if (found) {
			final Deque<ParseStackElement> que = parseStack
					.getParseStateNodeStack();
			final Iterator<ParseStackElement> iterator = que
					.descendingIterator();
			while (index >= 0 && found) {
				if (iterator.hasNext()) {
					final int stateNumber = iterator.next().getStateNumber();
					found = stateNumber == expectedStateNumberList[index--];
				} else {
					found = false;
				}
			}
		}
		return found;
	}

	/**
	 * 
	 * @return terminal value
	 * @throws ParserException
	 * @throws HasAmbiguityParserException
	 */
	private int parse(final boolean syntaxMode) throws ParserException,
			HasAmbiguityParserException {
		/* get position in input and read current terminal value */
		currentColumnNumber = parserInput.getColumnNumber();
		currentLineNumber = parserInput.getLineNumber();
		final int terminalValue = readTerminalValue();
		/* if not end of input */
		if (terminalValue != -1) {
			parse(syntaxMode, terminalValue);
		}
		return terminalValue;
	}

	/**
	 * 
	 * @param syntaxMode
	 * @param terminalValue
	 * @throws ParserException
	 * @throws HasAmbiguityParserException
	 */
	private void parse(final boolean syntaxMode, final int terminalValue)
			throws ParserException, HasAmbiguityParserException {
		final IParseNode parseNode =
		/**/new CharacterParseNode(
		/*  */terminalValue, currentLineNumber, currentColumnNumber);
		final IParseState parseState = syntaxMode ? syntaxParseState
				: lexParseState;
		final ParseStack[] localParseStacks = parseState.toArray();
		if (syntaxMode) {
			parseStacks = localParseStacks;
		} else {
			tokenParseStacks = localParseStacks;
		}
		parseState.clear();

		final List<ParseStack> checkForMatchList = new LinkedList<ParseStack>();
		for (ParseStack currentParseStack : localParseStacks) {
			checkShiftByGrammarSymbol(syntaxMode, terminalValue, parseNode,
					checkForMatchList, currentParseStack);
		}
		if (parserStackLogger.isDebugEnabled()) {
			parserStackLogger.logParseState(mode(syntaxMode)
					+ ": after shift by terminal",
					syntaxMode ? syntaxParseState : lexParseState);
		}
		for (ParseStack currentParseStack : checkForMatchList) {
			checkForMatch(syntaxMode, currentParseStack);
		}
	}

	private void shiftByToken() throws ParserException,
			HasAmbiguityParserException {
		final ParseStack[] parseStacks = syntaxParseState.toArray();
		final List<ParseStack> checkForMatchList = new LinkedList<ParseStack>();
		syntaxParseState.clear();
		for (IParseNode parseNode : matchedTokenList) {
			for (ParseStack currentParseStack : parseStacks) {
				final int nonTerminalIdentifier = parseNode
						.getGrammarSymbolIdentifier();
				checkShiftByGrammarSymbol(true/* syntaxMode */,
						nonTerminalIdentifier, parseNode, checkForMatchList,
						currentParseStack);
			}
		}

		if (parserStackLogger.isDebugEnabled()) {
			parserStackLogger.logParseState("syntax: after shift by token",
					syntaxParseState);
		}
		for (ParseStack currentParseStack : checkForMatchList) {
			checkForMatch(true/* syntaxMode */, currentParseStack);
		}
	}

	/**
	 * 
	 * @param syntaxMode
	 * @param grammarSymbolIdentifier
	 * @param parseNode
	 * @param checkForMatchList
	 * @param parseStack
	 * @throws ParserException
	 * @throws HasAmbiguityParserException
	 */
	private void checkShiftByGrammarSymbol(final boolean syntaxMode,
			final int grammarSymbolIdentifier, final IParseNode parseNode,
			final List<ParseStack> checkForMatchList,
			final ParseStack parseStack) throws ParserException,
			HasAmbiguityParserException {
		final int stateNumber = parseStack.getStateNumber();
		final ToStateAndWhiteSpaceFlag toStateAndWhiteSpaceFlag = toState(
				syntaxMode, stateNumber, grammarSymbolIdentifier);
		if (toStateAndWhiteSpaceFlag == null) {
			if (_log.isDebugEnabled()) {
				_log.debug(mode(syntaxMode) + ": no shift from " + stateNumber
						+ " by "
						+ grammarSymbolToString(grammarSymbolIdentifier) + " ("
						+ grammarSymbolIdentifier + ")");
			}
		} else {
			shift(syntaxMode, toStateAndWhiteSpaceFlag, parseStack, parseNode,
					grammarSymbolIdentifier);
			(syntaxMode ? syntaxParseState : lexParseState).add(parseStack);
			checkForMatchList.add(parseStack);
		}
	}

	/**
	 * read terminal value in {@link #parserInput}
	 * 
	 * @return current terminal value
	 * @throws ParserException
	 */
	private int readTerminalValue() throws ParserException {
		final int terminalValue;
		try {
			terminalValue = parserInput.read();
		} catch (IOException exception) {
			throw new ParserException(exception);
		}
		if (terminalValue == '\n' && listener != null) {
			listener.newLine(parserInput.getLineNumber());
		}
		if (_log.isDebugEnabled()) {
			final String toString = ParsingToStringUtil.getInstance()
					.terminalIdentifierToString(terminalValue);
			_log.debug("read " + toString + " (" + parserInput.getLineNumber()
					+ "," + parserInput.getColumnNumber() + ")");
		}
		return terminalValue;
	}

	/**
	 * @param syntaxMode
	 * @param toStateAndWhiteSpaceFlag
	 *            to state from current last state in parse stack
	 * @param parseStack
	 *            parse stack to update
	 * @param parseNode
	 *            from state to state by this parseNode ( parse not for grammar
	 *            symbol [terminal or non terminal] )
	 * @param grammarSymbolIdentifier
	 *            from state to state by this grammar symbol know by its
	 *            identifier
	 * @throws ParserException
	 * @throws HasAmbiguityParserException
	 *             reject ambiguity
	 */
	private void shift(final boolean syntaxMode,
			final ToStateAndWhiteSpaceFlag toStateAndWhiteSpaceFlag,
			final ParseStack parseStack, final IParseNode parseNode,
			final int grammarSymbolIdentifier) throws ParserException,
			HasAmbiguityParserException {
		if (_log.isDebugEnabled()) {
			_log.debug(mode(syntaxMode) + ": shift from "
					+ parseStack.getStateNumber() + " to "
					+ toStateAndWhiteSpaceFlag + " by "
					+ grammarSymbolToString(grammarSymbolIdentifier) + " ("
					+ grammarSymbolIdentifier + "), white space="
					+ toStateAndWhiteSpaceFlag.isWhiteSpace());
		}
		parseNode.setWhiteSpace(toStateAndWhiteSpaceFlag.isWhiteSpace());
		parseStack.shift(parseNode, toStateAndWhiteSpaceFlag.getToState());
	}

	/**
	 * @param syntaxMode
	 * @param parseStack
	 *            parse stack to update
	 * @throws ParserException
	 * @throws HasAmbiguityParserException
	 *             reject ambiguity
	 */
	private void checkForMatch(final boolean syntaxMode,
			final ParseStack parseStack) throws ParserException,
			HasAmbiguityParserException {
		final int stateNumber = parseStack.getStateNumber();
		final MatchEntrySet matchEntrySet;
		if (syntaxMode) {
			matchEntrySet = matchTable[stateNumber];
		} else {
			matchEntrySet = tokenMatchTable[stateNumber];
		}
		if (matchEntrySet.match()) {
			matchManagement(syntaxMode, parseStack, matchEntrySet);
		}
	}

	/**
	 * match management: manage all non terminal matched
	 * 
	 * @param syntaxMode
	 * @param parseStack
	 *            parse stack to update
	 * @param matchEntrySet
	 *            set of non terminal identifier matched informations
	 * @throws ParserException
	 * @throws HasAmbiguityParserException
	 */
	private void matchManagement(final boolean syntaxMode,
			final ParseStack parseStack, final MatchEntrySet matchEntrySet)
			throws ParserException, HasAmbiguityParserException {

		final Iterator<MatchEntry> entrySetIterator = matchEntrySet
				.matchEntrySetIterator();
		final List<IParseNode> currentMatchedTokenList =
		/**/new LinkedList<IParseNode>();
		while (entrySetIterator.hasNext()) {
			final MatchEntry matchEntry = entrySetIterator.next();
			final int matchedNonterminalIdentifier = matchEntry
					.getMatchedNonTerminalIdentifier();
			final int matchedLength = matchEntry.getMatchedLength();
			final IParseNode[] parseNodes = parseStack
					.getMatchedParseNode(matchedLength);
			final int[] notWhiteIndexes = parseStack.getNotWhiteIndexes();
			/* error status from son node, null if none */
			final IParseNodeInErrorStatus parseNodeInErrorStatus = parseStack
					.getParseNodeInErrorStatus();

			/*
			 * do reduction if confirmed
			 */
			if (matchClassInvoker.confirmReduction(
					matchedNonterminalIdentifier, parseNodes, notWhiteIndexes,
					parseSession)) {
				/*
				 * match not canceled
				 */
				if (_log.isDebugEnabled()) {
					_log
							.debug(mode(syntaxMode)
									+ ": state "
									+ parseStack.getStateNumber()
									+ " match "
									+ grammarSymbolToString(matchedNonterminalIdentifier)
									+ " (" + matchedNonterminalIdentifier + ")");
				}

				final int nodeType = matchEntry.getNodeType();
				final IParseNode reduceParseNode = reduce(syntaxMode,
						parseStack, matchedNonterminalIdentifier,
						matchedLength, nodeType, parseNodes, notWhiteIndexes,
						parseNodeInErrorStatus);
				if (!syntaxMode) {// if lex mode
					checkForTokenMatch(reduceParseNode, currentMatchedTokenList);
				}
			}
		}
		if (!currentMatchedTokenList.isEmpty()) {
			parserInput.mark();
			matchedTokenList = currentMatchedTokenList;
		}
	}

	/**
	 * check for token match in lexical analysis mode
	 * 
	 * @param reduceParseNode
	 * @param currentMatchedTokenList
	 */
	private void checkForTokenMatch(final IParseNode reduceParseNode,
			final List<IParseNode> currentMatchedTokenList) {
		final int nonTerminalIdentifier = reduceParseNode
				.getGrammarSymbolIdentifier();
		final NonTerminal nonTerminal = nonTerminalByIdentifierMap
				.get(nonTerminalIdentifier);
		if (nonTerminal.isToken()) {
			currentMatchedTokenList.add(reduceParseNode);
			if (_log.isDebugEnabled()) {
				_log.debug("token match "
						+ grammarSymbolToString(nonTerminalIdentifier));
			}
		}
	}

	/**
	 * do the reduction
	 * 
	 * @param syntaxMode
	 * @param parseStack
	 * @param matchedNonterminalIdentifier
	 * @param matchedLength
	 * @param nodeType
	 * @param parseNodes
	 * @param notWhiteIndexes
	 * @param parseNodeInErrorStatus
	 *            error status from son node, null if none
	 * @return reduce parse node, null if match start or exist in parse state
	 * @throws ParserException
	 * @throws HasAmbiguityParserException
	 */
	private IParseNode reduce(final boolean syntaxMode,
			final ParseStack parseStack,
			final int matchedNonterminalIdentifier, final int matchedLength,
			final int nodeType, final IParseNode[] parseNodes,
			final int[] notWhiteIndexes,
			final IParseNodeInErrorStatus parseNodeInErrorStatus)
			throws ParserException, HasAmbiguityParserException {
		final IParseNode result;
		final IParseNode reduceParseNode = parseNodeFactory.create(nodeType,
				parseNodes);
		if (parseNodeInErrorStatus != null) {
			reduceParseNode.setErrorStatus(parseNodeInErrorStatus);
		}
		reduceParseNode
				.setGrammarSymbolIdentifier(matchedNonterminalIdentifier);
		final ParseStack parseStackCopy = parseStack.copy();
		parseStackCopy.reduce(matchedLength);
		final IParseState parseState = syntaxMode ? syntaxParseState
				: lexParseState;
		if (matchedNonterminalIdentifier == rootNonTerminalIdentifier) {
			/*
			 * match all the grammar
			 */
			if (!syntaxMode) {
				throw new ParserException("start must not be a token");
			}
			parseStackCopy.setMatchAll(reduceParseNode);
			shift(syntaxMode, ENDING_STATE, parseStackCopy, reduceParseNode,
					matchedNonterminalIdentifier);
			parseState.add(parseStackCopy);
			reduceAction(matchedNonterminalIdentifier, parseNodes,
					notWhiteIndexes, reduceParseNode, parseNodeInErrorStatus);
			if (_log.isDebugEnabled()) {
				_log.debug("match all the grammar");
			}
			result = null;
		} else {
			final int stateNumber = parseStackCopy.getStateNumber();
			final ToStateAndWhiteSpaceFlag toStateAndWhiteSpaceFlag = toState(
					syntaxMode, stateNumber, matchedNonterminalIdentifier);
			if (toStateAndWhiteSpaceFlag == null) {
				if (syntaxMode) {
					result = null;
					throw new ParserException(
							mode(syntaxMode)
									+ ": must shift from "
									+ stateNumber
									+ " by "
									+ grammarSymbolToString(matchedNonterminalIdentifier));
				} else {
					reduceAction(matchedNonterminalIdentifier, parseNodes,
							notWhiteIndexes, reduceParseNode,
							parseNodeInErrorStatus);
					result = reduceParseNode;
				}
			} else {
				/* have from state number to state by identifier */
				shift(syntaxMode, toStateAndWhiteSpaceFlag, parseStackCopy,
						reduceParseNode, matchedNonterminalIdentifier);
				final boolean added = parseState.add(parseStackCopy);
				if (added) {
					reduceAction(matchedNonterminalIdentifier, parseNodes,
							notWhiteIndexes, reduceParseNode,
							parseNodeInErrorStatus);
					checkForMatch(syntaxMode, parseStackCopy);
					result = reduceParseNode;
				} else {
					result = null;
				}
			}
		}
		return result;
	}

	private void reduceAction(final int matchedNonterminalIdentifier,
			final IParseNode[] parseNodes, final int[] notWhiteIndexes,
			final IParseNode fatherParseNode,
			final IParseNodeInErrorStatus parseNodeInErrorStatus)
			throws ParserException {
		final int beginLineNumber;
		final int beginColumnNumber;
		final int endLineNumber;
		final int endColumnNumber;
		if (parseNodes.length == 0) {
			beginColumnNumber = endColumnNumber = currentColumnNumber;
			beginLineNumber = endLineNumber = currentLineNumber;
		} else {
			beginLineNumber = parseNodes[0].getBeginLineNumber();
			beginColumnNumber = parseNodes[0].getBeginColumnNumber();
			endLineNumber = parseNodes[parseNodes.length - 1]
					.getEndLineNumber();
			endColumnNumber = parseNodes[parseNodes.length - 1]
					.getEndColumnNumber();
		}
		fatherParseNode.setLineAndColumnNumber(beginLineNumber,
				beginColumnNumber, endLineNumber, endColumnNumber);

		if (parseNodeInErrorStatus == null) {
			matchClassInvoker.reduceAction(matchedNonterminalIdentifier,
					fatherParseNode, parseNodes, notWhiteIndexes, parseSession);
		}
		/*
		 * record first reduce error information, only first since the next
		 * reduce action will be skipped
		 */
		if (!skipReduceAfterError && matchClassInvoker.isSkipReduceAfterError()) {
			skipReduceAfterError = true;
			reduceErrorLineNumber = parserInput.getLineNumber();
			reduceErrorColumnNumber = parserInput.getColumnNumber();
			reduceException = matchClassInvoker.getReduceException();
		}
	}

	/**
	 * obtains from state to state by grammar symbol
	 * 
	 * @param syntaxMode
	 * @param stateNumber
	 *            from state
	 * @param grammarSymbolIdentifier
	 *            by grammar symbol
	 * @return to state from state by grammar symbol, null if no goto exist
	 */
	private ToStateAndWhiteSpaceFlag toState(final boolean syntaxMode,
			final int stateNumber, final int grammarSymbolIdentifier) {
		final ToStateAndWhiteSpaceFlag toStateAndWhiteSpaceFlag;
		if (stateNumber == -1) {
			toStateAndWhiteSpaceFlag = null;
			if (_log.isDebugEnabled()) {
				_log.debug(mode(syntaxMode)
						+ ": cancel match all grammar by grammar symbol "
						+ grammarSymbolToString(grammarSymbolIdentifier));
			}
		} else if (syntaxMode) {
			if (grammarSymbolIdentifier < 0x10000) {
				final GotoByTerminalEntry gotoByTerminalEntry =
				/**/gotoByTerminalTable[stateNumber];
				final int toState = gotoByTerminalEntry
						.toState(grammarSymbolIdentifier);
				if (toState == -1) {
					toStateAndWhiteSpaceFlag = null;
				} else {
					toStateAndWhiteSpaceFlag = new ToStateAndWhiteSpaceFlag(
							toState, false);
				}
			} else {
				final GotoByNonTerminalEntry gotoByNonTerminalEntry =
				/**/gotoByNonTerminalTable[stateNumber];
				toStateAndWhiteSpaceFlag = gotoByNonTerminalEntry
						.toStateAndWhiteSpaceFlag(grammarSymbolIdentifier);
			}
		} else {
			if (grammarSymbolIdentifier < 0x10000) {
				final int toState = tokenGotoByTerminalTable[stateNumber]
						.toState(grammarSymbolIdentifier);
				if (toState == -1) {
					toStateAndWhiteSpaceFlag = null;
				} else {
					toStateAndWhiteSpaceFlag = new ToStateAndWhiteSpaceFlag(
							toState, false);
				}
			} else {
				final GotoByNonTerminalEntry gotoByNonTerminalEntry =
				/**/tokenGotoByNonTerminalTable[stateNumber];
				toStateAndWhiteSpaceFlag = gotoByNonTerminalEntry
						.toStateAndWhiteSpaceFlag(grammarSymbolIdentifier);
			}
		}
		return toStateAndWhiteSpaceFlag;
	}

	/**
	 * facility method
	 * 
	 * @param grammarSymbolIdentifier
	 *            grammar symbol code
	 * @return name of grammar symbol
	 */
	private String grammarSymbolToString(final int grammarSymbolIdentifier) {
		return ParsingToStringUtil.getInstance()
				.grammarSymbolIdentifierToString(grammarSymbolIdentifier);
	}
}
