/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parser.service.data;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;

import net.sf.parser4j.parser.entity.data.ParserData;
import net.sf.parser4j.parser.service.ParserException;

import org.apache.log4j.Logger;

/**
 * {@link ParserData} input/output in file<br>
 * 
 * @author luc peuvrier
 * 
 */
public final class ParserDataIO {

	private static final Logger _log = Logger.getLogger(ParserDataIO.class);

	private static final ParserDataIO INSTANCE = new ParserDataIO();

	private ParserDataIO() {
		super();
	}

	public static ParserDataIO getInstance() {
		return INSTANCE;
	}

	public ParserData readFromClassPackage(final Class<?> clazz,
			final String resourceName) throws ParserException {
		final String path = clazz.getPackage().getName().replace('.', '/')
				+ "/" + resourceName;
		return readFromClassPath(path);
	}

	public ParserData readFromClassPath(final String path)
			throws ParserException {
		final ParserData parserData;
		final ClassLoader classLoader = // getClass().getClassLoader();
		Thread.currentThread().getContextClassLoader();
		final InputStream inputStream = classLoader.getResourceAsStream(path);
		if (inputStream == null) {
			throw new IllegalArgumentException(path
					+ " resource does not exist");
		}
		parserData = readFromInputStream(path, inputStream);
		return parserData;
	}

	public ParserData readFromFile(final File parserDataFile)
			throws ParserException {
		InputStream inputStream;
		try {
			inputStream = new FileInputStream(parserDataFile);
			final ParserData parserData;
			parserData = readFromInputStream(parserDataFile.getName(),
					inputStream);
			return parserData;
		} catch (FileNotFoundException exception) {
			throw new ParserException(exception);
		}
	}

	private ParserData readFromInputStream(final String path,
			final InputStream inputStream) throws ParserException {
		if (inputStream == null) {
			throw new IllegalArgumentException("input stream must be defined");
		}
		final ParserData parserData;
		try {
			final ObjectInputStream objectInputStream = new ObjectInputStream(
					inputStream);
			try {
				parserData = (ParserData) objectInputStream.readObject();
			} catch (ClassNotFoundException exception) {
				throw new ParserException("loading resource " + path, exception);
			} finally {
				objectInputStream.close();
			}
		} catch (IOException exception) {
			try {
				inputStream.close();
			} catch (IOException exception2) {
				_log.error("closing", exception2);
			}
			throw new ParserException("loading resource " + path, exception);
		}
		return parserData;
	}

	public void writeToFile(final File parserDataOutputFile,
			final ParserData parserData) throws ParserException {
		OutputStream outputStream = null;
		try {
			outputStream = new FileOutputStream(parserDataOutputFile);
			ObjectOutputStream objectOutputStream = null;
			try {
				objectOutputStream = new ObjectOutputStream(outputStream);
				objectOutputStream.writeObject(parserData);
			} catch (IOException exception) {
				throw new ParserException("saving resource "
						+ parserDataOutputFile, exception);
			} finally {
				try {
					objectOutputStream.close();
				} catch (IOException exception2) {
					_log.error("closing", exception2);
				}
			}

		} catch (FileNotFoundException exception) {
			if (outputStream != null) {
				try {
					outputStream.close();
				} catch (IOException exception2) {
					_log.error("closing", exception2);
				}
			}
			throw new ParserException(
					"saving resource " + parserDataOutputFile, exception);
		}
	}
}
