/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parser.service.parsenode;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.parser4j.parser.entity.ParseTreeInfo;
import net.sf.parser4j.parser.entity.data.NonTerminal;
import net.sf.parser4j.parser.entity.parsenode.IParseNode;
import net.sf.parser4j.parser.service.HasAmbiguityParserException;
import net.sf.parser4j.parser.service.ParserException;

import com.thoughtworks.xstream.XStream;

/**
 * output parse tree ( through its root parse node ) to xml in XStream forms
 * 
 * @author luc peuvrier
 * 
 */
public final class ParseTreeToXML {

	private static final ParseTreeToXML INSTANCE =
	/**/new ParseTreeToXML();

	public static ParseTreeToXML getInstance() {
		return INSTANCE;
	}

	private ParseTreeToXML() {
		super();
	}

	/**
	 * write parse tree with ambiguity in xml form
	 * 
	 * @param filePrefix
	 *            the file prefix for all output
	 * @param exception
	 *            in witch are stored parse tree with ambiguity
	 * @param nonTerminalByIdentifierMap
	 *            map between non terminal name and its identifier
	 * @throws ParserException
	 */
	public void toXml(final String filePrefix,
			final HasAmbiguityParserException exception,
			final Map<Integer, NonTerminal> nonTerminalByIdentifierMap)
			throws ParserException {
		final File prefix = new File(filePrefix);
		final File dirFile = prefix.getParentFile();
		final File[] fileList = dirFile.listFiles();
		final Pattern pattern = Pattern
				.compile(prefix.getName() + "\\d+\\.xml");
		for (File file : fileList) {
			final Matcher matcher = pattern.matcher(file.getName());
			if (matcher.matches()) {
				file.delete();
			}
		}
		final List<IParseNode> parseNodeList = exception
				.getWithAlternativeParseNodeList();
		int index = 0;
		for (IParseNode rootParseNode : parseNodeList) {
			final File outputFile =
			/**/new File(filePrefix + index + ".xml");// NOPMD
			toXml(outputFile, rootParseNode, nonTerminalByIdentifierMap);
			index++;
		}
	}

	/**
	 * write parse tree in xml form
	 * 
	 * @param outputFile
	 *            the file to store the parse tree
	 * @param rootParseNode
	 *            the parse tree root parse node
	 * @param nonTerminalByIdentifierMap
	 *            map between non terminal name and its identifier
	 * @throws ParserException
	 */
	public void toXml(final File outputFile, final IParseNode rootParseNode,
			final Map<Integer, NonTerminal> nonTerminalByIdentifierMap)
			throws ParserException {
		final XStream stream = new XStream();
		FileOutputStream outputStream = null;
		try {
			outputStream = new FileOutputStream(outputFile);
			final ParseTreeInfo parseTreeInfo = new ParseTreeInfo(
					rootParseNode, nonTerminalByIdentifierMap);
			stream.toXML(parseTreeInfo, outputStream);
		} catch (FileNotFoundException exception) {
			throw new ParserException(exception);
		} finally {
			if (outputStream != null) {
				try {
					outputStream.close();
				} catch (IOException exception) {// NOPMD
					// ignore error
				}
			}
		}
	}
}
